/* axes.c  */

#include <stdio.h>
#include <fcntl.h>
#include <math.h>
#include "specto.h"

#define MXNT    100
#define pow2(x) 	pow(2.0,(x))
#define pow10(x) 	pow(10.0,(x))

int     lpg_x();
int     lpg_y();
double  lpg_d();
double  ceil();
double  txtlen();
void    e_axis();
void    f_axis();
void    w_axis();
void    f_annot();
void    t_annot();
void    f_label();
void    t_label();
void    e_label();
void    w_label();
void    lpgadvto(int);
void    lpgrule(long, long, long, long);
void    lpgchup(double, double);
void    lpgchlw(double);
void    lpgchsz(double);
void    lpgchal(int, int);
void    lpgchxt(int);
void    lpgtext(double, double, char *);

static int ndx = 0;
static int ndy = 0;
static int t_tui = 0;
static int t_ndp = 0;
static int t_tts = 0;
static int t_annskp = 0;
static double max_ann_wid = 0;
static double max_tic_wid = 0;
static double lab_pos = 0;
static double t_ann_pos = 0;
static double last_time_ann = 0;
static double t_dt = 0;
static double t_tint = 0;
static double usf[] =
{1e-3, 1, 1e3, 1e6, 1e9};
static char *tu[] =
{"ksec", "sec", "msec", "usec", "nsec"};

extern int inf_type;

void
axes(int flg)
{
    int     any, n;
    double  margin, ttpi, aw;
    long    rx1, ry1, rxw, ryh;

    n = ucon_.lftmrgn - tick_.t_ys1;
    any = bpar_.eaxis || bpar_.faxis || bpar_.taxis || bpar_.waxis;
    ndx = (int) ((fpar_.axlnwd < lpg_d()) ? 1 : 1.5 + lpg_x(fpar_.axlnwd - lpg_d()));
    ndy = (int) ((fpar_.axlnwd < lpg_d()) ? 1 : 1.5 + lpg_y(fpar_.axlnwd - lpg_d()));
    if (flg == 1 && (n < 0 || !any)) {
	lpgadvto(ucon_.strtpos);
    } else if (flg == 1) {
	margin = fpar_.annsiz * 0.5 + (bpar_.taxis ? fpar_.ticsiz : 0.0);
	t_tui = (ucon_.tom + 5) / 3;
	t_ndp = (ucon_.tom % 3) ? 0 : 1;
	t_ann_pos = fpar_.yllc - margin;

	ttpi = (1 / (tick_.t_ti * ucon_.inps));
	if (ttpi < 1) {
	    t_tts = 1;
	} else if (ttpi < 2) {
	    t_tts = 2;
	} else if (ttpi < 5) {
	    t_tts = 5;
	} else {
	    t_tts = 1;
	}
	t_tint = (ttpi < 5) ? tick_.t_ti * 0.1 : tick_.t_ti;

	aw = (5 * fpar_.annsiz) / (ucon_.inps * tick_.t_ti);
	if (aw < 1) {
	    t_annskp = 1;
	} else if (aw < 2) {
	    t_annskp = 2;
	} else if (aw < 5) {
	    t_annskp = 5;
	} else {
	    t_annskp = 10;
	}
	t_dt = tick_.t_ti * t_annskp;

	max_tic_wid = any ? fpar_.ticsiz : 0.0;
	f_annot();
	margin = max_tic_wid + max_ann_wid + fpar_.annsiz + fpar_.labsiz * 0.5;
	lab_pos = fpar_.xllc - margin;
	f_label();
	t_label();
	e_label();
	w_label();
	t_annot(0.0);
	e_axis(0);
	f_axis(0);
	w_axis(0);
	if (bpar_.taxis) {
	    rx1 = ucon_.strtpos - ndx / 2;
	    ry1 = ucon_.ysp1 - tick_.t_xs1;
	    rxw = ndx;
	    ryh = tick_.t_xs1;
	    lpgrule(rx1, ry1, rxw, ryh);
	    ry1 = ucon_.ysp1 + ucon_.ysp2;
	    lpgrule(rx1, ry1, rxw, ryh);
	    rx1 = ucon_.strtpos - ndx / 2;
	    ry1 = ucon_.ysp1 - ndy / 2;
	    rxw = ndx + ucon_.framelength;
	    ryh = ndy;
	    lpgrule(rx1, ry1, rxw, ryh);
	    ry1 += ucon_.ysp2;
	    lpgrule(rx1, ry1, rxw, ryh);
	}
	lpgadvto(ucon_.strtpos);
    }
}

void
e_axis(int flg)
{
    int     i, iet;
    float   e, de1, ebot, adupdb;
    long    rx0, rx1, ry1, rxw, ryh;

    if (!bpar_.eaxis)
	return;
    ebot = fpar_.emin - fpar_.emax;
    adupdb = ucon_.yen2 / ebot;
    rx0 = ucon_.strtpos;
    rx1 = rx0 - ndx / 2;
    ry1 = ucon_.yen1 - ndy / 2;
    rxw = ndx;
    ryh = ucon_.yen2 + ndy;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 += ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
    if (inf_type == FT_TSD) {
	rx1 = rx0 - tick_.t_ys1;
	rxw = tick_.t_ys1;
	ry1 = ucon_.yen1 - ndy / 2;
	ryh = ndy;
	lpgrule(rx1, ry1, rxw, ryh);
	rx1 = ucon_.strtpos - ndx / 2 + ucon_.framelength;
	lpgrule(rx1, ry1, rxw, ryh);
	rx1 = rx0 - tick_.t_ys1;
	ry1 = ucon_.yen1 + ucon_.yen2 - ndy / 2;
	lpgrule(rx1, ry1, rxw, ryh);
	rx1 = ucon_.strtpos - ndx / 2 + ucon_.framelength;
	lpgrule(rx1, ry1, rxw, ryh);
	rx1 = rx0 - tick_.t_ys1;
	ry1 = ucon_.yen0 - ndy / 2;
	lpgrule(rx1, ry1, rxw, ryh);
	rx1 = ucon_.strtpos - ndx / 2 + ucon_.framelength;
	lpgrule(rx1, ry1, rxw, ryh);
    } else {
	de1 = -10;
	for (i = 0; i < MXNT; i++) {
	    if ((e = i * de1) < ebot)
		break;
	    iet = (int) (ucon_.yen2 - e * adupdb + 0.5);
	    if (iet >= 0) {
		if ((i % 2) == 0) {
		    rx1 = rx0 - tick_.t_ys1;
		    rxw = tick_.t_ys1;
		} else {
		    rx1 = rx0 - tick_.t_ys2;
		    rxw = tick_.t_ys2;
		}
		ry1 = ucon_.yen1 - ndy / 2 + iet;
		ryh = ndy;
		lpgrule(rx1, ry1, rxw, ryh);
		rx1 = ucon_.strtpos - ndx / 2 + ucon_.framelength;
		lpgrule(rx1, ry1, rxw, ryh);
	    }
	}
    }
}

void
f_axis(int flg)
{
    int     ift, ii, ts, ftpi;
    float   f, ft0, df1, ftop, aduphz;
    long    rx0, rx1, ry1, rxw, ryh;
    static float df0 = 100;

    if (!bpar_.faxis)
	return;

    ftop = ucon_.famx - ucon_.famn;
    if (ftop == 0)
	return;
    if (ftop < 0)
	ftop = -ftop;
    aduphz = ucon_.ysp2 / ftop;
    ftpi = (int) (ftop / (df0 * fpar_.ylen));
    if (ftpi < 10)
	ts = 1;
    else if (ftpi < 20)
	ts = 2;
    else if (ftpi < 50)
	ts = 5;
    else if (ftpi < 100)
	ts = 10;
    else if (ftpi < 200)
	ts = 20;
    else
	ts = 50;
    df1 = df0 * ts;
    ft0 = (float) ceil(ucon_.famn / df1) * df1 - ucon_.famn;
    ii = (int) (ceil(ucon_.famn / df1) * ts);
    rx0 = ucon_.strtpos;
    rx1 = rx0 - ndx / 2;
    ry1 = ucon_.ysp1 - ndy / 2;
    rxw = ndx;
    ryh = ucon_.ysp2 + ndy;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 += ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
    for (f = ft0; f <= ftop; f += df1, ii += ts) {
	ift = (int) (f * aduphz + 0.5);
	if (ift <= ucon_.ysp2) {
	    if ((ii % 10) == 0) {
		rxw = tick_.t_ys1;
	    } else if ((ii % 5) == 0) {
		rxw = tick_.t_ys2;
	    } else {
		rxw = tick_.t_ys3;
	    }
	    rx1 = rx0 - rxw;
	    ry1 = ucon_.ysp1 - ndy / 2 + ift;
	    ryh = ndy;
	    lpgrule(rx1, ry1, rxw, ryh);
	    rx1 = ucon_.strtpos - ndx / 2 + ucon_.framelength;
	    lpgrule(rx1, ry1, rxw, ryh);
	}
    }
}

void
w_axis(int flg)
{
    int     iwbl;
    long    rx0, rx1, ry1, rxw, ryh;

    if (!bpar_.waxis)
	return;
    rx0 = ucon_.strtpos;
    rx1 = rx0 - ndx / 2;
    ry1 = ucon_.ywv1 - ndy / 2;
    rxw = ndx;
    ryh = ucon_.ywv2 + ndy;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 += ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
    rxw = tick_.t_ys3;
    ryh = ndy;
    rx1 = rx0 - tick_.t_ys3;
    ry1 = ucon_.ywv1 - ndy / 2;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 = rx0 + ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 = rx0 - tick_.t_ys3;
    ry1 = ucon_.ywv1 - ndy / 2 + ucon_.ywv2;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 = rx0 + ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
    iwbl = (int) (ucon_.ywv2 * (wave_.wbl - wave_.wmn) / wave_.wrn);
    rx1 = rx0 - tick_.t_ys3;
    ry1 = ucon_.ywv1 - ndy / 2 + iwbl;
    lpgrule(rx1, ry1, rxw, ryh);
    rx1 = rx0 + ucon_.framelength;
    lpgrule(rx1, ry1, rxw, ryh);
}

void
t_axis(long et)
{
    int     itic;
    double  t, t1, t2;
    long    rx1, ry1, rxw, ryh;

    if (et == tick_.t_et) {
	if (bpar_.tannot) {
	    t = last_time_ann + t_dt;
	    t1 = et * ucon_.dtprn + t_dt + ucon_.dtprn;
	    t2 = ucon_.tmx - ucon_.tmn + ucon_.dtprn;
	    if (t <= t1 && t <= t2)
		t_annot(t);
	}
	tick_.t_type += t_tts;
	tick_.t_type %= 10;
	tick_.t_time += t_tts * t_tint;
	tick_.t_et = (long) (tick_.t_time / ucon_.dtprn + 0.5);
	if (tick_.t_type == 0)
	    itic = tick_.t_xs1;
	else if (tick_.t_type == 5)
	    itic = tick_.t_xs2;
	else
	    itic = tick_.t_xs3;
	t = ndx * ucon_.dtprn / 2;
	if (tick_.t_time <= ucon_.tmx - ucon_.tmn + t) {
	    rx1 = ucon_.strtpos - ndx / 2 + tick_.t_et;
	    ry1 = ucon_.ysp1 - itic;
	    rxw = ndx;
	    ryh = itic;
	    lpgrule(rx1, ry1, rxw, ryh);
	    ry1 = ucon_.ysp1 + ucon_.ysp2;
	    lpgrule(rx1, ry1, rxw, ryh);
	}
    }
}

void
f_annot()
{
    char    s[20];
    int     ift, ifmt, jfmt;
    float   xx, yy, dy, w, out;
    float   f, ft0, df1, ftop, aduphz;

    if (!bpar_.fannot)
	return;
    ftop = ucon_.famx - ucon_.famn;
    if (ftop == 0)
	return;
    if (ftop < 0)
	ftop = -ftop;
    aduphz = ucon_.ysp2 / ftop;
    dy = (float) (1.0 / ucon_.adupi_y);
    df1 = fpar_.fint;
    ft0 = (float) ceil(ucon_.famn / df1) * df1;	/*- ucon_.famn*/
    xx = (float) (fpar_.xllc - (max_tic_wid + fpar_.annsiz * 0.5));
    if (xx < fpar_.annsiz)
	return;
    max_ann_wid = 0;
    for (f = ft0; f <= ucon_.famx; f += df1) {
	ift = (int) ((f - ucon_.famn) * aduphz + 0.5);
	if (ift <= ucon_.ysp2) {
	    yy = (ucon_.ysp1 + ift) * dy;
	    ifmt = ucon_.ffmt + 1;
	    if (ucon_.faty == 0) {
		out = f;
                if (df1 >= 1000) {
    		    jfmt = 0;
    		} else if (df1 >= 100) {
    		    jfmt = 1;
    		    ifmt += 2;
    		}
	    } else if (ucon_.faty == 1) {
		out = (float) (ucon_.fmn * pow2(f / 1000));
		jfmt = ifmt + 2 - (int) log10(out);
	    } else {
		out = (float) (ucon_.fpma * (pow10(f * ucon_.fpmb) - ucon_.fpmc));
		jfmt = ifmt + 2 - (int) log10(out);
	    }
	    jfmt = limit(0, jfmt, ifmt - 1);
	    sprintf(s, "%*.*f", ifmt, jfmt, out / 1000);
	    lpgchup(0.0, 1.0);
	    lpgchlw(fpar_.annlwt);
	    lpgchsz(fpar_.annsiz);
	    lpgchal(3, 3);
	    lpgchxt(0);
	    lpgtext(xx, yy, s);
	    w = (float) txtlen(s);
	    if (max_ann_wid < w)
		max_ann_wid = w;
	}
    }
}

void
t_annot(t)
double  t;
{
    char    s[20];
    double  xx, yy, dx;

    if (!bpar_.tannot)
	return;
    dx = fpar_.xlen / (ucon_.tmx - ucon_.tmn);
    xx = fpar_.xllc + t * dx;
    yy = t_ann_pos;
    if (xx < fpar_.annsiz || yy < fpar_.annsiz)
	return;
    sprintf(s, t_ndp ? "%.1f" : "%.0f", t * usf[t_tui]);
    lpgchup(0.0, 1.0);
    lpgchlw(fpar_.annlwt);
    lpgchsz(fpar_.annsiz);
    lpgchal(2, 1);
    lpgchxt(0);
    lpgtext(xx, yy, s);
    last_time_ann = t;
}

void
f_label()
{
    double  xx, yy;

    if (!bpar_.flabel)
	return;
    xx = lab_pos;
    yy = fpar_.yllc + fpar_.ylen / 2;
    if (xx < fpar_.labsiz || yy < fpar_.labsiz)
	return;
    lpgchup(-1.0, 0.0);
    lpgchlw(fpar_.lablwt);
    lpgchsz(fpar_.labsiz);
    lpgchal(2, 5);
    lpgchxt(0);
    if (ucon_.faty == 2)
	lpgtext(xx, yy, "place, kHz");
    else
	lpgtext(xx, yy, "frequency, kHz");
}

void
t_label()
{
    char    s[20];
    double  xx, yy;

    if (!bpar_.tlabel)
	return;
    xx = fpar_.xllc + fpar_.xlen / 2;
    yy = t_ann_pos - (bpar_.tannot ? fpar_.annsiz * 2.0 : 0.0);
    if (xx < fpar_.labsiz || yy < fpar_.labsiz)
	return;
    lpgchup(0.0, 1.0);
    lpgchlw(fpar_.lablwt);
    lpgchsz(fpar_.labsiz);
    lpgchal(2, 1);
    lpgchxt(0);
    sprintf(s, "time, %s", tu[t_tui]);
    lpgtext(xx, yy, s);
}

void
e_label()
{
    double  xx, yy;

    if (!bpar_.elabel)
	return;
    xx = lab_pos;
    yy = fpar_.ellc + fpar_.elen / 2;
    if (xx < fpar_.labsiz || yy < fpar_.labsiz)
	return;
    lpgchup(-1.0, 0.0);
    lpgchlw(fpar_.lablwt);
    lpgchsz(fpar_.labsiz);
    lpgchal(2, 5);
    lpgchxt(0);
    if (inf_type == FT_TSD)
	lpgtext(xx, yy, "wnr");
    else
	lpgtext(xx, yy, "energy");
}

void
w_label()
{
    double  xx, yy;

    if (!bpar_.wlabel)
	return;
    xx = lab_pos;
    yy = fpar_.wllc + fpar_.wlen / 2;
    if (xx < fpar_.labsiz || yy < fpar_.labsiz)
	return;
    lpgchup(-1.0, 0.0);
    lpgchlw(fpar_.lablwt);
    lpgchsz(fpar_.labsiz);
    lpgchal(2, 5);
    lpgchxt(0);
    lpgtext(xx, yy, "waveform");
}

void
dogrid(long et)
{
    int     ift;
    float   f, ft0, df1, ftop, aduphz;
    long    rx1, ry1, rxw, ryh;

    if (et != grid_.g_et)
	return;

    ftop = ucon_.famx - ucon_.famn;
    if (ftop == 0)
	return;
    if (ftop < 0)
	ftop = -ftop;
    aduphz = ucon_.ysp2 / ftop;
    df1 = ucon_.fgr;
    ft0 = (float) ceil(ucon_.famn / df1) * df1 - ucon_.famn;
    rxw = 1 + (ndx - 1) / 2;
    ryh = 1 + (ndy - 1) / 2;
    for (f = ft0; f <= ftop; f += df1) {
	ift = (int) (f * aduphz + 0.5);
	if (ift <= ucon_.ysp2) {
	    rx1 = ucon_.strtpos - rxw / 2 + et;
	    ry1 = ucon_.ysp1 - ryh / 2 + ift;
	    lpgrule(rx1, ry1, rxw, ryh);
	}
    }
    grid_.g_time += grid_.g_incr;
    grid_.g_et = (long) (grid_.g_time / ucon_.dtprn + 0.5);
}
