/* emf.c */

#include <stdlib.h>
#include <stdio.h>
#include "emf.h"

#define XPAGE	    11
#define YPAGE	    8
#define XTRN(x)     (x)
#define YTRN(y)     (pgwid-(y))

void    lpr();
void    lprfl();
void    lprs();
void    lprrw();

extern double dotdiam;
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int p_color;
extern int bpp;
extern int p_gmode;
extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) (), (*pr_gray) ();

typedef struct {int l, t, r, b;} BB;	// bounding box

static double unmap(double, float *, int);
static void emf_idle(), emf_page(), emf_out(), emf_end();
static void emf_rule(), emf_line(), emf_gray();
static void emr_write(int, int), set_pen(int, int), set_brush(int);
static void line(int, int, int, int, int);
static void rectangle(int, int, int, int, int);
static BB bbox();

static int adupi = 0;
static int rcnt = 0;
static int rsiz = 0;
static int pnwd = 0;
static int havpen = 0;
static int havbrs = 0;
static int havbnd = 0;
static int lastpc = 0;
static int lastpw = 0;
static int lastbc = 0;
static short  desc[18] = {
    'B', 'T', 'N', 'R', 'H', ' ', 'S', 'P', 'E', 'C', 'T', 'O',
};
static BB	    b = {0};
static EMFRECORD    r;
static EMFHEADER    h;
static RECT 	    Scr;

void
emf_setup()
{
    if (p_gmode == 1) {
	adupi = 100;
    } else if (p_gmode == 2) {
	adupi = 200;
    } else if (p_gmode == 3) {
	adupi = 300;
    } else {
	adupi = 100;
    }
    adupi_x = adupi;
    adupi_y = adupi;
    pglen = adupi * XPAGE;
    pgwid = adupi * YPAGE;
    bpp = 8;
    dotdiam = (float) (1.0 / adupi);

    h.RecordType = EMR_HEADER;
    h.Size = sizeof(h) + sizeof(desc);
    h.BoundsLeft = 0;			    // update this on close
    h.BoundsTop = 0;			    // update this on close
    h.BoundsRight = pglen;		    // update this on close
    h.BoundsBottom = pgwid;		    // update this on close
    h.FrameLeft = 0;			    // 0.01 mm units
    h.FrameTop = 0;			    // 0.01 mm units
    h.FrameRight = (int) (XPAGE * 2540);    // 0.01 mm units
    h.FrameBottom = (int) (YPAGE * 2540);   // 0.01 mm units
    h.Signature = 0x464D4520;
    h.Version[0] = 0;
    h.Version[1] = 1;
    h.FileSize = h.Size;		    // update this on close
    h.NumOfRecords = 1;			    // update this on close
    h.NumOfHandles = 8;
    h.SizeOfDescrip = sizeof(desc) / 2;
    h.OffsOfDescrip = sizeof(h);
    h.NumPalEntries = 0;
    h.WidthDevPixels = (int) (XPAGE * adupi_x);
    h.HeightDevPixels = (int) (YPAGE * adupi_y);
    h.WidthDevMM = (int) (XPAGE * 25.4);
    h.HeightDevMM = (int) (YPAGE * 25.4);
    lpr(&h, sizeof(h));
    lpr(desc, sizeof(desc));
    rsiz = h.Size;
    rcnt = 1;

    havpen = havbrs = havbnd = 0;
    lastpc = lastbc = lastpw = -1;

    pr_idle = emf_idle;
    pr_page = emf_page;
    pr_out = emf_out;
    pr_end = emf_end;
    pr_rule = emf_rule;
    pr_line = emf_line;
    pr_gray = emf_gray;
}

void
emf_out(buf, n)
unsigned char *buf;
int     n;
{
}

void
emf_end(int prompt)
{

    // write EOF record
    r.P[0] = 0;		// Number of color palette entries
    r.P[1] = 4 * 4;	// Offset of color palette entries
    r.P[2] = 5 * 4;	// Offset to beginning of this record
    emr_write(EMR_EOF, 3);
    lprrw();
    h.FileSize = rsiz;
    h.NumOfRecords = rcnt;
    lpr(&h, sizeof(h));
    lprfl();
}

void
emf_idle()
{
}

void
emf_page()
{
}

void
emf_rule(long xllc, long yllc, long xlen, long ylen)
{
    if (xlen > ylen)
	line(xllc, yllc + ylen / 2, xllc + xlen, yllc + ylen / 2, ylen);
    else
	line(xllc + xlen / 2, yllc, xllc + xlen / 2, yllc + ylen, xlen);
}

void
emf_line(int n, long *x, long *y, double w)
{
    int i, f;
    BB a;

    pnwd = 1 + (int) (w * adupi - 0.5);
    set_pen(pnwd, RGB(0, 0, 0));
    emr_write(EMR_BEGINPATH, 0);
    for (i = 0; i < n; i++) {
	f = i ? EMR_LINETO : EMR_MOVETOEX;
	r.P[0] = XTRN(x[i]);
	r.P[1] = YTRN(y[i]);
        emr_write(f, 2);
    }
    emr_write(EMR_ENDPATH, 0);
    a = bbox(x, y, n);
    r.P[0] = a.l;
    r.P[1] = a.t;
    r.P[2] = a.r;
    r.P[3] = a.b;
    emr_write(EMR_STROKEPATH, 4);
}

void
emf_gray(long x, long y, long w, long h, long nx, unsigned char *px, float *mp)
{
    double a0, a1, dj, p1, p2, pm, a, b;
    int i, j, n, g, rd, gr, bl, c, y1, y2;
    static int lastn = 0;
    static double rpi = 100;	// rectangles per inch

    n = (int) ((h / adupi_y) * rpi + 0.5);
    if (n < nx)
	n = nx;
    if (n > h / 2)
	n = h;
    a = w;
    b = (double) h / n;
    for (i = 0; i < n; i++) {
	// interpolate i
	dj = unmap((double) (i) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	p1 = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 0.5
	dj = unmap((double) (i + 1) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	pm = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 1
	dj = unmap((double) (i + 0.5) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	p2 = a0 * px[j] + a1 * px[j + 1];
	g = (int) ((p1 + p2 + pm) / 3 + 0.5);
	if (p_color) {
	    if (g < 128) {
	        rd = 0;
		bl = 255 - g * 2;
	    } else {
	        rd = (g - 128) * 2;
		bl = 0;
	    }
	    gr = 255 - rd - bl;
	    c = RGB(rd, gr, bl);
	} else {
	    g = 255 - g;
	    c = RGB(g, g, g);
	}
	y1 = y + ((i + 0) * (h - 1)) / n;
	y2 = y + ((i + 1) * (h - 1)) / n;
	rectangle(x, y1, x + w, y2, c);
    }
}

/******************************************************************/

static void
emr_write(int f, int n)
{
    r.Function = f;
    r.Size = 4 * (2 + n);
    lpr(&r, r.Size);
    rsiz += r.Size;
    rcnt += 1;
}

static void
set_pen(int w, int c)
{
    if (c != lastpc  || w != lastpw) {
	if (havpen) {
	    r.P[0] = 1;		// handle
	    emr_write(EMR_DELETEOBJECT, 1);
	    havpen = 0;
        }
        if (c >= 0) {
	    r.P[0] = 1;		// handle
	    r.P[1] = 56;	// ???
	    r.P[2] = 0;		// ???
	    r.P[3] = 56;	// ???
	    r.P[4] = 0;		// ???
	    r.P[5] = 1 << 16;   // ???
	    r.P[6] = w;		// width
	    r.P[7] = 0;		// ???
	    r.P[8] = c;		// color
	    r.P[9] = 0;		// ???
	    r.P[10] = 0;	// ???
	    r.P[11] = 0;	// ???
	    emr_write(EMR_EXTCREATEPEN, 12);
	    r.P[0] = 1;		// handle
	    emr_write(EMR_SELECTOBJECT, 1);
	    havpen = 1;
	    lastpc = c;
	    lastpw = w;
	}
    }
}

static void
set_brush(int c)
{
    if (c != lastbc) {
	if (havbrs) {
	    r.P[0] = 2;		// handle
	    emr_write(EMR_DELETEOBJECT, 1);
	    havbrs = 0;
	}
	if (c >= 0) {
	    r.P[0] = 2;		// handle
	    r.P[1] = 0;		// solid
	    r.P[2] = c;		// color
	    r.P[3] = 0;		// ???
	    emr_write(EMR_CREATEBRUSH, 4);
	    r.P[0] = 2;		// handle
	    emr_write(EMR_SELECTOBJECT, 1);
	    havbrs = 1;
	    lastbc = c;
	}
    }
}

void
line(int x1, int y1, int x2, int y2, int w)
{
    set_pen(w + 1, RGB(0, 0, 0));
    r.P[0] = XTRN(x1);
    r.P[1] = YTRN(y1);
    emr_write(EMR_MOVETOEX, 2);
    r.P[0] = XTRN(x2);
    r.P[1] = YTRN(y2);
    emr_write(EMR_LINETO, 2);
}

static void
rectangle(int x1, int y1, int x2, int y2, int c)
{
    set_pen(1, c);
    set_brush(c);
    r.P[0] = XTRN(x1);
    r.P[1] = YTRN(y1);
    r.P[2] = XTRN(x2);
    r.P[3] = YTRN(y2);
    emr_write(EMR_RECTANGLE, 4);
}

static BB
bbox(int *xp, int *yp, int np)
{
    int i, x, y, p;
    BB a;

    p = pnwd;		// margin equals pen width
    x = XTRN(xp[0]);
    y = YTRN(yp[0]);
    a.l = x - p;
    a.t = y - p;
    a.r = x + p;
    a.b = y + p;
    for (i = 1; i < np; i++) {
	x = XTRN(xp[i]);
	y = YTRN(yp[i]);
	if (a.l > x - p)
	    a.l = x - p;
	if (a.t > y - p)
	    a.t = y - p;
	if (a.r < x + p)
	    a.r = x + p;
	if (a.b < y + p)
	    a.b = y + p;
    }
    if (!havbnd) {
	b.l = a.l;
	b.t = a.t;
	b.r = a.r;
	b.b = a.b;
	havbnd = 1;
    } else {
	if (b.l > a.l)
	    b.l = a.l;
	if (b.t > a.t)
	    b.t = a.t;
	if (b.r < a.r)
	    b.r = a.r;
	if (b.b < a.b)
	    b.b = a.b;
    }

    return (a);
}

static double
unmap(double u, float *mp, int nx)
{
    int i;

    for (i = 1; i <= nx; i++) {
	if (u < mp[i]) {
	    u = (i + (u - mp[i - 1]) / (mp[i] - mp[i - 1])) / nx;
	    break;
	}
    }
    return (u * (nx - 0.5));
}
