/* fx.c */

#include <stdlib.h>
#include <stdio.h>

#define ESC     27

void    fx_idle(), fx_page(), fx_out(), fx_end();
void    lpr();
void    lprfl();
void    lprs();

extern double dotdiam;
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int bpp;
extern int p_gmode;
extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) ();

static void fx_bitset();
static void fx_lin();

static unsigned char *fxbuf;
static int ov = 0;
static int ovlp = 0;
static int gmode = 0;
static int xdpi = 0;
static int ydpi = 0;
static int firstprint = 1;
static unsigned char m0 = 0x80;
static unsigned char m1 = 0x40;
static unsigned char m2 = 0x20;
static unsigned char m3 = 0x10;
static unsigned char m4 = 0x08;
static unsigned char m5 = 0x04;
static unsigned char m6 = 0x02;
static unsigned char m7 = 0x01;
static int nl = 0;

void
fx_setup()
{
    if (p_gmode == 3) {
	gmode = 3;
	ovlp = 3;
	xdpi = 216;
	ydpi = 240;
	nl = xdpi * 9;		/* allow for 9 inches across */
	fxbuf = (unsigned char *) calloc(nl * 3, 1);
    } else {
	gmode = 1;
	ovlp = 1;
	xdpi = 72;
	ydpi = 120;
    }
    adupi_x = xdpi;
    adupi_y = ydpi;
    pglen = adupi_x * 11;
    pgwid = adupi_y * 8;
    bpp = 1;
    dotdiam = (float) (1.0 / 72.0);

    pr_idle = fx_idle;
    pr_page = fx_page;
    pr_out = fx_out;
    pr_end = fx_end;
    pr_rule = NULL;
    pr_line = NULL;
}

void
fx_out(buf, n)
unsigned char *buf;
int     n;
{
    if (firstprint) {
	lprs("\r\033@\r");	/* reset printer */
	firstprint = 0;
    }
    if (ovlp == 1) {
	fx_lin(buf, n, 24);
    } else {
	fx_bitset(buf, n);
    }
}

void
fx_end(prompt)
{
    prompt = prompt;
    if (ovlp > 1 && ov > 0) {
	fx_lin(fxbuf, nl, 1);
	fx_lin(fxbuf + nl, nl, 1);
	fx_lin(fxbuf + 2 * nl, nl, 22);
	ov = 0;
    }
    lprs("\r\033@\r");		/* reset printer */
    lprfl();
}

void
fx_idle()
{
    lprs("\r");			/* send <CR> to printer to keep it on line */
}

void
fx_page()
{
}

static void
fx_bitset(buf, n)
unsigned char *buf;
int     n;
{
    int     b, i;

    if (ov == 0) {
	for (i = 0; i < n; i++) {
	    b = buf[i];
	    fxbuf[i] = (b & m0) | ((b & m3) << 2) | ((b & m6) << 4);
	    fxbuf[i + nl] = ((b & m1) << 1) | ((b & m4) << 3) |
		((b & m7) << 5);
	    fxbuf[i + 2 * nl] = ((b & m2) << 2) | ((b & m5) << 4);
	}
	ov++;
    } else if (ov == 1) {
	for (i = 0; i < n; i++) {
	    b = buf[i];
	    fxbuf[i + 2 * nl] |= ((b & m0) >> 2) | (b & m3) |
		((b & m6) << 2);
	    fxbuf[i] |= ((b & m1) >> 2) | (b & m4) |
		((b & m7) << 2);
	    fxbuf[i + nl] |= ((b & m2) >> 1) | ((b & m5) << 1);
	}
	ov++;
    } else {
	for (i = 0; i < n; i++) {
	    b = buf[i];
	    fxbuf[i + nl] |= ((b & m0) >> 5) | ((b & m3) >> 3) |
		((b & m6) >> 1);
	    fxbuf[i + 2 * nl] |= ((b & m1) >> 4) | ((b & m4) >> 2) |
		(b & m7);
	    fxbuf[i] |= ((b & m2) >> 4) | ((b & m5) >> 2);
	}
	fx_lin(fxbuf, nl, 1);
	fx_lin(fxbuf + nl, nl, 1);
	fx_lin(fxbuf + 2 * nl, nl, 22);
	ov = 0;
	for (i = 0; i < nl * 3; i++)
	    fxbuf[i] = 0;
    }
}

/* fx_lin - print a line of graphics commands with overprint if necessary */

static void
fx_lin(pb, n, lf)
unsigned char *pb;
int     n, lf;
{
    unsigned char *ovrbuf, s[20];
    int     c, j, ovrflg = 0, lc = 0;
    int     ac, mask;

    for (; n > 0 && pb[n - 1] == 0; n--)	/* trim */
	continue;
    if (n > 0) {
	s[0] = '\r';
	s[1] = ESC;
	s[2] = '*';
	s[3] = gmode;
	s[4] = n % 256;
	s[5] = n / 256;
	lpr((char *) s, 6);
	ovrbuf = calloc((unsigned) (n + 1), 1);
	for (j = 0; j < n; j++) {
	    c = pb[j] & 255;
	    ac = (lc & c);
	    if (ovlp > 1 && ac != 0) {
		ovrbuf[j] = ac;
		mask = ~((unsigned) ac);
		c &= mask;
		ovrflg++;
	    }
	    lc = c;
	    pb[j] = c;
	}
	lpr((char *) pb, (unsigned int) n);
	for (j = 0; j < n; j++) {
	    pb[j] = 0;
	}
	if (ovrflg > 0) {
	    fx_lin(ovrbuf, n, 0);
	}
	free(ovrbuf);
    }
    if (lf) {
	s[0] = ESC;
	s[1] = '3';
	s[2] = lf;
	s[3] = '\n';
	lpr((char *) s, 4);
    }
}
