/* grx.c - console routines for GRX version of SPECTO */

#include <stdio.h>
#include <grx20.h>

void    co_setup(), co_idle(), co_page(), co_out(), co_end();
void    co_rule(), co_line();

extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) ();
extern FILE *lgf;
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int bpp;
extern int p_gmode;
extern int p_color;
extern int noquery;
extern double dotdiam;
extern double a_coef, b_coef, c_coef;

static int ncolors = 256;
static int out_yet = 0;
static int xbyte = 0;
static int xpix = 640;
static int ypix = 480;
static int xmax = 0;
static int ymax = 0;
static int bgc = 0;
static int fgc = 0;
static GrContext *gcc;

static void
set_palette()
{
    int     i, m, n, r, g, b;
    static double l = 0.1, s = 1.0, hmx = 300;

    n = 1 << bpp;
    m = n - 1;
    GrSetColor(0, 255, 255, 255);
    GrSetColor(m, 0, 0, 0);
    for (i = 1; i < m; i++) {
	if (p_color == 0) {
	    r = g = b = (255 * (m  - 1 - i)) / (m - 2);
	} else {
	    if (i < m / 2) {
		r = 0;
		g = (i * 255 * 2) / m;
		b = 255 - g;
	    } else {
		b = 0;
		g = ((m - i - 1) * 255 * 2) / m;
		r = 255 - g;
	    }
	}
	GrSetColor(i, r, g, b);
    }
}

static void
grx_init()
{
    double p;
    int i, g, nin;

    GrSetMode(GR_width_height_color_graphics, xpix, ypix, ncolors);
    xbyte = 0;
    gcc = GrCreateContext(8, ypix, NULL, NULL);
    xmax = xpix - 1;
    ymax = ypix - 1;
    set_palette();
}

void
co_setup()
{
    bpp = 4;

    adupi_x = 60;
    adupi_y = 60;
    pglen = (int) (adupi_x * 10.5);
    pgwid = (int) (adupi_y * 8.0);
    dotdiam = 1.0 / adupi_x;

    pr_idle = co_idle;
    pr_page = co_page;
    pr_out = co_out;
    pr_end = co_end;
    pr_rule = co_rule;
    pr_line = co_line;
    pr_rule = NULL;
    pr_line = NULL;
}

void
co_idle()
{
}

void
co_page()
{
    if (xbyte > 0) {
	getch();
	xbyte = 0;
    }
}

void
co_out(unsigned char *buf, int n)
{
    int i, j, k, x, y, c, x0, m;

    if (!out_yet) {
	if (!noquery) {
	    fprintf(lgf, "\n    [press any key to continue]\n");
	    (void) getch();
	}
	lgf = fopen("specto.out", "w");
	grx_init();
	out_yet = 1;
    }
    x0 = xbyte * 8;
    for (i = 0; i < n / bpp; i++) {
        y = ypix - i;
        for (j = 0; j < 8; j++) {
	    x = x0 + j;
	    if (bpp == 8) {
		c = buf[i * bpp + j];
	    } else {
		c = 0;
		for (k = 0; k < bpp; k++)
		    c |= ((buf[i * bpp + k] >> (7 - j)) & 1) << (bpp - 1 - k);
	    }
	    GrPlotNC(x, y, c);
	}
    }
    GrFilledBox(x0, 0, x0 + 7, ymax - n, bgc);
    xbyte++;
}

void
co_end(int prompt)
{
    if (prompt && xbyte > 0)
	(void) getch();
    if (out_yet) {
	GrSetMode(GR_default_text);
	out_yet = 0;
    }
}

void
co_rule(long xllc, long yllc, long xlen, long ylen)
{
    int x1, y1, x2, y2;

    if (!out_yet) {
	grx_init();
	out_yet = 1;
    }
    x1 = xllc;
    y1 = ymax - yllc;
    x2 = xllc + (xlen - 1);
    y2 = ymax - yllc - (ylen - 1);
    GrFilledBox(x1, y1, x2, y2, fgc);
}

void
co_line(int n, long *x, long *y, double w)
{
    int  i;
    int x1, y1, x2, y2;

    if (!out_yet) {
	grx_init();
	out_yet = 1;
    }
    for (i = 1; i < n; i++) {
	x1 = x[i - 1];
	y1 = ypix - y[i - 1];
	x2 = x[i];
	y2 = ypix - y[i];
        GrLine(x1, y1, x2, y2, fgc);
    }
}
