/* ip.c */

#include <stdlib.h>
#include <stdio.h>

#define ESC         27
#define	BLACK       15
#define	BITMAP      235
#define	ENDPAGE     219
#define E_O_F       255
#define SET_ABS_H   135
#define SET_REL_H   136
#define SET_ABS_V   137
#define SET_REL_V   138

void    ip_idle(), ip_page(), ip_out(), ip_end();
void    lpr();
void    lprfl();
void    lprs();

extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) ();
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int bpp;
extern int p_gmode;
extern float dotdiam;

static void ip_lin();
static void putword();
static void putstring();
static void putbyte();
static int adjpat();
static void outpat();

static char *ipbuf;
static int ipdpi = 0;
static int ov = 0;
static int npat = 80;
static int ph = 32;
static int pw = 4;
static int xpos = 0;
static int firstprint = 1;
static unsigned int nl = 0;
static unsigned char m[8] =
{0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

void
ip_setup()
{
    ipdpi = 300;
    npat = 80;
    nl = npat * pw;
    ipbuf = (char *) calloc(nl * ph, 1);
    adupi_x = ipdpi;
    adupi_y = ipdpi;
    pglen = (int) (adupi_x * 10.5);
    pgwid = (int) (adupi_y * 8.0);
    bpp = 1;
    dotdiam = (float) (1.0 / ipdpi);

    pr_idle = ip_idle;
    pr_page = ip_page;
    pr_out = ip_out;
    pr_end = ip_end;
    pr_rule = NULL;
    pr_line = NULL;
}

void
ip_out(buf, n)
unsigned char *buf;
int     n;
{
    char   *pp;
    int     i, j;

    if (firstprint) {
	ov = 0;
	xpos = 0;
	firstprint = 0;
	putstring("@document(language impress)");
    }
    for (j = 0; j < 8; j++) {
	pp = ipbuf + (j + ov * 8) * nl;
	for (i = 0; i < n; i++) {
	    if (buf[i] & m[j])
		pp[i / 8] |= m[i % 8];
	}
    }
    ov++;
    if (ov == 4) {
	ip_lin(ipbuf, npat, 1);
	ov = 0;
    }
    xpos++;
}

void
ip_end(int prompt)
{
    prompt = prompt;
    if (xpos > 0) {
	if (ov > 0)
	    ip_lin(ipbuf, npat, 1);
	putbyte((char) ENDPAGE);
    }
    putbyte((char) E_O_F);
    lprfl();
}

void
ip_idle()
{
}

void
ip_page()
{
    if (xpos > 0)
	putbyte((char) ENDPAGE);
    ov = 0;
    xpos = 0;
}

/* ip_lin - print a row of bitmap 32x32 patches */

static void
ip_lin(pb, n, lf)
char   *pb;
int     n, lf;
{
    char   *pp;
    int     i, np;

    for (i = 0; i < npat; i++) {
	pp = pb + i * pw;
	if ((np = adjpat(pp, npat - i)) > 0) {
	    putbyte((char) SET_ABS_H);
	    putword(i * pw * 8);
	    outpat(pp, np);
	    i += np;
	}
    }
    putbyte((char) SET_REL_V);
    putword(ph * lf);
}

/*
 * putword - put a word to stdout in "big endian" mode
 */
static void
putword(w)
short   w;
{
    putbyte((char) (((unsigned short) w >> 8) & 0xff));
    putbyte((char) ((unsigned short) w & 0xff));
}

static void
putstring(s)
char   *s;
{
    while (*s)
	putbyte(*s++);
}

static void
putbyte(int c)
{
    char    s[1];

    s[0] = c;
    lpr(s, 1);
}

/* adjpat - determine the number of nonzero adjacent patches */

static int
adjpat(char *pp, int n)
{
    int     i, j, k, nz;

    for (i = 0; i < n; i++) {
	nz = 0;
	for (j = 0; j < ph && !nz; j++)
	    for (k = 0; k < pw && !nz; k++)
		if (pp[k + j * nl] != 0)
		    nz++;
	if (nz == 0)
	    break;
	pp += pw;
    }
    return (i);
}


static void
outpat(char *pp, int n)
{
    int     i, j, k;

    putbyte((char) BITMAP);
    putbyte((char) BLACK);	/* operation type */
    putbyte((char) n);		/* hsize */
    putbyte((char) 1);		/* vsize */
    for (i = 0; i < n; i++) {
	for (j = 0; j < ph; j++) {
	    for (k = 0; k < pw; k++) {
		putbyte(pp[k + j * nl]);
		pp[k + j * nl] = 0;
	    }
	}
	pp += pw;
    }
}
