/* lj.c */

#include <stdlib.h>
#include <stdio.h>

void    lj_idle(), lj_page(), lj_out(), lj_end();
void    lpr();
void    lprfl();
void    lprs();

extern double dotdiam;
extern double a_coef, b_coef, c_coef;
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int bpp;
extern int p_gmode;
extern void (*pr_idle)(), (*pr_page)(), (*pr_out)(), (*pr_end)();
extern void (*pr_rule)(), (*pr_line)();

static void lj_bitset();
static void lj_lin();

static char *ljbuf;
static int ljdpi = 0;
static int xpos = 0;
static int firstprint = 1;
static unsigned char m[8] =
{0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
static unsigned int nl = 0;

void
lj_setup()
{
    if (p_gmode == 3) {
	ljdpi = 300;
	a_coef = (float) 0.739;
	b_coef = (float) -0.973;
	c_coef = (float) 1.233;
    } else if (p_gmode == 1) {
	ljdpi = 150;
	a_coef = (float) 0.732;
	b_coef = (float) 0.178;
	c_coef = (float) 0.090;
    } else {
	ljdpi = 75;
    }
    nl = (ljdpi * 9 + 7) / 8;	/* allow for 9 inches across */
    ljbuf = (char *) calloc(nl, 1);

    adupi_x = ljdpi;
    adupi_y = ljdpi;
    pglen = (int) (adupi_x * 10.5);
    pgwid = (int) (adupi_y * 8.0);
    bpp = 1;
    dotdiam = (float) (1.0 / ljdpi);

    pr_idle = lj_idle;
    pr_page = lj_page;
    pr_out = lj_out;
    pr_end = lj_end;
    pr_rule = NULL;
    pr_line = NULL;
}

void
lj_out(buf, n)
unsigned char *buf;
int     n;
{
    char    s[20];

    if (firstprint) {
	lprs("\033E");		/* RESET */
	lprs("\033&l0E");	/* top margin = 0 */
	(void) sprintf(s, "\033*t%dR", ljdpi);
	lprs(s);		/* set resolution */
	lprs("\033*r0A");	/* start graphics at left margin */
	firstprint = 0;
	xpos = 0;
    }
    lj_bitset(buf, n);
    xpos++;
}

void
lj_end(int prompt)
{
    prompt = prompt;
    lprs("\033E");		/* RESET */
    lprfl();
}

void
lj_idle()
{
    lprs("\r");			/* send <CR> to printer to keep it on line */
}

void
lj_page()
{
    if (xpos > 0) {
	lprs("\f");		/* send <FORMFEED> to printer */
	xpos = 0;
    }
}

static void
lj_bitset(buf, n)
unsigned char *buf;
int     n;
{
    int     i, j;

    for (j = 0; j < 8; j++) {
	for (i = 0; i < n; i++)
	    if (buf[i] & m[j])
		ljbuf[i / 8] |= m[i % 8];
	lj_lin(ljbuf, nl);
    }
}

/* lj_lin - print a line of graphics */

static void
lj_lin(pb, n)
char   *pb;
unsigned int n;
{
    char    s[20];

    while (n > 1 && pb[n - 1] == 0)	/* trim zeros from end */
	n--;
    (void) sprintf(s, "\033*b%dW", n);
    lprs(s);
    lpr(pb, n);
    while (--n > 0)		/* zero buffer */
	pb[n] = 0;
}
