/* lpgtext.c */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "lpg.h"
#include "font.h"

#ifndef WIN32
#define _hypot hypot
#endif /* MSC */

#define MAXN 99
#define DRAW 0
#define MOVE 1
#define CIRC 5
#define ARC  6
#define ENDC 7
#define NSPOTMX	1024
#define NCIRCMX	256

#define xbyte(x, y)        (((x)/BPS)+(y)*(ov->xbyt))
#define dotset(x, y)	(ov->pbb[xbyte(x, y)] |= mask[(x)%BPS])
#define dotclr(x, y)	(ov->pbb[xbyte(x, y)] &= ~mask[(x)%BPS])

OVRLAY *getovrly();
double  txtlen();
int     bb_alloc(double);
int     onspot();
int     text_extent();
void    lpgrule(long, long, long, long);

extern void (*pr_line) ();
extern FILE *lgf;
extern int cpu_lsb;

static void bb_box();
static void bb_draw();
static void bb_line();
static void circset();
static void drachar();
static void dranodes();
static void initfont();
static void make_spot();
static void sktext();
static void spotset();

static double pnw = 0.01;
static double chh = 0.2;
static double chux = 0.0;
static double chuy = 1.0;
static double chhal[4] ={0.0, 0.0, 0.5, 1.0};
static double chval[6] ={0.0};
static double chlwt = 1.0;
static double lastpen = 0;
static int nspot = 0;
static int ncirc = 0;
static int xtent_draw = 1;
static int xspot[NSPOTMX];
static int yspot[NSPOTMX];
static int xcirc[NCIRCMX];
static int ycirc[NCIRCMX];
static int tx_hal = 1;
static int tx_val = 4;
static long ex1 = 0;
static long ex2 = 0;
static long ey1 = 0;
static long ey2 = 0;
static long tx_xtnt_x[4] ={0};
static long tx_xtnt_y[4] ={0};
static long tx_xllc = 0;
static long tx_yllc = 0;
static OVRLAY *ov = NULL;
static FONT *skfont = NULL;
static TRANSF ndt;

void
lpgtext(double x, double y, char *str)
{
    initfont();
    if (text_extent(x, y, str) == 0)
	return;
    pnw = chlwt * chh * 0.1;
    if (pr_line == NULL) {
	if (bb_alloc(pnw))
	    return;
    }
    if (xtent_draw != 0)
	bb_box(tx_xtnt_x, tx_xtnt_y);
    sktext(str);
}

void
lpgchup(double x, double y)
{
    double  d;

    d = _hypot(x, y);
    if (d <= 0.0) {
	fprintf(lgf, "*** Warning: Invalid character orientation\n");
	return;
    }
    chux = x / d;
    chuy = y / d;
}

void
lpgchlw(double lwt)
{
    chlwt = lwt;
}

void
lpgchsz(double siz)
{
    chh = siz;
}

void
lpgchal(int hal, int val)
{
    tx_hal = hal;
    tx_val = val;
}

void
lpgchxt(int xtnt)
{
    xtent_draw = xtnt;
}

int
text_extent(x, y, str)
double  x, y;
char   *str;
{
    int i;
    double  len, x0, y0, x1, y1, x2, y2, xllc, yllc;

    if ((len = txtlen(str)) <= 0)
	return (0);
    x0 = len * (chhal[1] - chhal[tx_hal]);
    y0 = chh * (chval[4] - chval[tx_val]);
    x1 = len * (chhal[1] - chhal[1]);
    y1 = chh * (chval[5] - chval[4]);
    x2 = len * (chhal[3] - chhal[1]);
    y2 = chh * (chval[1] - chval[4]);
    xllc = x + x0 * chuy + y0 * chux;
    yllc = y - x0 * chux + y0 * chuy;
    tx_xllc = (long) (xllc * adupi_x);
    tx_yllc = (long) (yllc * adupi_y);
    tx_xtnt_x[0] = (long) ((xllc + x1 * chuy + y1 * chux) * adupi_x);
    tx_xtnt_y[0] = (long) ((yllc - x1 * chux + y1 * chuy) * adupi_y);
    tx_xtnt_x[1] = (long) ((xllc + x2 * chuy + y1 * chux) * adupi_x);
    tx_xtnt_y[1] = (long) ((yllc - x2 * chux + y1 * chuy) * adupi_y);
    tx_xtnt_x[2] = (long) ((xllc + x2 * chuy + y2 * chux) * adupi_x);
    tx_xtnt_y[2] = (long) ((yllc - x2 * chux + y2 * chuy) * adupi_y);
    tx_xtnt_x[3] = (long) ((xllc + x1 * chuy + y2 * chux) * adupi_x);
    tx_xtnt_y[3] = (long) ((yllc - x1 * chux + y2 * chuy) * adupi_y);

    ex1 = ex2 = tx_xtnt_x[0];
    ey1 = ey2 = tx_xtnt_y[0];
    for (i = 1; i < 4; i++) {
	if (tx_xtnt_x[i] < ex1)
	    ex1 = tx_xtnt_x[i];
	if (tx_xtnt_y[i] < ey1)
	    ey1 = tx_xtnt_y[i];
	if (tx_xtnt_x[i] > ex2)
	    ex2 = tx_xtnt_x[i];
	if (tx_xtnt_y[i] > ey2)
	    ey2 = tx_xtnt_y[i];
    }

    return (1);
}

int
bb_alloc(double w)
{
    int     nn, x1, y1, x2, y2, xo, yo, dx, dy, xm, ym;

    if ((ov = getovrly()) == NULL) {
	fprintf(lgf, "*** Warning: failed to get overlay\n");
	return (1);
    }

    make_spot(w);
    dx = (int) (1 + w * adupi_x / 2);
    dy = (int) (1 + w * adupi_y / 2);
    x1 = ex1 - dx;
    y1 = ey1 - dy;
    x2 = ex2 + dy;
    y2 = ey2 + dy;
    if (x1 < 0 || y1 < 0 || y2 > pgwid) {
	fprintf(lgf, "*** Warning: bit block off page\n");
    }
    xm = x1 % BPS;
    ym = y1 % BPS;
    xo = (xm < 0) ? BPS + xm : xm;
    yo = (ym < 0) ? BPS + ym : ym;
    x1 -= xo;
    y1 -= yo;
    xm = x2 % BPS;
    ym = y2 % BPS;
    if (xm > 0)
	x2 += BPS - xm;
    if (ym > 0)
	y2 += BPS - ym;
    ov->xoff = xo;
    ov->yoff = yo;
    ov->xllc = x1;
    ov->yllc = y1;
    ov->xlen = x2 - ov->xllc;
    ov->ylen = y2 - ov->yllc;
    ov->xbyt = (ov->xlen + BPS - 1) / BPS;
    nn = ov->ylen * ov->xbyt;
    if ((ov->pbb = (char *) calloc((unsigned int) nn, 1)) == NULL) {
        fprintf(lgf, "*** Warning: Bit block allocation error\n");
        return (3);
    }
    ov->outs = 0;
    ov->otyp = OV_BTBK;

    return (0);
}

int
ov_line(int n, long *x, long *y, double w)
{
    int i;
    long *xx, *yy, xmn, xmx;
    
    if ((ov = getovrly()) == NULL) {
	fprintf(lgf, "*** Warning: failed to get overlay\n");
	return (1);
    }
    if ((ov->pbb = (char *) calloc(2 * n, sizeof(long))) == NULL) {
        fprintf(lgf, "*** Warning: Overly line allocation error\n");
        return (2);
    }
    xx = (long *) ov->pbb;
    yy = xx + n;
    xmn = xmx = xx[0];
    for (i = 0; i < n; i++) {
	xx[i] = x[i];
	yy[i] = y[i];
	if (xmn > xx[i])
	    xmn = xx[i];
	if (xmx < xx[i])
	    xmx = xx[i];
    }
    ov->xoff = 0;
    ov->yoff = 0;
    ov->xllc = xmn;
    ov->yllc = n;
    ov->xlen = xmx - xmn + 1;
    ov->ylen = (long) (w * 1000);
    ov->xbyt = 0;
    ov->outs = 0;
    ov->otyp = OV_LINE;

    return (0);
}

static void
bb_draw(int n, long *x, long *y, double w)
{
    int     i;

    if (n < 2)
	return;
    if (pr_line) {
	//pr_line(n, x, y, w);
	ov_line(n, x, y, w);
    } else if (ov) {
	for (i = 1; i < n; i++)
	    bb_line(x[i - 1], y[i - 1],	x[i], y[i]);
    }
}

/* bb_line - convert a single line segment to a seqence of pixels */

void
bb_line(long x0, long y0, long x1, long y1)
{
    int     dx, dy;
    int     xinc, yinc;
    int     res1;
    int     res2;

    x0 -= ov->xllc;
    y0 -= ov->yllc;
    x1 -= ov->xllc;
    y1 -= ov->yllc;

    if (x1 >= x0) {
	dx = x1 - x0;
	xinc = 1;
    } else {
	dx = x0 - x1;
	xinc = -1;
    }
    if (y1 >= y0) {
	dy = y1 - y0;
	yinc = 1;
    } else {
	dy = y0 - y1;
	yinc = -1;
    }
    res1 = 0;
    res2 = 0;
    if (dx == dy) {
	while (x0 != x1) {
	    circset(x0, y0);
	    y0 += yinc;
	    x0 += xinc;
	}
    } else if (dx > dy) {
	while (x0 != x1) {
	    circset(x0, y0);
	    if (res1 > res2) {
		res2 += dx - res1;
		res1 = 0;
		y0 += yinc;
	    }
	    res1 += dy;
	    x0 += xinc;
	}
    } else {
	while (y0 != y1) {
	    circset(x0, y0);
	    if (res1 > res2) {
		res2 += dy - res1;
		res1 = 0;
		x0 += xinc;
	    }
	    res1 += dx;
	    y0 += yinc;
	}
    }
    spotset(x1, y1);
}


/* make_spot - define a pattern of pixels to approximate a circular spot
 *             The diameter of the spot should be about pen/300 inches.
 */
static void
make_spot(pen)
double  pen;
{
    double  siz;
    int     i, j, i1, j1, i2, j2, x, y;
    int     xsize, ysize, xdsq, ydsq;
    long    xsc, ysc, sizesq, dsq;

    if (pen == lastpen)
	return;
    lastpen = pen;
    xsize = (int) (1 + (pen - dotdiam) * adupi_x);
    if (xsize < 1)
	xsize = 1;
    ysize = (int) (1 + (pen - dotdiam) * adupi_y);
    if (ysize < 1)
	ysize = 1;
    nspot = 0;
    if (xsize > 1 || ysize > 1) {	/* size is bigger than 1 dot */
	i1 = -xsize / 2;
	j1 = -ysize / 2;
	i2 = xsize / 2;
	j2 = ysize / 2;
	xsc = ((long) adupi_y * adupi_y);
	ysc = ((long) adupi_x * adupi_x);
	siz = (pen - dotdiam) * adupi_x * adupi_y;
	sizesq = (long) (1 + siz * siz);
	for (i = i1; i <= i2; i++) {
	    for (j = j1; j <= j2; j++) {
		xdsq = (xsize % 2) ? (i * i) * 4 : (i * i + i) * 4 + 1;
		ydsq = (ysize % 2) ? (j * j) * 4 : (j * j + j) * 4 + 1;
		dsq = xdsq * xsc + ydsq * ysc;
		if (dsq < sizesq && nspot < NSPOTMX) {
		    xspot[nspot] = i;
		    yspot[nspot] = j;
		    nspot++;
		}
	    }
	}
    }
    if (nspot == 0) {
	nspot = 1;
	xspot[0] = 0;
	yspot[0] = 0;
    }
    ncirc = 0;
    for (i = 0; i < nspot; i++) {
	x = xspot[i];
	y = yspot[i];
	if (ncirc < NCIRCMX && (
		!onspot(x - 1, y) || !onspot(x + 1, y) || 
		!onspot(x, y - 1) || !onspot(x, y + 1) ||
		!onspot(x - 1, y - 1) || !onspot(x + 1, y - 1) || 
		!onspot(x - 1, y + 1) || !onspot(x + 1, y + 1))) {
	    xcirc[ncirc] = x;
	    ycirc[ncirc] = y;
	    ncirc++;
	}
    }
}

int
onspot(int x, int y)
{
    int     i;

    for (i = 0; i < nspot; i++)
	if (xspot[i] == x && yspot[i] == y)
	    return (1);
    return (0);
}


/* spotset - set pixels at (x,y) according to predefined spot pattern */
static void
spotset(x, y)
long    x, y;
{
    int     i;

    if (nspot == 1)
	dotset(x, y);
    else
	for (i = 0; i < nspot; i++)
	    dotset(x + xspot[i], y + yspot[i]);
}

/* circset - set pixels at (x,y) according to predefined circ pattern */
static void
circset(x, y)
long    x, y;
{
    int     i;

    if (ncirc == 1)
	dotset(x, y);
    else
	for (i = 0; i < ncirc; i++)
	    dotset(x + xcirc[i], y + ycirc[i]);
}

static void
bb_box(x, y)
long   *x, *y;
{
    int     i;
    long    xx[5], yy[5];

    for (i = 0; i < 4; i++) {
	xx[i] = x[i];
	yy[i] = y[i];
    }
    xx[4] = xx[0];
    yy[4] = yy[0];
    bb_draw(5, xx, yy, pnw);
}

static void
sktext(pstr)
char   *pstr;
{
    int     k;
    double xb, yb, xt, yt, sc;
    float   chr_xleft, chr_width, chr_ybase;

    xt = tx_xllc;
    yt = tx_yllc;
    sc = chh / skfont->fchh;
    ndt.txx = (float) (chuy * sc * adupi_x);
    ndt.txy = (float) (chux * sc * adupi_x);
    ndt.tyx = (float) (-chux * sc * adupi_y);
    ndt.tyy = (float) (chuy * sc * adupi_y);
    chr_ybase = skfont->fval[4];/* Vertical offset to base of char */
    while (*pstr) {
	k = *pstr++;
	if (32 <= k && k < 127) {	/* printable ascii characters */
	    chr_xleft = skfont->xlft[k];
	    chr_width = skfont->xadv[k];
	    xb = ndt.txx * chr_xleft - ndt.txy * chr_ybase;
	    yb = ndt.tyx * chr_xleft - ndt.tyy * chr_ybase;
	    drachar(xt + xb, yt + yb, k);
	    xt += ndt.txx * chr_width;
	    yt += ndt.tyx * chr_width;
	}
    }
}

/*
 * drachar - draw a character using table of symbol node definitions
 */
static void
drachar(xllt, yllt, k)
double  xllt, yllt;
int     k;
{
    short int *fnode, *lnode;

    /* check range of character code */
    if (k < 0 || k >= skfont->nkhar) {
	return;
    }
    /* find first and last nodes in table */
    fnode = skfont->node + skfont->khar[k] - 1;
    lnode = skfont->node + skfont->khar[k + 1] - 2;
    ndt.txo = (float) (xllt - (ndt.txx + ndt.txy) * skfont->fnx0);
    ndt.tyo = (float) (yllt - (ndt.tyx + ndt.tyy) * skfont->fny0);
    dranodes(fnode, lnode, &ndt);
}

static void
dranodes(fnode, lnode, n)
short int *fnode, *lnode;
TRANSF *n;
{
    short int *pnode;
    int     nodtyp, lsb, msb;
    float   x, y;
    int     ii = 0;
    long    xt[MAXN], yt[MAXN];

    lsb = cpu_lsb ? 1 : 0;
    msb = cpu_lsb ? 0 : 1;
    for (pnode = fnode; pnode <= lnode; pnode++) {
	y = ((unsigned char *) pnode)[lsb];
	x = ((unsigned char *) pnode)[msb];
	nodtyp = (int) ((x == 255) ? y : DRAW);
	switch (nodtyp) {
	case DRAW:
	    xt[ii] = (long) (n->txx * x + n->txy * y + n->txo);
	    yt[ii] = (long) (n->tyx * x + n->tyy * y + n->tyo);
	    ii++;
	    if (ii >= MAXN) {
		bb_draw(ii, xt, yt, pnw);
		xt[0] = xt[ii - 1];
		yt[0] = yt[ii - 1];
		ii = 0;
	    }
	    break;
	case MOVE:
	    bb_draw(ii, xt, yt, pnw);
	    ii = 0;
	    break;
	case ENDC:
	    lnode = pnode;
	    break;
	}
    }
    bb_draw(ii, xt, yt, pnw);
}

double
txtlen(pstr)
char   *pstr;
{
    int     k;
    double  xt, txx;

    xt = 0;
    txx = chh / skfont->fchh;
    while (*pstr) {
	k = *pstr++;
	if (32 <= k && k < 127) {	/* printable ascii characters */
	    xt += txx * skfont->xadv[k];
	}
    }
    return (xt);		/* return string length in inches */
}

static void
initfont()
{
    int     i;
    double  txx;

    skfont = &stdfont_;
    txx = 1.0 / skfont->fchh;
    for (i = 0; i < 6; i++)
	chval[i] = txx * skfont->fval[i];
}


double
line_extent(double x1, double y1, double x2, double y2)
{
    double  xmin, ymin, xmax, ymax;

    if (x1 < x2) {
	xmin = x1;
	xmax = x2;
    } else {
        xmin = x2;
        xmax = x1;
    }
    if (y1 < y2) {
        ymin = y1;
        ymax = y2;
    } else {
        ymin = y2;
        ymax = y1;
    }
    ex1 = (long) (xmin * adupi_x + 0.5);
    ey1 = (long) (ymin * adupi_y + 0.5);
    ex2 = (long) (xmax * adupi_x + 0.5);
    ey2 = (long) (ymax * adupi_y + 0.5);

    return ((xmax - xmin) * (ymax - ymin));
}

void
lpgline(double x1, double y1, double x2, double y2, double w)
{
    double  a, xm, ym;
    long    x[2], y[2];
    static double amax = 0.25;

    x[0] = (long) (x1 * adupi_x + 0.5);
    y[0] = (long) (y1 * adupi_y + 0.5);
    x[1] = (long) (x2 * adupi_x + 0.5);
    y[1] = (long) (y2 * adupi_y + 0.5);
    if (pr_line) {
	//pr_line(2, x, y, w);
	ov_line(2, x, y, w);
	return;
    }
    a = line_extent(x1, y1, x2, y2);
    if (a > amax) {   			/* divide long lines */
        xm = (x1 + x2) / 2;
        ym = (y1 + y2) / 2;
        lpgline(x1, y1, xm, ym, w);
        lpgline(xm, ym, x2, y2, w);
        return;
    }
    if (bb_alloc(w))
	return;
    bb_line(x[0], y[0], x[1], y[1]);
}

void
lpgdraw(int n, double *x, double *y, double w)
{
    int i;

    for (i = 1; i < n; i++)
	lpgline(x[i - 1], y[i - 1], x[i], y[i], w);

}
