/* pc.c - console routines for PC version of SPECTO */

#ifdef MSDOS
#include <stdio.h>
#include <conio.h>
#include <dos.h>

#define UNKNOWN	        0x00	/* Unknown board type */
#define MDA             0x01	/* IBM Monochrome card */
#define CGA             0x02	/* IBM Color graphics adapter (CGA) */
#define HGC             0x04	/* Hercules monochrome graphics card */
#define EGA_MONO        0x10	/* IBM Enhanced graphics adapter (EGA)
				   w/monochrome display */
#define EGA_COLOR       0x20	/* EGA w/color display */
#define EGA_HIGH        0x30	/* EGA w/high resolution color display */
#define PEGA            0x40	/* Paradise Autoswitch EGA, 640 x 480 */
#define VGA_MONO        0x50	/* IBM VGA hi res 2 color, 640x480 */
#define VGA_COLOR       0x60	/* IBM VGA hi res 16 color, 640x480 */

#define ESC     	27
#define NPAL            16
#ifndef _lint
#define CGA_ADDR	0xB8000000
#define EGA_ADDR        0xA0000000
#define HGC_ADDR        0xB0000000
#define PEGA_ADDR	0xC0000058
#define UCHAR           unsigned char far
#else
#define CGA_ADDR	0
#define EGA_ADDR        0
#define HGC_ADDR        0
#define PEGA_ADDR	0
#define UCHAR           unsigned char
#endif
#define	CGAbyte(x,y)	(CGAaddr + (((y)&1)*0x2000) + (((y)>>1)*80) + (x))
#define	HGCbyte(x,y)	(HGCaddr + (ytbl[y]) + (x))
#define	EGAbyte(x,y)	(EGAaddr + ((((y)<<2) + (y))<<4) + (x))
#define MAPMSK(m)	{_disable(); _outp(0x3C4, 2); _outp(0x3C5, m); _enable();}

void    mouse();
void    co_setup(), co_idle(), co_page(), co_out(), co_end();

extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) ();
extern FILE *lgf;
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int bpp;
extern int p_gmode;
extern int p_color;
extern int noquery;
extern double dotdiam;
extern double a_coef, b_coef, c_coef;

int     xpix = 0;
int     ypix = 0;

static int video_test();
static void co_prompt();
static void fill_mem();
static void m6845();
static void delay();
static void gr_init();
static void gr_check();
static void gr_clearscr();
static void gr_colm();
static void gr_reset();
static void gr_setbyte();
static void egaclearscr();
static void egacolm();
static void hgcclearscr();
static void hgccolm();
static void hgc_graph_mode();
static void cgaclearscr();
static void cgacolm();
static void hls_palette();
static void ega_palette();
static void vga_palette();

static int xdpi = 0;
static int ydpi = 0;
static int firstprint = 1;
static int xbyte = 0;

static short graphics_mode = 0;
static short can_do_graphics = 0;
static short _pc_adapt;
static short _vid_mode;
static short _old_vid_mode;
static union _REGS rg;		/* cpu register for use of DOS calls */
static UCHAR *CGAaddr = (UCHAR *) CGA_ADDR;
static UCHAR *HGCaddr = (UCHAR *) HGC_ADDR;
static UCHAR *EGAaddr = (UCHAR *) EGA_ADDR;
static unsigned int index_reg = 0x3b4;
static unsigned int data_reg = 0x3b5;
static unsigned int mode_select_reg = 0x3b8;
static unsigned int status_reg = 0x3ba;
static unsigned int ytbl[348];
static int gr_fntwid = 8;
static int gr_fnthgt = 8;

static float hue[16] = {
     0,   0,   8,  16,  24,  32,  42, 58,
    59, 80, 150, 165, 170, 180, 240,  0
};

void
co_setup()
{
    gr_check();
    if (p_gmode > 0) {
	_vid_mode = p_gmode;	/* override automatic setting */
	can_do_graphics = 1;
    }
    switch (_vid_mode) {
    case 6:			/* CGA 640x200x1 */
	xdpi = 60;
	ydpi = 25;
	xpix = 640;
	ypix = 200;
	bpp = 1;
	/* the default values for a_coef, b_coef, and c_coef are OK. */
	break;
    case 8:			/* HGC 720x348x1 */
	xdpi = 60;
	ydpi = 40;
	xpix = 720;
	ypix = 348;
	bpp = 1;
	/* the default values for a_coef, b_coef, and c_coef are OK. */
	break;
    case 16:			/* EGA 640x350x4 */
	xdpi = 60;
	ydpi = 40;
	xpix = 640;
	ypix = 350;
	bpp = p_color ? 4 : 2;
	if (!p_color) {
	    a_coef = 0.89;
	    b_coef = -1.36;
	    c_coef = 1.45;
	}
	break;
    case 17:			/* VGA 640x480x1 */
	xdpi = 60;
	ydpi = 60;
	xpix = 640;
	ypix = 480;
	bpp = 1;
	/* the default values for a_coef, b_coef, and c_coef are OK. */
	break;
    case 18:			/* VGA 640x480x4 */
	xdpi = 60;
	ydpi = 60;
	xpix = 640;
	ypix = 480;
	bpp = p_color ? 4 : 2;
	if (!p_color) {
	    a_coef = 0.89;
	    b_coef = -1.36;
	    c_coef = 1.45;
	}
	break;
    case 80:			/* PEGA 640x480x4 */
	xdpi = 60;
	ydpi = 60;
	xpix = 640;
	ypix = 480;
	bpp = p_color ? 4 : 2;
	/*
	   a_coef, b_coef and c_coef have not been tested. Default values
	   used
	*/
	break;
    default:
	fprintf(lgf, "PC must have CGA, EGA, VGA or HGC\n");
	leave(1);
    }

    adupi_x = xdpi;
    adupi_y = ydpi;
    pglen = (int) (adupi_x * 10.5);
    pgwid = (int) (adupi_y * 8.0);
    dotdiam = 1.0 / xdpi;

    pr_idle = co_idle;
    pr_page = co_page;
    pr_out = co_out;
    pr_end = co_end;
    pr_rule = NULL;
    pr_line = NULL;
}

void
co_out(unsigned char *buf, int n)
{
    fflush(lgf);
    if (firstprint) {
	if (_isatty(_fileno(lgf))) {
            if (!noquery) {
	        fprintf(lgf, "\n    [press any key to continue]\n");
	        (void) _getch();
	    }
	    lgf = fopen("specto.out", "w");
	}
	gr_init();
	xbyte = 0;
	firstprint = 0;
    }
    gr_colm(xbyte++, buf, n);
}

void
co_end(int prompt)
{
    if (prompt && xbyte > 0)
	co_prompt();
    gr_reset();
}

void
co_idle()
{
}

void
co_page()
{
    if (xbyte > 0) {
	co_prompt();
	xbyte = 0;
    }
}

void
co_prompt()
{
    if (!noquery) {
        (void) _putch('\007');
        mouse();
    }
    gr_clearscr();
}

/*****************************************************************************/
int
HLS2RGB(float n1, float n2, float h)
{
    int     v;

    if (h > 360)
	h -= 360;
    if (h < 0)
	h += 360;

    if (h < 60)
	v = (int) (255.0 * (n1 + (n2 - n1) * h / 60));
    else if (h < 180)
	v = (int) (255.0 * n2);
    else if (h < 240)
	v = (int) (255.0 * (n1 + (n2 - 1) * (240 - h) / 60));
    else
	v = (int) (255.0 * n1);

    return v;
}

/*****************************************************************************/

void
gr_init()
{
    int     i;
    float   l = 0.5, s = 1;

    if (can_do_graphics) {
	if (_vid_mode == 8) {
	    hgc_graph_mode();
	} else {
	    rg.h.ah = 0;	/* function number: set video mode */
	    rg.h.al = _vid_mode;
	    (void) _int86(0x10, &rg, &rg);	/* go for it! */
	    if (bpp == 4) {
		hls_palette(15, 1.0, 1.0, 0.0); /* white background */
		hls_palette( 0, 0.0, 1.0, 0.0); /* black foreground */
		for (i = 1; i < 15; i++) {
		    hls_palette(i, l, s, hue[i]);
		}
	    }
	}
	gr_clearscr();
    }
    graphics_mode = can_do_graphics;
}

void
gr_check()
{
    rg.h.ah = 15;		/* function number: get video mode */
    (void) _int86(0x10, &rg, &rg);	/* go for it! */
    _old_vid_mode = rg.h.al;
    _vid_mode = rg.h.al;

    _pc_adapt = video_test();

    switch (_pc_adapt) {
    case CGA:
	_vid_mode = 6;
	can_do_graphics = 1;
	break;
    case MDA:
	_vid_mode = 7;
	can_do_graphics = 0;
	break;
    case HGC:
	_vid_mode = 8;
	can_do_graphics = 1;
	break;
    case EGA_MONO:
    case EGA_COLOR:
    case EGA_HIGH:
	_vid_mode = 16;
	can_do_graphics = 1;
	break;
    case VGA_MONO:
	_vid_mode = 17;
	can_do_graphics = 1;
	break;
    case VGA_COLOR:
	_vid_mode = 18;
	can_do_graphics = 1;
	break;
    case PEGA:
	_vid_mode = 80;
	can_do_graphics = 1;
	break;
    default:
	_vid_mode = 7;
	can_do_graphics = 0;
    }
    graphics_mode = 0;
}

void
gr_clearscr()
{
    switch (_vid_mode) {
	case 6:
	cgaclearscr();
	break;
    case 8:
	hgcclearscr();
	break;
    case 16:
    case 17:
    case 18:
	egaclearscr();
	break;
    }
}

void
gr_setbyte(x, y, c)
{
    UCHAR  *pb;

    x /= 8;
    switch (_vid_mode) {
    case 6:
	pb = CGAbyte(x, y);
	break;
    case 8:
	pb = HGCbyte(x, y);
	break;
    case 16:
    case 17:
    case 18:
	pb = EGAbyte(x, y);
	break;
    }
    *pb++ = c;
}

void
gr_putc(x, y, c)
int     x, y, c;
{
    int     i;
    unsigned char *fp;
    unsigned char *gr_fntptr = (unsigned char *) 0xF000FA6E;

    fp = gr_fntptr + gr_fnthgt * c;
    for (i = 0; i < gr_fnthgt; i++) {
	gr_setbyte(x, y - gr_fnthgt + i, 0xFF);
	gr_setbyte(x, y - gr_fnthgt + i, ~fp[i]);
    }
}

void
gr_text(x, y, s)
int     x, y;
char   *s;
{
    int     c;

    for (; c = (*s++) & 0xFF; x += gr_fntwid)
	gr_putc(x, y, c);
}

void
gr_colm(x, b, n)
int     x, n;
unsigned char *b;
{
    switch (_vid_mode) {
    case 6:
	cgacolm(x + 5, b, n);
	break;
    case 8:
	hgccolm(x + 5, b, n);
	break;
    case 16:
    case 17:
    case 18:
	egacolm(x, b, n);
	break;
    }
}

void
gr_reset()
{
    if (graphics_mode) {
	if (_vid_mode != _old_vid_mode) {
	    rg.h.ah = 0;	/* function number: set video mode */
	    rg.h.al = _old_vid_mode;
	    (void) _int86(0x10, &rg, &rg);	/* go for it! */
	    _vid_mode = _old_vid_mode;
	}
	graphics_mode = 0;
    }
}

/*****************************************************************************/

static int
video_test()
{
    char   *pega_key = "PEGA BIOS";
    char far *pega_sig = (char far *) PEGA_ADDR;
    unsigned char far *equip_stat = (unsigned char far *) 0x410;
    unsigned int *page1 = (unsigned int *) CGA_ADDR;
    int     i, tmp, dipsw, mda_regs = 0, cga_regs = 0;


    /* Check for MCGA or VGA. */

    rg.x.ax = 0x1a00;
    (void) _int86(0x10, &rg, &rg);
    if (rg.h.al == 0x1a) {
	switch (rg.h.bl) {
	case 7:
	    return (VGA_MONO);
	case 8:
	    return (VGA_COLOR);
	default:
	    return (UNKNOWN);
	}
    }
    /* Check for EGA card */

    rg.h.ah = 0x12;		/* undocumented function for INT 10H */
    rg.h.bl = 0x10;
    (void) _int86(0x10, &rg, &rg);
    if (rg.h.bl != 0x10) {
	dipsw = rg.h.cl & 0x0e;
	dipsw = (dipsw >= 6) ? dipsw - 6 : dipsw;
	for (i = 0; pega_key[i] != '\0' && pega_key[i] == pega_sig[i]; i++)
	    continue;
	if (pega_key[i] == 0)
	    return (PEGA);
	if (dipsw == 4)
	    return (EGA_MONO);
	else if (dipsw == 2)
	    return (EGA_HIGH);
	return (EGA_COLOR);
    }
    /* Check for CGA registers */

    (void) _outp(0x3D4, 0xF);	/* address the cursor register */
    tmp = _inp(0x3D5);		/* read it and save current value */
    (void) _outp(0x3D5, 100);	/* send arbitrary value to register */
    if (_inp(0x3D5) == 100) {	/* if same value, then card is CGA is present */
	cga_regs++;
	(void) _outp(0x3D5, tmp);/* restore previous value */
    }
    if (cga_regs && (*equip_stat & 48) == 32)	/* must be CGA */
	return (CGA);

    /* Check for MDA registers */

    (void) _outp(0x3B4, 0xF);	/* address the cursor register */
    tmp = _inp(0x3B5);		/* read it and save current value */
    (void) _outp(0x3B5, 100);	/* send arbitrary value to register */
    if (_inp(0x3B5) == 100) {	/* if same value, then card is MDA is present */
	mda_regs++;
	(void) _outp(0x3B5, tmp);/* restore previous value */
    }
    if (mda_regs) {		/* MDA is present */
	(void) _outp(0x3BF, 3);	/* try to enable full HGC configuration */
	tmp = *page1;		/* read and save first word in page 1 */
	*page1 = 1234;		/* write arb. value to page 1 */
	if (*page1 == 1234) {
	    *page1 = tmp;	/* restore first word to prev. value */
	    return (HGC);
	}
	return (MDA);
    }
    return (UNKNOWN);
}

/*****************************************************************************/

void
egaclearscr()
{
    UCHAR  *pb;
    int     x, y, c;

    c = 0xFF;
    for (y = 0; y < ypix; y++) {
	pb = EGAbyte(0, y);
	x = xpix / 8;
	while (x-- > 0)
	    *pb++ = c;
    }
}

void
egacolm(x, b, n)
int     x, n;
unsigned char *b;
{
    int     y, t;

    if (bpp == 1) {
	t = ypix - n;
	for (y = ypix - 1; y >= t; y--)
	    *EGAbyte(x, y) = ~(*b++);
    } else if (bpp == 2) {
	t = ypix - n / 2;
	for (y = ypix - 1; y >= t; y--) {
	    MAPMSK(7);
	    *EGAbyte(x, y) = ~(*b++);
	    MAPMSK(8);
	    *EGAbyte(x, y) = ~(*b++);
	}
	MAPMSK(15);
    } else if (bpp == 4) {
	t = ypix - n / 4;
	for (y = ypix - 1; y >= t; y--) {
	    MAPMSK(8);
	    *EGAbyte(x, y) = ~(*b++);
	    MAPMSK(4);
	    *EGAbyte(x, y) = ~(*b++);
	    MAPMSK(2);
	    *EGAbyte(x, y) = ~(*b++);
	    MAPMSK(1);
	    *EGAbyte(x, y) = ~(*b++);
	}
	MAPMSK(15);
    }
}

/*****************************************************************************/

void
hgcclearscr()
{
    UCHAR  *pb;
    int     x, y;

    for (y = 0; y < ypix; y++) {
	pb = HGCbyte(0, y);
	x = xpix / 8;
	while (x-- > 0)
	    *pb++ = 0xFF;
    }
}

void
hgccolm(x, b, n)
int     x, n;
unsigned char *b;
{
    int     y, t;

    t = ypix - n;
    for (y = ypix - 1; y >= t; y--)
	*HGCbyte(x, y) = ~(*b++);
}

void
hgc_graph_mode()
{
    int     i, j;
    static char seq1[16] =
    {0x35, 0x2d, 0x2e, 0x7, 0x5b, 0x2, 0x57, 0x57,
	0x2, 0x3, 0, 0, 0, 0, 0, 0};

    for (i = 0, j = 0; i < 348; j += 90) {
	ytbl[i++] = 0x0000 + j;
	ytbl[i++] = 0x2000 + j;
	ytbl[i++] = 0x4000 + j;
	ytbl[i++] = 0x6000 + j;
    }
    (void) _outp(mode_select_reg, 3);
    m6845(seq1);
    fill_mem(0);
    delay(10000);
    (void) _outp(mode_select_reg, 11);
}

/* Fills the screen entirely with the specified bit pattern. */

static void
fill_mem(pattern)
register pattern;
{
    register count, row;
    int far *v;

    pattern = (pattern << 8) | pattern;
    row = 0;
    while (row < 348) {
	v = (int far *) (HGCaddr + ytbl[row++]);
	count = 45;
	while (count--)
	    *v++ = pattern;
    }
}


/* Used to send data to the hercules card to control its status. */
static void
m6845(seq)
char   *seq;
{
    int     i;

    while (_inp(status_reg) & 0x80);
    _disable();
    for (i = 0; i < 16; i++) {
	(void) _outp(index_reg, i);
	(void) _outp(data_reg, seq[i]);
    }
    _enable();
}


/* Allows a delay, determined by the count parameter "c". */
static void
delay(c)
{
    while (c--);
}

/*****************************************************************************/

void
cgaclearscr()
{
    UCHAR  *pb;
    int     x, y;

    for (y = 0; y < ypix; y++) {
	pb = CGAbyte(0, y);
	x = xpix / 8;
	while (x-- > 0)
	    *pb++ = 0xFF;
    }
}

void
cgacolm(x, b, n)
int     x, n;
unsigned char *b;
{
    int     y, t;

    t = ypix - n;
    for (y = ypix - 1; y >= t; y--)
	*CGAbyte(x, y) = ~(*b++);
}

/*****************************************************************************/

void
hls_palette(i, l, s, h)
double  l, s, h;
{
    int     r, g, b;

    if (l <= 0.5)
	s = l * (1 + s);
    else
	s = l + s - l * s;
    l = 2 * l - s;
    r = HLS2RGB(l, s, h + 120);
    g = HLS2RGB(l, s, h);
    b = HLS2RGB(l, s, h - 120);

    rg.x.ax = 0x1a00;
    (void) _int86(0x10, &rg, &rg);	/* Check for MCGA or VGA. */

    if (rg.h.al == 0x1a && rg.h.bl == 8) {	/* VGA COLOR */
	vga_palette(i, r, g, b);
    } else {
	ega_palette(i, r, g, b);
    }
}

void
ega_palette(int i, int rr, int gg, int bb)
{
    int     cv;			/* ega color value */

    rr >>= 6;
    gg >>= 6;
    bb >>= 6;
    cv = ((rr & 2) << 1) | ((rr & 1) << 5)
	| ((gg & 2)) | ((gg & 1) << 4)
	| ((bb & 2) >> 1) | ((bb & 1) << 3);

    rg.h.ah = 0x10;
    rg.h.al = 0;
    rg.h.bh = cv;
    rg.h.bl = i;
    (void) _int86(0x10, &rg, &rg);
}

void
vga_palette(int i, int rr, int gg, int bb)
{
    rg.h.ah = 0x10;
    rg.h.al = 0;
    rg.h.bh = i;
    rg.h.bl = i;
    (void) _int86(0x10, &rg, &rg);

    rg.h.ah = 0x10;
    rg.h.al = 16;
    rg.h.bh = 0;
    rg.h.bl = i;
    rg.h.dh = (rr & 0xFF) >> 2;
    rg.h.cl = (bb & 0xFF) >> 2;
    rg.h.ch = (gg & 0xFF) >> 2;
    rg.h.dl = 0;
    (void) _int86(0x10, &rg, &rg);
}

#endif				/* MSDOS */
