/* pc_mouse.c */

#include <dos.h>
#include <conio.h>
#include <math.h>
#include <string.h>
#include "lpg.h"
#include "frame.h"

#define	MINT		0x33	/* mouse interrupt */
#define	MMV             (0|MS)	/* mouse moved */
#define	MDL             (1|MS)	/* mouse down left */
#define	MDR             (2|MS)	/* mouse down right */
#define	FN              256	/* function key event */
#define	MS              512	/* mouse event */
#define	ESC             27

double spec_value();
int wave_value();

extern int ypix;

typedef unsigned char uchar;
typedef unsigned int uint;

static int gr_minit();
static int getevent();
static void gr_mshow();
static void gr_mstatus();

static double xg = 0, yg = 0;
static int mouse_detected = 0;
static int mlb, mrb, mcx, mcy;
static int spcflg = 1;

/* returns 1 if mouse is present and resets mouse, otherwise returns 0 */
int
gr_minit()
{
    union _REGS regs;

    regs.x.ax = 0;		/* function 0, reset and status */
    (void) _int86(MINT, &regs, &regs);
    return ((regs.x.ax == 0xFFFF) ? 1 : 0);
}


/* use nonzero flag to turn cursor on,  flag = 0 to turn cursor off */
void
gr_mshow(flag)
int     flag;
{
    union _REGS regs;

    regs.x.ax = (flag == 0) ? 2 : 1;	/* function 1: show cursor, funct 2:
					   hide */
    (void) _int86(MINT, &regs, &regs);
    return;
}


/* get current status of buttons and location, store in m */
void
gr_mstatus(but1, but2, x, y)
int    *but1;
int    *but2;
int    *x;
int    *y;
{
    union _REGS regs;

    regs.x.ax = 3;		/* function 3: get button status and mouse
				   position */
    (void) _int86(MINT, &regs, &regs);
    *but1 = (regs.x.bx & 1) == 1;	/* check bit 0 */
    *but2 = (regs.x.bx & 2) == 2;	/* check bit 1 */
    *x = regs.x.cx;
    *y = regs.x.dx;
    return;
}

int
getevent()
{
    int     c, llb, lrb;

    for (;;) {
	if (_kbhit()) {
	    c = _getch();
	    if (c == 0)
		c = FN | _getch();
	    return (c);
	}
	if (mouse_detected) {
	    llb = mlb;
	    lrb = mrb;
	    gr_mstatus(&mlb, &mrb, &mcx, &mcy);
	    if (mlb && !llb)
		return (MDL);
	    if (mrb && !lrb)
		return (MDR);
	}
    }
}

void
tstr(t, st)
double  t;
char   *st;
{
    if (t < 1e-5)
	sprintf(st, "%4.2f us", t * 1e6);
    else if (t < 1e-4)
	sprintf(st, "%4.1f us", t * 1e6);
    else if (t < 1e-3)
	sprintf(st, "%4.0f us", t * 1e6);
    else if (t < 1e-2)
	sprintf(st, "%4.2f ms", t * 1e3);
    else if (t < 0.1)
	sprintf(st, "%4.1f ms", t * 1e3);
    else if (t < 1)
	sprintf(st, "%4.0f ms", t * 1e3);
    else if (t < 10)
	sprintf(st, "%4.2f s", t);
    else if (t < 100)
	sprintf(st, "%4.1f s", t);
    else if (t < 1e6)
	sprintf(st, "%4.0f s", t);
    else
	sprintf(st, "INF");
}

void
fstr(f, sf)
double  f;
char   *sf;
{
    if (f > 1e12)
	sprintf(sf, "INF");
    else if (f > 1e6)
	sprintf(sf, "%4.2f MHz", f / 1e6);
    else if (f > 1e5)
	sprintf(sf, "%4.0f kHz", f / 1e3);
    else if (f > 1e4)
	sprintf(sf, "%4.1f kHz", f / 1e3);
    else if (f > 1e3)
	sprintf(sf, "%4.2f kHz", f / 1e3);
    else if (f > 1e2)
	sprintf(sf, "%4.0f Hz", f);
    else if (f > 10)
	sprintf(sf, "%4.1f Hz", f);
    else
	sprintf(sf, "%4.2f Hz", f);
}

double
pfmap(p, i)
double  p;
{
    double  f, a, b, c, p0, p1, p2;

    switch (fr_[i].fa_typ) {
    case 1:
	f = fr_[i].fa_min * exp(p * fr_[i].fa_nat);
	break;
    case 2:
	a = fr_[i].fa_pma;
	b = fr_[i].fa_pmb;
	c = fr_[i].fa_pmc;
	p1 = log10(fr_[i].fa_min / a + c) / b;
	p2 = log10(fr_[i].fa_max / a + c) / b;
	p0 = p1 + p * (p2 - p1);
	f = a * (pow(10.0, p0 * b) - c);
	break;
    default:
	f = fr_[i].fa_min + p * (fr_[i].fa_max - fr_[i].fa_min);
    }
    return (f);
}

void
info_line()
{
    double  x, y, t, f, p, v, m, d, sps;
    int     i, ix1, iy1, ix2, iy2, iw1, iw2, cx, cy, iv, nwn, nft;
    char    s[80], st[20], sf[20], *sdf;
    static  double eps = 1e-9;

    cx = mcx;
    cy = ypix - mcy;
    x = cx * xg;
    y = cy * yg;
    t = 0;
    f = 0;
    *s = '\0';
    for (i = 0; i < nfr; i++) {
	ix1 = fr_[i].ix_min;
	iy1 = fr_[i].iy_min;
	ix2 = fr_[i].ix_max;
	iy2 = fr_[i].iy_max;
	iw1 = fr_[i].iw_min;
	iw2 = fr_[i].iw_max;
        sdf = fr_[i].sd_fn;
        nwn = fr_[i].wn_wid;
        nft = fr_[i].wn_nft;
        sps = fr_[i].wv_sps;
	if (ix2 <= ix1 || iy2 <= iy1)
	    continue;
	if (cx >= ix1 && cx <= ix2) {
	    p = ((float) (cx - ix1)) / (ix2 - ix1);
	    t = fr_[i].ta_min + p * (fr_[i].ta_max - fr_[i].ta_min);
	    tstr(t, st);
	    if (cy >= iy1 && cy <= iy2) {
		p = ((float) (cy - iy1)) / (iy2 - iy1);
		f = pfmap(p, i);
		fstr(f, sf);
                v = spec_value(sdf, t, f, sps, nwn, nft);
                m = (fr_[i].sm_max < eps) ? eps : fr_[i].sm_max;
                d = (v < eps) ? -999 : 20 * log10(v / m);
		sprintf(s, " (t,f,v) = (%s, %s, %.1f dB) ", st, sf, d);
	    } else if (cy >= iw1 && cy <= iw2) {
                iv = wave_value(sdf, t, sps);
		sprintf(s, " (t,v) = (%s, %d) ", st, iv);
	    } else {
		sprintf(s, " t = %s ", st);
	    }
	    break;
	}
    }
    if (*s == '\0')
        sprintf(s, " (x,y) = (%3.2f in, %3.2f in)", x, y);
    for (i = strlen(s); i < 70; i++)
	s[i] = ' ';
    s[i] = '\0';
    gr_text(0, 10, s);
}

void
mouse()
{
    int     i;

    if (!mouse_detected) {
	if (gr_minit()) {
	    mouse_detected = 1;
	    gr_mshow(1);
	    xg = 1.0 / adupi_x;
	    yg = 1.0 / adupi_y;
	}
    }
    for (i = 0; i < 1;) {
	switch (getevent()) {
	case ESC:
	case '\n':
	case '\r':
	case MDR:
	    i++;
	    break;
	case MDL:
	    info_line();
    	    break;
        case 's':
            spcflg = 1;
	    info_line();
    	    break;
        case 'x':
            spcflg = 0;
	    info_line();
    	    break;
	}
    }
    if (mouse_detected)
	gr_mshow(0);
}
