/* process.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <fcntl.h>
#include "specto.h"
#include "frame.h"
#include "cparam.h"

#ifdef MGW
#define _kbhit kbhit
#define _getch getch
#endif /* MGW */

#define NF 	8192		/* maximum samples in each FFT frame */
#define NY 	(NF+2)		/* maximum points in frame buffer = NF */
#define NQ 	(NY/2)		/* maximum number of frequencies */
#define OK 	-1		/* status is okay */
#define EEOF 	-2		/* status is end-of-file */
#define MAXLPC	40
#define LPCSIZ	(MAXLPC+1)
#ifndef PI
#define PI  	3.1415926535898
#endif

#define limit(mn,x,mx) 	((mn>x)?mn:((mx<x)?mx:x))
#define pow2(x) 	pow(2.0,(x))
#define pow10(x) 	pow(10.0,(x))

double  lpgcomp();
float  *rd_tsd_data();
int     autoc(int, float *, int, float *, float *, float *, float *);
int     lpg_x();
int     lpg_y();
int     getwav();
int     match();
int     open_sdf();
int     open_tsd();
int     rfft(float *, int);
void    axes(int);
void    close_sdf();
void    close_tsd();
void    setup();
void    do_tsd();
void    dowave();
void    dospec();
void    dofile();
void    dospl();
void    dofft();
void    incomp();
void    doenrg();
void    do_wnr();
void    dohist();
void    lpfft();
void    scnwav();
void    fscnwav();
void    rfft_unpack();
void    guess_sdf_type();
void    leave(int);
void    lpgidle();
void    lpgini();
void    lpgpage();
void    lpgchup(double, double);
void    lpgchlw(double);
void    lpgchsz(double);
void    lpgchal(int, int);
void    lpgchxt(int);
void    lpggray(unsigned char *, int, int, int, int);
void    lpggrmp(float *);
void    lpgtext(double, double, char *);
void    t_axis(long);
void    dogrid(long);
void    lpgadv(int);
void    lpgrule(long, long, long, long);

static int atendfr();
static void initfr();
static void namedate();
static void info_sdf();
static void info_tsd();
static void process_sdf();
static void process_tsd();

extern char file_name[80];
extern char time_date[40];
extern int (*fetch) ();
extern int setup_lpg_called;
extern int printer, npts, nbyt, nstp, npm1;
extern int sdf_type;
extern float tm0, wnap;
extern FILE *lgf;

static int *iy = NULL;
static float *w = NULL;
static float *y = NULL;

static double __max_mag = 0;
static float *bbbbb;
static double meansq = 0;
static float ff_mx = 0;
static float apmn = 0;
static float apmx = 0;
static float appp = 0;
static double inprn = 0.0;
static double tpi = 2 * PI;
static long isamp = 0;
static long id[256] = {0};
static unsigned char px[NQ+1];
static struct _wave w0_;

void
process(int ft)
{
    double  xl, dx, fr, xmn, xmx, dxx, e10;
    int     fd, i, start_page = 0;
    static double xxmx = 10000.0;
    static float map[NQ+1];

    inprn = 0;
    if (ft == FT_TSD) {
	if (!open_tsd()) {
	    (void) fprintf(lgf, "ERROR :  can't open %s\n", wave_.sdfn);
	    leave(1);
	}
	strcpy(wave_.sdfn, pa_.stimfn);
	guess_sdf_type(pa_.stimfn);
	fd = open_sdf(sdf_type);
	if (!fd) {
	    (void) fprintf(lgf, "ERROR : can't open stimulus file %s\n", wave_.sdfn);
	    leave(1);
	}
	setup();
	scnwav();
	fscnwav();
	ucon_.fpma = (float) (limit(0, pa_.fpma, 1) * 1000);
	ucon_.fpmb = (float) (limit(0, pa_.fpmb, 1) * 100 / xxmx);
	ucon_.fpmc = (float) (limit(0, pa_.fpmc, 1));
	xmn = 0;
	xmx = xxmx;
	if (fpar_.fmin > 0 && fpar_.fmax > fpar_.fmin) {
	    xmn = log10(fpar_.fmin / ucon_.fpma + ucon_.fpmc) / ucon_.fpmb;
	    xmx = log10(fpar_.fmax / ucon_.fpma + ucon_.fpmc) / ucon_.fpmb;
	}
	ucon_.famn = (float) limit(0, xmn, xxmx);
	ucon_.famx = (float) limit(0, xmx, xxmx);
	ucon_.faty = 2;
	ucon_.ffmt = 3;
	dxx = (npm1 > 1) ? xxmx / npm1 : xxmx;
	ucon_.ifsp = (int) (ucon_.famn / dxx + 0.5);
	ucon_.nfsp = (int) (ucon_.famx / dxx + 0.5) - ucon_.ifsp + 1;
	ucon_.fgr = (float) (xxmx / 10);
	e10 = limit(-30, ucon_.famn * ucon_.fpmb, 30);
	ucon_.fmn = (float) (ucon_.fpma * (pow10(e10) - ucon_.fpmc));
	e10 = limit(-30, ucon_.famx * ucon_.fpmb, 30);
	ucon_.fmx = (float) (ucon_.fpma * (pow10(e10) - ucon_.fpmc));

	dx = 1.0 / ucon_.nfsp;
	for (i = 0; i <= ucon_.nfsp; i++)
	    map[i] = (float) (i * dx);
    } else {
	fd = open_sdf(ft);
	if (!fd) {
	    wave_.sampsc = fpar_.fmax * 2;
	}
	setup();
	scnwav();
	if (fpar_.fpma > 0 && fpar_.fpmb > 0) {
	    ucon_.fpma = fpar_.fpma * 1000;
	    ucon_.fpmb = (float) (fpar_.fpmb * 100 / xxmx);
	    ucon_.fpmc = fpar_.fpmc;
	    xmn = 0;
	    xmx = xxmx;
	    if (fpar_.fmin > 0 && fpar_.fmax > fpar_.fmin) {
		xmn = log10(fpar_.fmin / ucon_.fpma + ucon_.fpmc) / ucon_.fpmb;
		xmx = log10(fpar_.fmax / ucon_.fpma + ucon_.fpmc) / ucon_.fpmb;
	    }
	    ucon_.famn = (float) limit(0, xmn, xxmx);
	    ucon_.famx = (float) limit(0, xmx, xxmx);
	    dx = (ucon_.fmx - ucon_.fmn) / ucon_.nfsp;
	    dxx = ucon_.fpmb * xxmx;
	    for (i = 0; i <= ucon_.nfsp; i++) {
		fr = ucon_.fmn + i * dx;
		map[i] = (float) (log10(fr / ucon_.fpma + ucon_.fpmc) / dxx);
	    }
	    ucon_.faty = 2;
	    ucon_.ffmt = 3;
	} else if (ucon_.fnat <= 0) {
	    dx = 1.0 / ucon_.nfsp;
	    for (i = 0; i <= ucon_.nfsp; i++)
		map[i] = (float) (i * dx);
	    ucon_.famn = ucon_.fmn;
	    ucon_.famx = ucon_.fmx;
	    ucon_.faty = 0;
	    ucon_.ffmt = 0;
	} else {
	    dx = (ucon_.fmx - ucon_.fmn) / ucon_.nfsp;
	    for (i = 0; i <= ucon_.nfsp; i++) {
		fr = ucon_.fmn + i * dx;
		map[i] = (float) (log(fr / ucon_.fmn) / ucon_.fnat);
	    }
	    ucon_.famn = 0;
	    ucon_.famx = (float) (1000 * ucon_.fnat / log(2.0));
	    ucon_.faty = 1;
	    i = (int) (ucon_.fmn + 0.5);
	    if (i % 1000 == 0)
		ucon_.ffmt = 0;
	    else if (i % 100 == 0)
		ucon_.ffmt = 1;
	    else if (i % 10 == 0)
		ucon_.ffmt = 2;
	    else
		ucon_.ffmt = 3;
	}
    }
    lpggrmp(map);
    newframe();
    while (ucon_.tmn < ucon_.tstp) {
	if (start_page) {
	    iniframe();
	    inprn = 0.0;
	    start_page = 0;
	}
	lpgidle();		/* keep printer on line */
	if (ft == FT_TSD)
	    process_tsd();
	else
	    process_sdf();
	if (bpar_.newpage) {
	    lpgpage();
	    start_page = 1;
	}
	ucon_.tmn = ucon_.tmx;
	ucon_.tmx = ucon_.tmn + ucon_.tfrm;
	if (ucon_.tmx >= ucon_.tstp && ucon_.tfrm > 0) {
	    xl = fpar_.xlen * (ucon_.tstp - ucon_.tmn) / ucon_.tfrm;
	    ucon_.framelength = (long) (xl * ucon_.adupi_x + 0.5);
	    ucon_.tmx = ucon_.tstp;
	}
    }
    if (ft == FT_TSD) {
	close_tsd();		/* close the TSD file */
    }
    lpgidle();			/* keep printer on line */
    close_sdf();		/* close the file */
}

void
process_sdf()
{
    double  tprn, tfil, dx, wwms;
    int     nt, ww;
    long    et;

    if (ucon_.lftmrgn < 0) {
	(void) fprintf(lgf, "ERROR(%s): x=%.1f and xllc=%.1f\n",
	    wave_.sdfn, inprn, fpar_.xllc);
	leave(0);
    }
    ww = ucon_.winwid;
    if (ww <= 0) {
	(void) fprintf(lgf, "invalid window width %d\n", ww);
	leave(0);
    }
    wwms = ww * 1000.0 / wave_.sampsc;
    if (flags_.verbose) {
	info_sdf();		/* print sdf info */
    }
    if (bpar_.date && inprn == 0.0)
	namedate(bpar_.date);	/* put date on spectrograph */
    initfr(ww);			/* initialize FFT frame buffers */
    axes(1);
    if (bpar_.wrfile) {
        fprintf(lgf, ";%s third-octave bands, window-width=%.1f ms\n", 
            bpar_.toba ? "ANSI": "EQUAL", wwms);
        fprintf(lgf, ";   I1      I2      250         500         1000        2000        4000    \n");
        fprintf(lgf, ";==========================================================================\n");
    }
    dx = 1.0 / ucon_.adupi_x;
    inprn += ucon_.lftmrgn * dx;
    tprn = tfil = 0.0;
    nt = (int) (ucon_.dtprn * wave_.sampsc + 0.5);
    for (et = 0; et < ucon_.framelength; et++) {
	if (atendfr(et) != 0)
	    break;		/* quit if end of frame */
	if (bpar_.prspec)
	    dospec(et);		/* spectrograph */
	if (bpar_.wrfile)
	    dofile(et);		/* write SPL to file */
	if (bpar_.taxis)
	    t_axis(et);
	if (bpar_.grid)
	    dogrid(et);
	if (bpar_.prwave)
	    dowave(et, ww, nt);	/* waveform */
	if (bpar_.prenrg)
	    doenrg(et, ww);	/* energy */
	lpgidle();
	lpgadv(1);
	inprn += dx;
	tprn += ucon_.dtprn;
	nt = (int) ((tprn - tfil) * wave_.sampsc + 0.5);
	if (getwav(nt, ww) == EEOF)	/* get ww samples */
	    break;		/* quit if end of file */
	tfil += ucon_.dtfil * nt;
	lpgidle();
    }
    if (bpar_.taxis)
	t_axis(et);
    axes(1);
    axes(2);
    inprn += tick_.t_ys3 * dx;
}

void
process_tsd()
{
    double  tprn, tfil, dx;
    int     nt, ww;
    long    et;

    if (ucon_.lftmrgn < 0) {
	(void) fprintf(lgf, "ERROR(%s): x=%.1f and xllc=%.1f\n",
	    wave_.sdfn, inprn, fpar_.xllc);
	leave(0);
    }
    ww = ucon_.winwid;
    if (ww <= 0) {
	(void) fprintf(lgf, "invalid window width %d\n", ww);
	leave(0);
    }
    if (flags_.verbose) {
	info_tsd();		/* print tsd info */
    }
    if (bpar_.date && inprn == 0.0)
	namedate(bpar_.date);	/* put date on spectrograph */
    initfr(ww);
    axes(1);
    dx = 1.0 / ucon_.adupi_x;
    inprn += ucon_.lftmrgn * dx;
    tprn = tfil = 0.0;
    nt = (int) (ucon_.dtprn * wave_.sampsc + 0.5);
    for (et = 0; et < ucon_.framelength; et++) {
	if (atendfr(et) != 0)
	    break;		/* quit if end of frame */
	do_tsd(tfil);
	if (bpar_.taxis)
	    t_axis(et);
	if (bpar_.grid)
	    dogrid(et);
	if (bpar_.prwave)
	    dowave(et, ww, nt);	/* waveform */
	if (bpar_.prenrg)
	    do_wnr(et, npm1);
	lpgidle();
	lpgadv(1);
	inprn += dx;
	tprn += ucon_.dtprn;
	nt = (int) ((tprn - tfil) * wave_.sampsc + 0.5);
	if (getwav(nt, ww) == EEOF)	/* get ww samples */
	    break;		/* quit if end of file */
	tfil += ucon_.dtfil * nt;
	lpgidle();
    }
    if (bpar_.taxis)
	t_axis(et);
    axes(1);
    axes(2);
    inprn += tick_.t_ys3 * dx;
}

void
finish_page()
{
    if (inprn > 0)
	lpgpage();
}

/* initfr -initialize the buffers */
void
initfr(ww)
int     ww;
{
    int     i, bl;
    long    nn;

    for (i = 0; i <= ucon_.nfsp; i++)
	px[i] = 0;
    bl = (int) ((bpar_.rmdc) ? wave_.wdc : wave_.wbl);
    for (i = 0; i < ww; i++)
	iy[i] = bl;		/* clear the frame buffer */

    nn = (long) (ucon_.tmn * wave_.sampsc);
    isamp = limit(0, nn, wave_.nsamp) - ww / 2;
    (void) getwav(ww, ww);
    tick_.t_type = 0;
    tick_.t_et = 0;
    tick_.t_time = 0.0;
    grid_.g_et = 0;
    grid_.g_time = 0.0;
}

int
is_nan(double x)
{
    int e;
   
    (void) frexp(x, &e);
    return (e < -1021 || e > 1024);	/* test for NaN */
}
        
void
fscnwav()
{
    double   bb, wavmin, wavmax, sum;
    float   range, aver;
    long    i, is, ns;
    int     j;

    sum = 0;
    wavmin = wavmax = 0;
    is = (long) (ucon_.tmn / tm0);
    if (is < 0)
	is = 0;
    ns = (long) (ucon_.tstp / tm0);
    if (ns > nstp)
	ns = nstp;
    for (i = is; i < ns; i++) {
	bbbbb = rd_tsd_data(i);
        if (is_nan(wnap) || fabs(wnap) > 1e30)
	    wnap = 0;
	if (apmx < wnap)
	    apmx = wnap;
	if (apmn > wnap)
	    apmn = wnap;
	for (j = 1; j < npm1; j++) {
	    bb = bbbbb[j];
	    if (is_nan(bb) || fabs(bb) > 1e30)
		bb = 0;
	    if (i == 0 && j == 0)
		wavmin = wavmax = bb;
	    else if (bb < wavmin)
		wavmin = bb;
	    else if (bb > wavmax)
		wavmax = bb;
	}
    }
    appp = (apmx > apmn) ? (apmx - apmn) : 1;
    ucon_.yen0 = (int) (ucon_.yen1 - ucon_.yen2 * apmn / appp);
    aver = (float) ((nstp > 0) ? sum / nstp / npm1 : 0.0);
    range = (float) (wavmax - wavmin);
    w0_.wbits = 16;
    w0_.wsign = 1;
    w0_.wrn = (range > 0) ? range : 1;
    w0_.wmn = (float) wavmin;
    w0_.wmx = (float) wavmax;
    w0_.wbl = 0;
    w0_.wdc = aver;
    w0_.lstwv = 0;
    if (flags_.debug) {
	(void) fprintf(lgf,
	    "fscnwav: wmn=%.0f wmx=%.0f wrn=%.0f wdc=%.1f nstp=%d\n",
	    w0_.wmn, w0_.wmx, w0_.wrn, w0_.wdc, nstp);
    }
}

/* scnwav - scan all samples */
void
scnwav()
{
    int     bb, wavmin, wavmax;
    double  v, sum, sumsq;
    float   range, aver;
    long    ii, is, ns;

    is = (long) (ucon_.tmn * wave_.sampsc);
    if (is < 0)
	is = 0;
    ns = (long) (ucon_.tstp * wave_.sampsc);
    if (ns > wave_.nsamp)
	ns = wave_.nsamp;
    sum = 0.0;
    sumsq = 0.0;
    if (fetch != NULL) {
        wavmin = wavmax = fetch(is);
        for (ii = is; ii < ns; ii++) {
	    bb = fetch(ii);
	    if (bb < wavmin)
		wavmin = bb;
	    else if (bb > wavmax)
		wavmax = bb;
	    v = bb;
	    sum += v;
	    sumsq += v * v;
	}
    }
    aver = (float) ((ns > is) ? sum / (ns - is) : 0.0);
    meansq = (ns > is) ? (sumsq / (ns - is)) - (aver * aver) : 0;
    meansq *= ucon_.winwid * 2;
    range = ((float) wavmax) - ((float) wavmin);
    if (bpar_.bits >= 8 && bpar_.bits <= 16) {
	wave_.wbits = bpar_.bits;
	wave_.wsign = bpar_.sign;
    } else {
	if (range > 4096 || aver > 4096) {
	    wave_.wbits = 16;
	    wave_.wsign = 1;
	} else if (range > 256 || aver > 256) {
	    wave_.wbits = 12;
	    wave_.wsign = (aver < 1024) ? 1 : 0;
	} else {
	    wave_.wbits = 8;
	    wave_.wsign = (aver < 64) ? 1 : 0;
	}
    }
    if (bpar_.aswv != 0) {
	wave_.wrn = (range > 0) ? range : 1;
	wave_.wmn = (float) wavmin;
	wave_.wmx = (float) wavmax;
    } else {
	wave_.wrn = (float) pow2((double) wave_.wbits);
	wave_.wmn = (float) ((wave_.wsign) ? (-0.5 * wave_.wrn) : 0);
	wave_.wmx = wave_.wmn + wave_.wrn;
    }
    wave_.wbl = (float) ((wave_.wsign) ? 0.0 : pow2(wave_.wbits - 1.0));
    wave_.wdc = aver;
    wave_.lstwv = wave_.wbl;

    if (flags_.debug) {
	(void) fprintf(lgf,
	    "scnwav: wmn=%.0f wmx=%.0f wrn=%.0f wdc=%.1f ns=%ld\n",
	    wave_.wmn, wave_.wmx, wave_.wrn, wave_.wdc, ns);
    }
}

/* getwav - get a frame of samples */
int
getwav(int nt, int ww)
{
    int     i, it, nn;

    if (ww <= 0)
	return (0);
    if (fetch == NULL)
	return (0);
    if (nt <= 0)                    /* frame slip must be positive */
	return (0);
    nn = ww - nt;
    if (nn < 0) {	            /* gap between frames ? */
        for (i = 0; i < -nn; i++) {  /* skip nn samples */
            it = fetch(isamp);
            isamp++;
        }
	nn = 0;
    } else {
	for (i = 0; i < nn; i++)    /* slide frame by nt samples */
	    iy[i] = iy[nt + i];
    }
    for (it = nn; it < ww; it++) {  /* get nt new samples */
	iy[it] = fetch(isamp);  /* new sample at end of frame */
	isamp++;
    }
    return (OK);
}

/* info_sdf - print sdf spectrograph info */
void
info_sdf()
{
    fprintf(lgf, "FFT speech spectrograph           ");
    fprintf(lgf, "speech filename: %s \n", wave_.sdfn);
    fprintf(lgf, "sampling rate = %12.3f kHz        ", wave_.sampsc / 1000);
    fprintf(lgf, "ADC = %d bits (%ssigned)\n",
	wave_.wbits, wave_.wsign ? "" : "un");
    fprintf(lgf, "print rate    = %12.3f in/sec\n", ucon_.inps);
    fprintf(lgf, "file length   = %8ld samples", wave_.nsamp);
    fprintf(lgf, " = %8.3f sec\n", wave_.nsamp / wave_.sampsc);
    fprintf(lgf, "window width  = %8d samples", ucon_.winwid);
    fprintf(lgf, " = %8.3f msec\n", (ucon_.winwid / wave_.sampsc) * 1000);
    fprintf(lgf, "FFT width     = %8d samples", ucon_.nfft);
    fprintf(lgf, " = %8.3f msec\n", (ucon_.nfft / wave_.sampsc) * 1000);
    fprintf(lgf, "frequency range from %.0f to %.0f Hz\n",
	ucon_.fmn, ucon_.fmx);
}

/* info_tsd - print tsd spectrograph info */
void
info_tsd()
{
    fprintf(lgf, "Cochlear spectrograph           ");
    fprintf(lgf, "stimulus file: %s \n", wave_.sdfn);
    fprintf(lgf, "npts=%d, nstp=%d\n", npts, nstp);
    fprintf(lgf, "frequency-to-place map:  a=%6.4f b=%6.4f c=%6.4f\n",
	pa_.fpma, pa_.fpmb, pa_.fpmc);
    fprintf(lgf, "sampling rate = %12.3f kHz        ", wave_.sampsc / 1000);
    fprintf(lgf, "ADC = %d bits (%ssigned)\n",
	wave_.wbits, wave_.wsign ? "" : "un");
    fprintf(lgf, "print rate    = %12.3f in/sec\n", ucon_.inps);
    fprintf(lgf, "window width  = %8d samples", ucon_.winwid);
    fprintf(lgf, " = %8.3f msec\n", (ucon_.winwid / wave_.sampsc) * 1000);
    fprintf(lgf, "FFT width     = %8d samples", ucon_.nfft);
    fprintf(lgf, " = %8.3f msec\n", (ucon_.nfft / wave_.sampsc) * 1000);
    fprintf(lgf, "frequency range from %.0f to %.0f Hz\n",
	ucon_.fmn, ucon_.fmx);
}

/* namedate - print file name and current date on spectrogram */
void
namedate(int i)
{
    lpgchup(0.0, 1.0);
    lpgchlw(fpar_.lablwt);
    lpgchsz(fpar_.labsiz);
    lpgchxt(0);
    if (i > 1) {
        lpgchal(1, 3);
        lpgtext(0.1, 0.1, file_name);
    }
    lpgchal(3, 3);
    lpgtext(10.4, 0.1, time_date);
}

void
setup_lpg()
{
    if (!setup_lpg_called) {
	lpgini();
	ucon_.adupi_x = lpg_x(1.0);
        ucon_.adupi_y = lpg_y(1.0);
	setup_lpg_called++;
    }
}


void
setup()
{
    int     i, j, ww, minskp, ftpi;
    double  f, d, o, npw;
    float   df, nyqf, twave, smpcy;
    unsigned int nn, mm, nm;

    if (iy == NULL) {
        iy = malloc(NF * sizeof(int));
        y = malloc(NY * sizeof(float));
        w = malloc(NF * sizeof(float));
        if (w == NULL) {
            fprintf(stderr, "memory allocation error!\n");
            exit(1);
        }
    }
    strcpy(file_name, wave_.sdfn);
    setup_lpg();
    if (fpar_.xllc > 32000.0 / ucon_.adupi_x) {
	ucon_.lftmrgn = 32000;
	ucon_.strtpos = 32000;
    } else {
	ucon_.lftmrgn = (int) ((fpar_.xllc - inprn) * ucon_.adupi_x + 0.5);
	ucon_.strtpos = (int) (fpar_.xllc * ucon_.adupi_x + 0.5);
    }
    ucon_.ysp1 = (int) (fpar_.yllc * ucon_.adupi_y + 0.5);
    ucon_.ysp2 = (int) (fpar_.ylen * ucon_.adupi_y + 0.5);
    nn = ucon_.ysp2 * sizeof(char);
    ucon_.ywv1 = (int) (fpar_.wllc * ucon_.adupi_y + 0.5);
    ucon_.ywv2 = (int) (fpar_.wlen * ucon_.adupi_y + 0.5);
    ucon_.yen1 = (int) (fpar_.ellc * ucon_.adupi_y + 0.5);
    ucon_.yen2 = (int) (fpar_.elen * ucon_.adupi_y + 0.5);
    tick_.t_xs1 = (int) (fpar_.ticsiz * ucon_.adupi_y + 0.5);
    tick_.t_ys1 = (int) (fpar_.ticsiz * ucon_.adupi_x + 0.5);
    tick_.t_xs2 = (int) (fpar_.ticsiz * ucon_.adupi_y * 0.75 + 0.5);
    tick_.t_ys2 = (int) (fpar_.ticsiz * ucon_.adupi_x * 0.75 + 0.5);
    tick_.t_xs3 = (int) (fpar_.ticsiz * ucon_.adupi_y * 0.5 + 0.5);
    tick_.t_ys3 = (int) (fpar_.ticsiz * ucon_.adupi_x * 0.5 + 0.5);
    ucon_.pagelength = (long) (fpar_.xpage * ucon_.adupi_x + 0.5);
    ucon_.pagewidth = (int) (fpar_.ypage * ucon_.adupi_y + 0.5);
    ucon_.framelength = (long) (fpar_.xlen * ucon_.adupi_x + 0.5);


    ucon_.fmn = (fpar_.fmin < 0) ? 0 : fpar_.fmin;
    ucon_.fmx = fpar_.fmax;
    ucon_.fgr = fpar_.fgrid;
    if (bpar_.nflim) {
	nyqf = wave_.sampsc / 2;
	ucon_.fmn = (float) limit(0.0, ucon_.fmn, nyqf);
	ucon_.fmx = (float) limit(0.0, ucon_.fmx, nyqf);
    }
    if (ucon_.fmn == ucon_.fmx) {
	ucon_.fmn = 0;
	ucon_.fmx = wave_.sampsc / 2;
    } else if (ucon_.fmn > ucon_.fmx) {
	f = ucon_.fmn;
	ucon_.fmn = ucon_.fmx;
	ucon_.fmx = (float) f;
    }
    if (fpar_.foct <= 0) {
	ucon_.fnat = 0;
    } else {
	o = pow2(fpar_.foct);
	ucon_.fmn = (float) ((ucon_.fmn <= 0) ? ucon_.fmx / o : ucon_.fmn);
	ucon_.fmx = (float) ((ucon_.fmx <= 0) ? ucon_.fmn * o : ucon_.fmx);
	ucon_.fnat = (float) log(ucon_.fmx / ucon_.fmn);
    }
    twave = (wave_.sampsc > 0) ? wave_.nsamp / wave_.sampsc : 0;
    ucon_.tmn = limit(0, fpar_.tmin, twave);
    ucon_.tmx = fpar_.tmax;
    if (ucon_.tmx <= 0.0)
	ucon_.tmx = twave;
    if (ucon_.tmn >= ucon_.tmx) {
	(void) fprintf(stderr, "invalid time range\n");
	leave(1);
    }
    ucon_.tstp = ucon_.tmx;
    if (fpar_.tpage <= 0.0 || bpar_.newpage == 0) {
	ucon_.tfrm = 0;
    } else {
	ucon_.tfrm = fpar_.tpage;
	ucon_.tmx = limit(ucon_.tmn, ucon_.tmn + fpar_.tpage, ucon_.tstp);
    }
    ucon_.inps = (ucon_.tmx > ucon_.tmn) ? fpar_.xlen / (ucon_.tmx - ucon_.tmn) : 1;
    ucon_.tom = (int) log10(ucon_.inps);
    tick_.t_ti = pow(0.1, (double) (1 + ucon_.tom));
    if (flags_.debug)
	(void) fprintf(stderr, "tom=%d, t_ti=%.3f\n", ucon_.tom, tick_.t_ti);

    if (bpar_.window == 0) {
	ucon_.lobwid = 1.0;
    } else if (bpar_.window == 1) {
	ucon_.lobwid = 2.0;
    } else if (bpar_.window == 2) {
	ucon_.lobwid = 3.0;
    } else {
	ucon_.lobwid = 1.0;
    }
    if (fpar_.bandwidth <= 0.0)
	smpcy = NF / 2;
    else if (fpar_.windwidth <= 0.0)
	smpcy = ucon_.lobwid * wave_.sampsc / fpar_.bandwidth;
    else
	smpcy = fpar_.windwidth * wave_.sampsc;
    ucon_.winwid = (int) limit(0.0, smpcy, NF);
    if ((d = fpar_.eeps / ucon_.inps) < 1 || d > 100)
	fpar_.eeps = 10 * ucon_.inps;
    ww = ucon_.winwid;
    d = tpi / (ww - 1);
    if (bpar_.window == 0) {
	for (i = 0; i < ww; i++) {
	    w[i] = 1.0;		/* rectangular */
	}
    } else if (bpar_.window == 1) {
	for (i = 0; i < ww; i++) {
	    j = i - ww / 2;
	    w[i] = (float) (0.54 + 0.46 * cos(j * d));	/* Hamming */
	}
    } else if (bpar_.window == 2) {
	for (i = 0; i < ww; i++) {
	    j = i - ww / 2;
	    w[i] = (float) (0.42 + 0.5 * cos(j * d));
	    w[i] += (float) (0.08 * cos(2 * j * d));	/* Blackman */
	}
    }
    bpar_.nlpc = limit(0, bpar_.nlpc, MAXLPC);
    if (bpar_.mfft > 0) {
	nn = 1;
	nm = bpar_.mfft;
	for (mm = 0; mm < nm; mm++)
	    nn *= 2;
	ucon_.nfft = limit(8, nn, NF);
    } else {
	nm = ucon_.winwid;
	for (nn = 1; nn < nm; nn *= 2)
	    continue;
	ucon_.nfft = limit(8, 2 * nn, NF);
    }
    if (ucon_.winwid > ucon_.nfft) {
	(void) fprintf(stderr, "window width is greater than FFT width\n");
	leave(1);
    }
    ucon_.nfrq = ucon_.nfft / 2 + 1;
    df = wave_.sampsc / ucon_.nfft;
    ucon_.ifsp = (int) (ucon_.fmn / df + 0.5);
    ucon_.nfsp = (int) ((ucon_.fmx - ucon_.fmn) / df + 0.5);
    ucon_.dtprn = (float) (1.0 / (ucon_.adupi_x * ucon_.inps));
    ucon_.dtfil = (float) (1.0 / wave_.sampsc);
    d = ucon_.dtprn * wave_.sampsc * bpar_.ftpw;
    ftpi = (bpar_.ftpi < 1) ? 1 : bpar_.ftpi;
    minskp = (ucon_.adupi_x + ftpi / 2) / ftpi;
    npw = ucon_.winwid / minskp;
    ucon_.fftskp = (int) ((d <= 0 || d > npw) ? minskp : ucon_.winwid / d);
    if (flags_.debug) {
	(void) fprintf(stderr, "fftskp=%d\n", ucon_.fftskp);
    }
    grid_.g_incr = 0.1;
    while (grid_.g_incr > ucon_.dtprn * ucon_.adupi_x * 0.5)
	grid_.g_incr *= 0.1;
    while (grid_.g_incr > ucon_.dtprn * ucon_.adupi_x * 0.05)
	grid_.g_incr *= 0.5;
}

void
dowave(et, ww, nt)
long    et;
int     nt, ww;
{
    int     i1, i2, it, iymn, iydf, iyit;
    float   ymn, ymx;
    long    x0, y0, xw, yh;

    i1 = (ww - nt) / 2;
    if (i1 < 0)
	i1 = 0;
    i2 = i1 + nt;
    if (i2 > ww)
	i2 = ww;
    ymn = ymx = wave_.lstwv;
    for (it = i1; it < i2; it++) {	/* find min and max */
	iyit = iy[it];
	if (ymn > iyit)
	    ymn = (float) iyit;
	if (ymx < iyit)
	    ymx = (float) iyit;
	wave_.lstwv = (float) iyit;
    }
    if (bpar_.invwav == 0)
	iymn = (int) (ucon_.ywv2 * (ymn - wave_.wmn) / wave_.wrn);
    else
	iymn = (int) (ucon_.ywv2 * (wave_.wmx - ymx) / wave_.wrn);
    iydf = (int) (ucon_.ywv2 * (ymx - ymn) / wave_.wrn);
    x0 = ucon_.strtpos + et;
    y0 = ucon_.ywv1 + iymn;
    xw = 1;
    yh = 1 + iydf;
    lpgrule(x0, y0, xw, yh);	/* waveform */
}

void
do_tsd(t)
double  t;
{
    double  a, bb;
    float   mn, rn;
    float   c, d, g;
    int     i, j, k;
    long    n;
    static long lastn = -1;

    mn = w0_.wmn;
    rn = w0_.wmx - mn;

    n = (long) ((t + ucon_.tmn) / tm0);
    c = fpar_.contrast;
    d = fpar_.darkness;
    g = fpar_.floor;
    j = (npm1 - 1) - ucon_.ifsp;
    k = ucon_.nfsp;
    if (n >= nstp) {
	for (i = 0; i <= k; i++)
	    px[i] = 0;
    } else if (n != lastn) {
	lastn = n;
	bbbbb = rd_tsd_data(n);
	for (i = 0; i < k; i++) {
	    bb = bbbbb[j - i];
	    if (is_nan(bb) || fabs(bb) > 1e10)
		bb = 0;
	    a = (bb - mn) / rn;
	    if (a == 0)
		a = 1e-38;
	    a = 1 + log(a) / log(g);
	    a = (a - 0.25) * 1.8 * c + d;
	    a = lpgcomp(a) * 255;
	    if (a < 0)
		px[i] = 0;
	    else if (a > 255.)
		px[i] = 255;
	    else
		px[i] = (unsigned char) a;
	}
    }
    if (bpar_.prspec)
	lpggray(px, ucon_.nfsp, ucon_.ysp1, ucon_.ysp2, 1);
}

void
dospec(et)
long    et;
{
    int w;

    if ((et % ucon_.fftskp) == 0) {
	if (bpar_.nlpc > 0)
	    lpfft();
	else
	    dofft();
	incomp();
	w = ucon_.fftskp;
	if (w > ucon_.framelength - et)
	    w = ucon_.framelength - et;
        lpggray(px, ucon_.nfsp, ucon_.ysp1, ucon_.ysp2, w);
    }
}

void
dofft()
{
    int     i, j, ww;
    double  yr, yi, bl;

    ww = ucon_.winwid;
    if (bpar_.hfemph != 0) {
	y[0] = (float) iy[0];
	for (i = 1; i < ww; i++)/* emphasize high freq. */
	    y[i] = ((float) iy[i]) - ((float) iy[i - 1]);
    } else {
	bl = (bpar_.rmdc) ? wave_.wdc : wave_.wbl;
	for (i = 0; i < ww; i++)/* copy to float array */
	    y[i] = (float) (iy[i] - bl);
    }
    for (i = 0; i < ww; i++)	/* window */
	y[i] = w[i] * y[i];
    for (i = ww; i < ucon_.nfft; i++)	/* zero remainder */
	y[i] = 0.0;
    (void) rfft(y, ucon_.nfft);
    rfft_unpack(y, ucon_.nfft);

    for (i = 0; i < ucon_.nfrq; i++) {	/* complex magnitude squared */
	j = 2 * i;
	yr = y[j];
	yi = y[j + 1];
	y[i] = (float) (yr * yr + yi * yi);
	if (y[i] > __max_mag) {
	    __max_mag = y[i];
	}
    }
    for (i = ucon_.nfrq; i < ucon_.nfft; i++)	/* zero remainder */
	y[i] = 0.0;
}

/*
 * incomp - intensity compression
 */
void
incomp()
{
    int     i;
    double  yy, c, d, g, scl, sumsq;

    sumsq = 0.0;
    for (i = 0; i < ucon_.nfsp; i++)
	sumsq += y[i + ucon_.ifsp];
    sumsq /= ucon_.nfsp;
    sumsq *= 3.0;
    if (meansq == 0 || sumsq == 0)
	scl = 100;
    else
	scl = pow(meansq / sumsq, fpar_.agcsp) / meansq;
    c = fpar_.contrast;
    d = fpar_.darkness;
    g = fpar_.floor;
    for (i = 0; i < ucon_.nfsp; i++) {
	yy = y[i + ucon_.ifsp] * scl;
	if (ff_mx < yy)
	    ff_mx = (float) yy;
	yy = (yy > 0) ? 1 + log(yy) / log(g) : -1;
	yy = (yy - 0.3) * 1.5 * c + d;
	yy = lpgcomp(yy) * 255;
	if (yy > 255) {
	    px[i] = 255;
	} else if (yy < 0) {
	    px[i] = 0;
	} else {
	    px[i] = (unsigned char) yy;
	}
	id[px[i]]++;
    }
}

void
dofile(et)
long    et;
{
    int ww;

    ww = ucon_.winwid;
    if ((et % ucon_.fftskp) == 0) {
        fprintf(lgf, "%7ld %7ld", isamp - ww, isamp);
        dospl();	
    }
}

double
dbe(double e)
{
    return ((e < 1e-40) ? -400 : 10 * log10(e));
}

void
dospl()
{
    int     i, j, ww, ib = 0;
    double  yr, yi, bl, ys = 0;
    double  df, f, *fb1, *fb2;
    static int nb = 15;
    static double fbe_equal[16] = {
        177, 223, 281, 354, 445, 561, 707, 891, 1122,
        1414, 1782, 2245, 2828, 3564, 4490, 5657
    };
    static double fbe_ansi[16] = {
        180, 224, 280, 355, 450, 560, 710, 900, 1120,
        1400, 1800, 2240, 2800, 3550, 4500, 5600
    };

    if (bpar_.toba) {
        fb1 = fbe_ansi;
        fb2 = fbe_ansi + 1;
    } else {
        fb1 = fbe_equal;
        fb2 = fbe_equal + 1;
    }
    ww = ucon_.winwid;
    bl = wave_.wbl;
    for (i = 0; i < ww; i++) {
        y[i] = (float) (iy[i] - bl);    /* copy to float array */
	y[i] *= w[i];	                /* window */
    }
    for (i = ww; i < ucon_.nfft; i++)	/* zero remainder of array */
	y[i] = 0.0;

    (void) rfft(y, ucon_.nfft);         /* real FFT (in place) */

    df = wave_.sampsc / ucon_.nfft;
    for (i = 1; i < ucon_.nfrq; i++) {	/* complex magnitude squared */
        f = i * df;
	j = 2 * i;
	yr = y[j];
	yi = y[j + 1];
	y[i] = (float) (yr * yr + yi * yi);
        if (ib < nb) {
            if (f >= fb1[ib]) {
	        if (f < fb2[ib]) {
	            ys += y[i];
	        } else {
                    fprintf(lgf, "%4.0f", dbe(ys));
	            ib++;
	            ys = y[i];
	        }
	    }
	}
    }
    if (f >= fb1[0] && f < fb1[nb])
        fprintf(lgf, "%4.0f", dbe(ys));
    fprintf(lgf, "\n");
}

void
do_wnr(et, ww)
int     ww;
long    et;
{
    int     iydb;
    long    x0, y0, xw, yh;
    static int lsty = 0;

    if (et == 0)
	lsty = 0;
    if (is_nan(wnap) || fabs(wnap) > 1e30)
        wnap = 0;
    iydb = (int) (ucon_.yen2 * (wnap - apmn) / appp);
    x0 = ucon_.strtpos + et;
    xw = 1;
    if (iydb > lsty) {
	y0 = ucon_.yen1 + lsty;
	yh = iydb - lsty + 1;
    } else {
	y0 = ucon_.yen1 + iydb;
	yh = lsty - iydb + 1;
    }
    lpgrule(x0, y0, xw, yh);	/* wnr */
    lsty = iydb;
}

void
doenrg(et, ww)
int     ww;
long    et;
{
    int     i, iydb;
    double  d, dbnrg, yy, sumsq, nrg;
    long    x0, y0, xw, yh;
    static double e_time, e_dt, maxsq;
    static long e_et;

    if (et == 0) {
	e_dt = (fpar_.eeps > 0.0) ? 1.0 / fpar_.eeps : 0.01;
	e_time = 0.0;
	e_et = 0;
	d = pow2((double) wave_.wbits - 1.0);
	maxsq = d * d * ww;
    }
    if (et < e_et) {
	d = fpar_.emin;
    } else {
	sumsq = 0.0;
	for (i = 0; i < ww; i++) {
	    yy = iy[i] - wave_.wbl;
	    sumsq += yy * yy;
	}
	nrg = sumsq / maxsq;
	if (nrg < 1e-9) {
	    d = fpar_.emin;
	} else {
	    dbnrg = 10.0 * log10(nrg);
	    d = limit(fpar_.emin, dbnrg, fpar_.emax);
	}
	e_time += e_dt;
	e_et = (long) (e_time / ucon_.dtprn + 0.5);
    }
    iydb = (int) (ucon_.yen2 * (d - fpar_.emin) / (fpar_.emax - fpar_.emin));
    xw = 1;
    yh = 1 + iydb;
    x0 = ucon_.strtpos + et;
    y0 = ucon_.yen1;
    lpgrule(x0, y0, xw, yh);	/* waveform */
}

int
atendfr(long et)
{
#ifdef INFO
    endframe(__max_mag);
#endif /* INFO */
#ifdef MSDOS

    if (_kbhit()) {
        int c;
	c = _getch();		/* get a character from the keyboard */
	if (c == ' ')		/* if character is a space, then quit */
	    leave(0);
	else if (c == 't' && printer > 0)
	    fprintf(lgf, "................ time = %9.3f\n", et * ucon_.dtprn);
    }
#endif /* MSDOS */
    return (0);
}

void
dohist()
{
    int     i, j;
    float   sum1, sum2, per;

    sum1 = 0.0;
    for (i = 0; i < 256; i++)
	sum1 += id[i];
    if (sum1 > 0.0) {
	for (i = 0; i < 16; i++) {
	    sum2 = 0.0;
	    for (j = 0; j < 16; j++)
		sum2 += id[i * 16 + j];
	    per = 100 * sum2 / sum1;
	    fprintf(lgf, "%4.0f%c", per, (i % 16 == 15) ? '\n' : ' ');
	}
    }
}

void
lpfft()
{
    int     i, j, nc, ww, status;
    double  yr, yi, bl, as, dn;
    float   alpha[1], a[LPCSIZ], r[LPCSIZ], rc[LPCSIZ];

    ww = ucon_.winwid;
    nc = bpar_.nlpc;
    if (bpar_.hfemph != 0) {
	y[0] = (float) iy[0];
	for (i = 1; i < ww; i++)/* emphasize high freq. */
	    y[i] = ((float) iy[i]) - ((float) iy[i - 1]);
    } else {
	bl = (bpar_.rmdc) ? wave_.wdc : wave_.wbl;
	for (i = 0; i < ww; i++)/* copy to float array */
	    y[i] = (float) (iy[i] - bl);
    }
    status = autoc(ww, y, nc, a, alpha, rc, r);
    if (status != 0) {
	nc = 0;
	a[0] = 1.0;
	alpha[0] = 0.0;
    }
    for (i = 0; i <= nc; i++)
	y[i] = a[i];
    for (i = nc + 1; i < ucon_.nfft; i++)	/* zero remainder */
	y[i] = 0.0;
    (void) rfft(y, ucon_.nfft);
    rfft_unpack(y, ucon_.nfft);

    as = alpha[0];
    for (i = 0; i < ucon_.nfrq; i++) {	/* complex magnitude squared */
	j = 2 * i;
	yr = y[j];
	yi = y[j + 1];
	dn = (yr * yr + yi * yi);
	y[i] = (float) ((dn > 0.0) ? as / dn : as);
    }
    for (i = ucon_.nfrq; i < ucon_.nfft; i++)	/* zero remainder */
	y[i] = 0.0;

}

/* autoc - autocorrelation method of linear prediction analysis */

int
autoc(int n, float *x, int m, float *a, float *alpha, float *rc, float *r)
{
/*
 *   inputs:   n        - no. of data points
 *             x[n]     - input data sequence
 *             m        - order of filter
 *
 *  outputs:   a[m+1]   - filter coefficients
 *             alpha[1] - residual "energy"
 *             rc[m]    - reflection coefficients
 *             r[m+1]   - autocorrelation coefficients
 *
 */
    int     ib, ip, k;
    int     mh, minc, mp, nk, np;
    float   at, s;

    a[0] = 1.0;			/* first coefficient is always 1 */
    mp = m + 1;
    for (k = 0; k < mp; k++) {
	r[k] = 0.0;
	nk = n - k;
	for (np = 0; np < nk; np++) {
	    r[k] += x[np] * x[np + k];
	}
    }
    if (r[0] <= 0.0)		/* check for zero energy in signal */
	return (1);
    rc[0] = -r[1] / r[0];
    a[1] = rc[0];
    alpha[0] = r[0] + r[1] * rc[0];
    for (minc = 2; minc <= m; minc++) {
	s = 0.0;
	for (ip = 0; ip < minc; ip++) {
	    s += r[minc - ip] * a[ip];
	}
	rc[minc - 1] = -s / alpha[0];
	mh = minc / 2 + 1;
	for (ip = 1; ip < mh; ip++) {
	    ib = minc - ip;
	    at = a[ip] + rc[minc - 1] * a[ib];
	    a[ib] += rc[minc - 1] * a[ip];
	    a[ip] = at;
	}
	a[minc] = rc[minc - 1];
	alpha[0] += rc[minc - 1] * s;
	if (alpha[0] <= 0.0)
	    return (1);
    }
    return (0);
}

void
rfft_unpack(ww, n)
float  *ww;
int     n;
{
    ww[n] = ww[1];
    ww[1] = 0.0;
    ww[n + 1] = 0.0;
}

double
spcval(is, p)
long is;
double p;
{
    int ip1, ip2, ww;
    double val;
    
    ip1 = (int) p;
    ip2 = ip1 + 1;
    
    ww = ucon_.winwid;
    isamp = limit(0, is, wave_.nsamp) - ww / 2;
    (void) getwav(ww, ww);
    dofft();
    val = y[ip1] + (y[ip2] - y[ip1]) * (p - ip1);
    return (val);
}
