/* ps.c */

#include <stdio.h>
#include "version.h"

unsigned int strlen();
void    lpr();
void    lprfl();
void    lprs();

extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) (), (*pr_gray) ();
extern char file_name[80];
extern char time_date[40];
extern int adupi_x;
extern int adupi_y;
extern int pglen;
extern int pgwid;
extern int p_color;
extern int bpp;
extern int p_gmode;
extern double dotdiam;

static double unmap(double, float *, int);
static void ps_idle(), ps_page(), ps_out(), ps_end();
static void ps_rule(), ps_line(), ps_gray();
static void ps_header();

static int psdpi = 0;
static long xpos = 0;
static long xprn = 0;
static long yprn = 0;
static int firstprint = 1;

/******************************************************************/

void
ps_setup()
{
    psdpi = 300;
    adupi_x = psdpi;
    adupi_y = psdpi;
    pglen = (int) (psdpi * 10.5);
    pgwid = (int) (psdpi * 8.0);
    bpp = 1;
    dotdiam = (float) (1.0 / psdpi);

    pr_idle = ps_idle;
    pr_page = ps_page;
    pr_out = NULL;
    pr_end = ps_end;
    pr_rule = ps_rule;
    pr_line = ps_line;
    pr_gray = ps_gray;
}

void
ps_out(unsigned char *buf, int n)
{
    char    str[80];
    int     i, y, xtrn, ytrn;

    if (firstprint) {
	ps_header();
	firstprint = 0;
    }
    if (n > 0) {
	for (y = 0; buf[y] == 0; y++)
	    continue;
	xtrn = xpos - xprn;
	ytrn = y - yprn;
	sprintf(str, "%d %d %d VS\n", n - y, xtrn, ytrn);
	lprs(str);
	for (i = y; i < n; i++) {
	    sprintf(str, "%.2x", buf[i] & 0xFF);
	    lprs(str);
	    if (((i - y) % 32) == 31 || i == (n - 1))
		lprs("\n");
	}
	xprn = xpos;
	yprn = y;
    }
    xpos += 8;
}

void
ps_end(int prompt)
{
    prompt = prompt;
    if (xpos > 0) {
	lprs("showpage\n");
    }
    lprs("%%EOF\n");
    lprfl();
}

void
ps_idle()
{
}

void
ps_page()
{
    if (xpos > 0) {
	lprs("showpage\n");
    }
    xpos = 0;
    xprn = 0;
    yprn = 0;
}

void
ps_rule(long xllc, long yllc, long xlen, long ylen)
{
    char    str[80];

    if (firstprint) {
	ps_header();
	firstprint = 0;
    }
    sprintf(str, "%ld 0 ", -xlen);
    lprs(str);
    sprintf(str, "0 %ld ", ylen);
    lprs(str);
    sprintf(str, "%ld 0 ", xlen);
    lprs(str);
    sprintf(str, "%ld %ld ", xllc - xprn, yllc - yprn);
    lprs(str);
    lprs("R\n");
}

void
ps_line(int n, long *x, long *y, double w)
{
    int     i, nc;
    char    str[80];

    if (n < 2)
	return;
    if (firstprint) {
	ps_header();
	firstprint = 0;
    }
    sprintf(str, "%ld %ld M ", x[0] - xprn, y[0] - yprn);
    nc = strlen(str);
    lprs(str);
    for (i = 1; i < n; i++) {
	sprintf(str, "%ld %ld L ", x[i] - x[i - 1], y[i] - y[i - 1]);
	nc += strlen(str);
	if (nc > 64) {
	    lprs("\n");
	    nc = strlen(str);
	}
	lprs(str);
    }
    sprintf(str, "%.0f S\n", w * psdpi);
    lprs(str);
}

void
ps_gray(long x, long y, long w, long h, long nx, unsigned char *px, float *mp)
{
    char s[256];
    double a0, a1, dj, p1, p2, pm, a, b;
    int i, j, n, g, rd, gr, bl;
    static int lastn = 0;
    static double rpi = 100;	// rectangles per inch

    n = (int) ((h / adupi_y) * rpi + 0.5);
    if (n < nx)
	n = nx;
    if (n > h / 2)
	n = h;
    a = w;
    b = (double) h / n;
    if (lastn != n) {
        sprintf(s, "/gs_str %d string def\n", n);
        lprs(s);
	lastn = n;
    }
    sprintf(s, "%d %d 8 ", 1, n);	// dimensions of source image
    lprs(s);
    sprintf(s, "[%.4f 0 0 %.4f %.1f %.1f] %s\n",
	1 / a, 1 / b, -x / a, -y / b,	// map unit square to source
	p_color ? "CS" : "GS");		// color or grayscale ?
    lprs(s);
    for (i = 0; i < n; i++) {
	// interpolate i
	dj = unmap((double) (i) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	p1 = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 0.5
	dj = unmap((double) (i + 1) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	pm = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 1
	dj = unmap((double) (i + 0.5) / (n - 1), mp, nx);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	p2 = a0 * px[j] + a1 * px[j + 1];
	g = (int) ((p1 + p2 + pm) / 3 + 0.5);
	if (p_color) {
	    if (g < 128) {
	        rd = 0;
		bl = 255 - g * 2;
	    } else {
	        rd = (g - 128) * 2;
		bl = 0;
	    }
	    gr = 255 - rd - bl;
	    sprintf(s, "%02X%02X%02X%s", rd, gr, bl,
		((i == (n - 1)) || ((i % 12) == 11)) ? "\n" : "");
	} else {
	    sprintf(s, "%02X%s", 255 - g, 
		((i == (n - 1)) || ((i % 32) == 31)) ? "\n" : "");
	}
	lprs(s);
    }
}

/******************************************************************/

static double
unmap(double u, float *mp, int nx)
{
    int i;

    for (i = 1; i <= nx; i++) {
	if (u < mp[i]) {
	    u = (i + (u - mp[i - 1]) / (mp[i] - mp[i - 1])) / nx;
	    break;
	}
    }
    return (u * (nx - 0.5));
}

static void
ps_header()
{
    char    str[80];

    lprs("%!PS-Adobe-2.0 EPSF-2.0\n");
    sprintf(str, "%%%%Title: %s\n", file_name);
    lprs(str);
    sprintf(str, "%%Creator: %.20s\n", VER);
    lprs(str);
    sprintf(str, "%%%%CreationDate: %s", time_date);
    lprs(str);
    lprs("%%BoundingBox: 18 18 594 774\n");
    lprs("%%Pages: 1\n");
    lprs("%%EndComments\n");
    lprs("/vs_str 1 string def\n");
    lprs("/VS {\n");
    lprs("    translate 8 exch true [1 0 0 1 0 0]\n");
    lprs("    {currentfile vs_str readhexstring pop}\n");
    lprs("    imagemask\n");
    lprs("} def\n");
    lprs("/GS {\n");
    lprs("    {currentfile gs_str readhexstring pop}\n");
    lprs("    false 1 colorimage\n");
    lprs("} def\n");
    lprs("/CS {\n");
    lprs("    {currentfile gs_str readhexstring pop}\n");
    lprs("    false 3 colorimage\n");
    lprs("} def\n");
    lprs("/M /moveto load def\n");
    lprs("/L /rlineto load def\n");
    lprs("/R {moveto rlineto rlineto rlineto fill} bind def\n");
    lprs("/S {setlinewidth stroke} bind def\n");
    lprs(str);
    lprs("%%EndPrologue\n");
    lprs("%%BeginSetup\n");
    lprs("18 774 translate 270 rotate % landscape\n");
    sprintf(str, "%.6f %.6f scale\n", 72.0 / psdpi, 72.0 / psdpi);
    lprs(str);
    lprs("0 setgray ");		/* set pen color to black */
    lprs("1 setlinejoin ");	/* set line joints: rounded join */
    lprs("1 setlinecap\n");	/* set line end: rounded ends */
    lprs("%%EndSetup\n");
    lprs("%%Page: 1 1\n");
}
