/* sdf.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h> 
#include "specto.h"

#ifdef WIN32
#include <io.h>
#define OPEN_MODE   _O_RDONLY|_O_BINARY
#define BUF_TYPE    void *
#else
#include <unistd.h>
#define _stat       stat
#define _fstat      fstat
#define _access     access
#define _open       open
#define _close      close
#define _read       read
#define _lseek      lseek
#define _swab       swab
#define OPEN_MODE   O_RDONLY
#define BUF_TYPE    char *
#endif
#define MALLOC(n,s) malloc((n)*(s))
#define FREE(p)     free(p)
typedef short *SIGPTR;

#define DTYP_UNKNOWN            0
#define DTYP_LINEAR_8		1
#define DTYP_MULAW_8		2
#define DTYP_UNSIGNED_8		3
#define DTYP_UNSIGNED_12        4
#define DTYP_LINEAR_16		5
#define DTYP_LINEAR_32		6
#define DTYP_FLOAT		7
#define DTYP_DOUBLE		8

#define BS          256		/* block size (# of samples) */
#define ILS_MAGIC   -32000	/* ILS sampled data file */
#define TOK_MAGIC   11111	/* BTNI token file */

int     (*fetch) ();
int     is_tsd();
short   ulaw2linear();
void    swab();

extern FILE *lgf;
extern int sdf_type;
extern int cpu_lsb;
extern int noheader;

typedef struct {
    long int key;	/* must be equal to snd_key */
    long int ofst;	/* Offset to raw data */
    long int size;	/* Number of bytes of data */
    long int code;	/* Data format code */
    long int rate;	/* Sample rate */
    long int chan;	/* Number of channels */
    char info[4];	/* Information about the sound. */
}   SND_fmt;
typedef struct {
    long key[2];
    long npts;
    long cntr;
    unsigned short chan;
    unsigned short bits;
    unsigned short code;
    float rate;
    long peak;
    char fill[466];
    float gain1;
    float gain2;
    float vrms;
    float dbref;
}   ADF_fmt;

static double scale_mat();
static int is_ils();
static int is_adf();
static int is_snd();
static int is_mat();
static int is_wav();
static int is_ascii();
static int open_snd();
static int open_lst();
static int open_mat();
static int open_wav();
static int open_noh();
static int open_ils();
static int open_adf();
static int fetch_list();
static int fetch_data();
static int rdhdr_ils();
static int rdhdr_snd();
static int valid_data();
static int wxchnb();
static long lxchnb();
static void exch_hdr();
static void dswapb();
static void lswapb();
static void wswapb();
static void setsiz();

static char *adf_key = "CSRE40";
static double sdf_scal = 0;
static int sdf_bps[] = { 0, 1, 1, 1, 2, 2, 4, 4, 8, };
static int bf = 0;
static int ib = 0;
static int hs = 0;
static int ts = 0;
static int oflg = (int) OPEN_MODE;
static int sdf_dtyp = 0;
static int sdf_bsiz = 0;
static int sdf_bswp = 0;
static int sdf_sblk = 0;
static int tok = 0;
static int itok = 0;
static int ntok = 0;
static int toksiz = 0;
static long sdf_ofst = 0;
static long snd_key = 0x2E736E64L;
static short buff[BS] = {0};
static SIGPTR lstbuf;
static SND_fmt *sound;

void
guess_sdf_type(fn)
char   *fn;
{
    if (is_ils(fn))
        sdf_type = (tok) ? FT_TOK : FT_ILS;
    else if (is_adf(fn))
        sdf_type = FT_ADF;
    else if (is_mat(fn))
        sdf_type = FT_MAT;
    else if (is_wav(fn))
        sdf_type = FT_WAV;
    else if (is_snd(fn))
        sdf_type = FT_SND;
     else if (is_ascii(fn))
        sdf_type = FT_LST;
    else if (_access(fn, 0) == 0)
        sdf_type = FT_NOH;
}

int
guess_file_type(fn, sd_asc)
char   *fn;
int     sd_asc;
{
    int     ft;

    if (noheader) {
        ft = sdf_type = FT_NOH;
    } else if (is_ascii(fn) && !sd_asc) {
	ft = sdf_type = FT_SPC;
    } else if (is_tsd(fn)) {
	ft = sdf_type = FT_TSD;
    } else {
	ft = sdf_type = FT_SDF;
	guess_sdf_type(fn);
    }
    return (ft);
}

int
is_ils(fn)
char   *fn;
{
    int     img, hdsz;

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    hdsz = rdhdr_ils();
    _close(bf);			/* close the file */
    if (hdsz < 64) {
        img = 0;
    } else {
        img = (buff[62] == ILS_MAGIC || buff[62] == wxchnb(ILS_MAGIC));
        tok = (buff[63] == TOK_MAGIC || buff[63] == wxchnb(TOK_MAGIC));
    }
    return (img);
}

int
is_adf(fn)
char   *fn;
{
    ADF_fmt *adf_hdr;
    unsigned int len, status;

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    adf_hdr = (ADF_fmt *) buff;
    len = sizeof(ADF_fmt);
    status = _read(bf, (char *) adf_hdr, len);
    if (status != len) {
	return (0);
    }
    _close(bf);			/* close the file */
    if (strcmp((char *) adf_hdr, adf_key) == 0)
	return (1);

    return (0);
}

int
is_mat(fn)
char   *fn;
{
    char    type[4];
    int     bswp, hdsz;
    long    hdr[5];

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    hdsz = _read(bf, hdr, 20);
    _close(bf);			/* close the file */
    if (hdsz != 20)
	return (0);
    bswp = (hdr[0] < 0 || hdr[0] > 9999 ||
	((hdr[0] / 1000) % 10) != cpu_lsb);
    if (bswp)
	lswapb(hdr, 5);
    type[0] = (hdr[0] / 1000) % 10;		/* M */
    type[1] = (hdr[0] / 100) % 10;		/* O */
    type[2] = (hdr[0] / 10) % 10;		/* P */
    type[3] = (hdr[0] / 1) % 10;		/* T */
    if (type[3] == 0)
	if (type[0] == 0 || type[0] == 1)
	    if (type[1] == 0 || type[1] == 1)
                if (type[2] >= 0 && type[2] <= 4)
		    if (hdr[3] == 0)
			if (hdr[1] == 1 || hdr[2] == 1)
			    if (hdr[4] > 0)
				return (1);
    return (0);
}

int
is_wav(fn)
char   *fn;
{
    int     hdsz;
    long    rif_hdr[4], magic[2];

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    hdsz = _read(bf, (BUF_TYPE) rif_hdr, 12);
    _close(bf);			/* close the file */
    if (hdsz < 12)
	return (0);

    magic[0] = 0x46464952;	/* "RIFF" */
    magic[1] = 0x45564157;	/* "WAVE" */
    if (rif_hdr[0] == magic[0] && rif_hdr[2] == magic[1])
	return (1);
    lswapb(magic, 2);
    if (rif_hdr[0] == magic[0] && rif_hdr[2] == magic[1])
	return (1);

    return (0);
}

int
is_ascii(fn)
char   *fn;
{
    char   *h;
    long    offset;
    int count, rd, i;

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    count = BS * sizeof(short);
    offset = 0;
    (void) _lseek(bf, offset, 0);
    h = (char *) buff;
    rd = _read(bf, (BUF_TYPE) h, count);
    _close(bf);			/* close the file */
    for (i = 0; i < rd && h[i] > 0; i++)
	continue;
    return (i == rd);
}

int
rdhdr_ils()
{
    long    offset;
    short   hitch[4], sw;
    int count, rd;

    count = BS * sizeof(short);
    offset = 0;
    (void) _lseek(bf, offset, 0);
    rd = _read(bf, (BUF_TYPE) buff, count);
    if (rd <= 0)
	return (0);
    itok = ntok = toksiz = 0;
    if (buff[63] == TOK_MAGIC || buff[63] == wxchnb(TOK_MAGIC)) {
	sw = (buff[63] == wxchnb(TOK_MAGIC));
	if (sw)
	    wswapb(buff, BS);
	offset = (long) (buff[5] + 1) * count;
	(void) _lseek(bf, offset, 0);
	rd = _read(bf, (BUF_TYPE) hitch, 4 * sizeof(short));
	if (sw)
	    wswapb(hitch, 4);
	if (rd > 0) {
	    toksiz = hitch[1];
	    ntok = hitch[0];
	    itok = (int) fpar_.token;
	}
	count = BS * sizeof(short);
	offset = 0;
	(void) _lseek(bf, offset, 0);
	rd = _read(bf, (BUF_TYPE) buff, count);
    }
    return (rd / ((int) sizeof(short)));
}

int
open_sdf(int ft)
{
    int     ww;

    if (sdf_type == FT_ILS) {
	ww = open_ils();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_ADF) {
	ww = open_adf();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_MAT) {
	ww = open_mat();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_WAV) {
	ww = open_wav();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_SND) {
	ww = open_snd();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_LST) {
	ww = open_lst();
	fetch = (int (*) ()) fetch_list;
    } else if (sdf_type == FT_NOH) {
	ww = open_noh();
	fetch = (int (*) ()) fetch_data;
    } else if (sdf_type == FT_TOK) {
	ww = open_ils();
	fetch = (int (*) ()) fetch_data;
    } else {
	ww = 0;
    }
    return (ww);
}

int
open_ils()
{
    char    str[40];
    int     sw;
    long    sr;
    struct _stat ss;

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    if (rdhdr_ils() < 64) {
        _close(bf);
	return (0);
    }

    if (tok) {
	sw = (buff[63] == wxchnb(TOK_MAGIC));
	hs = (long) (512 + (itok - 1) * toksiz * sizeof(short));
    } else if (buff[62] == ILS_MAGIC) {
	sw = 0;
	hs = 512;
    } else if (buff[62] == wxchnb(ILS_MAGIC)) {
	sw = 1;
	hs = 512;
    } else {
        _close(bf);
	return (0);
    }
    if (sw)
	wswapb(buff, BS);	/* swap bytes */
    ts = 0;
    sr = (long) (buff[61] * pow(10.0, (double) buff[60]));
    if (fpar_.rate > 0) {
	wave_.sampsc = fpar_.rate;
    } else if (sr > 0) {
	wave_.sampsc = (float) sr;
    } else {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    if (ntok > 0) {
	wave_.nsamp = toksiz;
    } else {
	(void) _fstat(bf, &ss);
	wave_.nsamp = (ss.st_size - hs - ts) / sizeof(short);
    }
    sdf_dtyp = DTYP_LINEAR_16;
    sdf_ofst = hs;
    sdf_bswp = sw;
    sdf_scal = 1;
    setsiz();
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

static void
adf_swab(ADF_fmt *h)
{
    lswapb(h->key, 4);
    wswapb((short *) &(h->chan), 3);
    lswapb((long *) &(h->rate), 2);
    lswapb((long *) &(h->gain1), 4);
}

int
open_adf()
{
    int     bswp = 0;
    ADF_fmt *adf_hdr;
    unsigned int len, status;
    static long adf_key_0 = 0x45525343L;

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    adf_hdr = (ADF_fmt *) buff;
    len = sizeof(ADF_fmt);
    status = _read(bf, (char *) adf_hdr, len);
    if (status != len) {
	return (0);
    }
    if (adf_hdr->key[0] != adf_key_0) {
	adf_swab(adf_hdr);
	if (adf_hdr->key[0] != adf_key_0) {
	    return (0);
	}
	bswp = 1;
    }
    wave_.nsamp = adf_hdr->npts;
    wave_.sampsc = adf_hdr->rate * 1000;
    sdf_dtyp = DTYP_LINEAR_16;
    sdf_ofst = sizeof(ADF_fmt);
    sdf_bswp = bswp;
    sdf_scal = 1;
    setsiz();
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

int
open_noh()
{
    char    str[40];
    int     sw;
    struct _stat ss;

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    if (rdhdr_ils() <= 0) {
        _close(bf);
	return (0);
    }

    sw = bpar_.swap;
    if (sw)
	wswapb(buff, BS);	/* swap bytes */
    hs = (int) ((fpar_.head < 0) ? 0 : fpar_.head);
    ts = (int) fpar_.tail;

    if (fpar_.rate > 0) {
	wave_.sampsc = fpar_.rate;
    } else {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    (void) _fstat(bf, &ss);
    wave_.nsamp = (ss.st_size - hs - ts) / sizeof(short int);
    sdf_dtyp = DTYP_LINEAR_16;
    sdf_ofst = hs;
    sdf_bswp = sw;
    sdf_scal = 1;
    setsiz();
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

static double
rd_mat_val(int fd, int dtyp, int bswp)
{
    double dtmp;
    float  ftmp;
    long   ltmp;
    short  stmp;
    unsigned long utmp;
    
    if (dtyp == 0) {
        (void) _read(fd, &dtmp, 8);
        if (bswp) dswapb(&dtmp, 1);
    } else if (dtyp == 1) {
        (void) _read(fd, &ftmp, 4);
        if (bswp) lswapb((long *) &ftmp, 1);
        dtmp = ftmp;
    } else if (dtyp == 2) {
        (void) _read(fd, &ltmp, 4);
        if (bswp) lswapb(&ltmp, 1);
        dtmp = ltmp;
    } else if (dtyp == 3) {
        (void) _read(fd, &stmp, 2);
        if (bswp) wswapb((short *) &stmp, 1);
        dtmp = stmp;
    } else if (dtyp == 4) {
        (void) _read(fd, &utmp, 4);
        if (bswp) lswapb((long *) &utmp, 1);
        dtmp = utmp;
    }
    return (dtmp);
}

int
open_mat()
{
    char    str[40], nam[40], type[4];
    float   rate = 0;
    long    ofst, nsmp, hdr[5];
    unsigned int len, status;
    static  int dtyp[] = {
	DTYP_DOUBLE, DTYP_FLOAT, DTYP_LINEAR_32,
	DTYP_LINEAR_16, DTYP_LINEAR_32
    };

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    ofst = 0;
    for (;;) {
	(void) _lseek(bf, ofst, 0);
	len = 20;
	status = _read(bf, (BUF_TYPE) hdr, len);
	if (status != len) {
            _close(bf);
	    return (0);
	}
        sdf_bswp = (hdr[0] < 0 || hdr[0] > 9999 ||
            ((hdr[0] / 1000) % 10) != cpu_lsb);
	if (sdf_bswp)
	    lswapb(hdr, 5);
	type[0] = (hdr[0] / 1000) % 10;		/* M */
	type[1] = (hdr[0] / 100) % 10;		/* O */
	type[2] = (hdr[0] / 10) % 10;		/* P */
	type[3] = (hdr[0] / 1) % 10;		/* T */
	sdf_dtyp = dtyp[type[2]];
	setsiz();
	ofst += 20 + (short) hdr[4];
	len = (unsigned int) hdr[4];
	status = _read(bf, (BUF_TYPE) nam, len);
        if (strcmp(nam, "fmax") == 0) {
            rate = (float) rd_mat_val(bf, type[2], sdf_bswp) * 2;
	    fprintf(lgf, "    %s = %.0f\n", nam, rate / 2);
        } else if (strcmp(nam, "rate") == 0) {
            rate = (float) rd_mat_val(bf, type[2], sdf_bswp);
	    fprintf(lgf, "    %s = %.0f\n", nam, rate);
	} else {
	    fprintf(lgf, "    %s%s\n", nam, hdr[3] ? "*" : "");
        }
	nsmp = hdr[1] * hdr[2];
	if (nsmp > 1 && type[3] == 0 && hdr[3] == 0)
	    break;
	ofst += nsmp * sdf_bsiz;
	if (hdr[3])                         /* complex data */
	    ofst += nsmp * sdf_bsiz;
    }

    if (nsmp <= 0) {
        _close(bf);
	return (0);
    }
    if (fpar_.rate > 0) {
	wave_.sampsc = fpar_.rate;
    } else if (rate > 0) {
	wave_.sampsc = rate;
    } else {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    wave_.nsamp = nsmp;
    sdf_ofst = ofst;
    sdf_scal = scale_mat(bf, sdf_dtyp, ofst, nsmp, sdf_bswp);
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

int
read_data(int b)
{
    int     i;
    long    offset;
    int count, rd;
    static char *cbuf = (char *) buff;
    static double *dbuf = (double *) buff;
    static long *lbuf = (long *) buff;
    static float *fbuf = (float *) buff;
    static unsigned char *ucbuf = (unsigned char *) buff;

    count = sdf_sblk * sdf_bsiz;
    offset = sdf_ofst + b * ((long) count);
    (void) _lseek(bf, offset, 0);
    rd = _read(bf, (BUF_TYPE) buff, count);
    if (rd <= 0)
	return (0);
    switch (sdf_dtyp) {
    case DTYP_UNSIGNED_8:
	for (i = rd; i >= 0; i--)
	    buff[i] = (short) ucbuf[i];
	break;
    case DTYP_LINEAR_8:
	for (i = rd; i >= 0; i--)
	    buff[i] = (short) cbuf[i];
	break;
    case DTYP_MULAW_8:
	for (i = rd; i >= 0; i--)
	    buff[i] = ulaw2linear(cbuf[i]);
	break;
    case DTYP_LINEAR_16:
	if (sdf_bswp)
	    wswapb(buff, sdf_sblk);	/* swap bytes */
	break;
    case DTYP_LINEAR_32:
	if (sdf_bswp)
	    lswapb(lbuf, sdf_sblk);	/* swap bytes */
	break;
    case DTYP_FLOAT:
	if (sdf_bswp)
	    lswapb(lbuf, sdf_sblk);	/* swap bytes */
	for (i = 0; i < sdf_sblk; i++)
	    buff[i] = (int) (fbuf[i] * sdf_scal);
	break;
    case DTYP_DOUBLE:
	if (sdf_bswp)
	    dswapb(dbuf, sdf_sblk);	/* swap bytes */
	for (i = 0; i < sdf_sblk; i++)
	    buff[i] = (int) (dbuf[i] * sdf_scal);
	break;
    }
    ib = b;
    return (((int) rd) / sdf_bsiz);
}

int
fetch_data(long i)
{
    int     jb, bl, ip, nw;

    bl = (int) wave_.wbl;
    if (i < 0 || i >= wave_.nsamp) {
	if (bpar_.periodic) {
	    while (i < 0)
		i += wave_.nsamp;
	    i %= wave_.nsamp;
	} else {
	    return (bl);
	}
    }
    jb = i / sdf_sblk;
    if (jb != ib) {		/* read another block of points, if necessary */
	nw = read_data(jb);
	if (nw < sdf_sblk) {
	    for (ip = nw; ip < sdf_sblk; ip++)
		buff[ip] = bl;
	}
    }
    ip = i % sdf_sblk;
    return (buff[ip]);
}

void
close_sdf()
{
    if (sdf_type == FT_LST) {
	FREE(lstbuf);
    } else if (bf >= 0) {
	_close(bf);		/* close the file */
    }
}

int
open_lst()
{
    char    str[80];
    int     f;
    float   v, scl, maxmag;
    long    nsmp, rate, cnt;
    FILE   *ifp;

    ifp = fopen(wave_.sdfn, "r");
    if (ifp == NULL)
	return (0);
    rate = (long) fpar_.rate;
    nsmp = 0;
    f = 0;
    maxmag = 0;
    while (fgets(str, 80, ifp))
	if (valid_data(str, &v, &f)) {
	    if (maxmag < v)
		maxmag = v;
	    if (maxmag < -v)
		maxmag = -v;
	    nsmp++;
	}
    lstbuf = (SIGPTR) MALLOC(nsmp, sizeof(SIGPTR *));
    if (lstbuf == (SIGPTR) NULL) {
        fclose(ifp);
	return (0);
    }
    cnt = 0;
    scl = (float) ((f && maxmag > 0) ? 32767.0 / maxmag : 1);
    fseek(ifp, 0L, 0);		/* rewind to start of file */
    while (fgets(str, 80, ifp) && cnt < nsmp)
	if (valid_data(str, &v, &f)) {
	    lstbuf[cnt++] = (int) (f ? v * scl : v);
	} else if (strncmp("; rate=", str, 7) == 0) {
            sscanf(str + 7, "%ld", &rate);
        }
    fclose(ifp);
    if (rate > 0) {
	wave_.sampsc = (float) rate;
    } else {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    wave_.nsamp = cnt;

    return (1);
}

int
fetch_list(i)
long    i;
{
    if (i < 0 || i >= wave_.nsamp)
	return ((int) wave_.wbl);
    return (lstbuf[i]);
}

int
valid_data(s, v, f)
char   *s;
float  *v;
int    *f;
{
    while (*s == ' ' || *s == '\t')
	s++;
    if (*s < '+' || *s > '9')
	return (0);
    if (strchr(s, ',') != NULL) {
	s = strchr(s, ',') + 1;
	while (*s == ' ' || *s == '\t')
	    s++;
	if (*s < '+' || *s > '9')
	    return (0);
    }
    if (strchr(s, '.') != NULL) {
	*v = (float) atof(s);
	if (*v != 0)
	    *f = 1;
    } else {
	*v = (float) atoi(s);
    }
    return (1);
}


double
scale_mat(int file, int format, long ip, long np, int swap)
{
    double  v, vmx, sc;
    int     ns;
    long    k, j;
    unsigned int len, status;
    static char *cbuf = (char *) buff;
    static double *dbuf = (double *) buff;
    static long *lbuf = (long *) buff;
    static float *fbuf = (float *) buff;

    if (format != DTYP_FLOAT && format != DTYP_DOUBLE)
	return (1.0);
    vmx = 0;
    (void) _lseek(file, ip, 0);
    status = len = sizeof(buff);
    for (k = 0, j = 0; k < np && status == len; k += j) {
	status = _read(file, (BUF_TYPE) cbuf, len);
	switch (format) {
	case DTYP_FLOAT:
	    ns = status / 4;
	    if (swap)
		lswapb(lbuf, ns);
	    for (j = 0; j < ns && (k + j) < np; j++) {
		v = fbuf[j];
		if (vmx < v)
		    vmx = v;
		if (vmx < -v)
		    vmx = -v;
	    }
	    break;
	case DTYP_DOUBLE:
	    ns = status / 8;
	    if (swap)
		dswapb(dbuf, ns);
	    for (j = 0; j < ns && (k + j) < np; j++) {
		v = dbuf[j];
		if (vmx < v)
		    vmx = v;
		if (vmx < -v)
		    vmx = -v;
	    }
	    break;
	}
    }
    sc = (vmx > 0) ? 32767.0 / vmx : 1;
    return (sc);
}


int
rdhdr_snd()
{
    long    offset;
    int count, rd;

    sound = (SND_fmt *) buff;
    count = sizeof(SND_fmt);
    offset = 0;
    (void) _lseek(bf, offset, 0);
    rd = _read(bf, (BUF_TYPE) buff, count);
    if (rd <= 0)
	return (0);
    if (sound->key == snd_key)
	return (1);
    if (sound->key == lxchnb(snd_key)) {
	exch_hdr();		/* byte-swap header */
	sdf_bswp = 1;
	return (1);
    }
    return (0);
}

int
is_snd(fn)
char   *fn;
{
    int dfmt;
    static int sndfmt[] = {
	DTYP_UNKNOWN, DTYP_MULAW_8, DTYP_LINEAR_8, DTYP_LINEAR_16,
	DTYP_UNKNOWN, DTYP_UNKNOWN, DTYP_FLOAT, DTYP_DOUBLE,
    };
    static int nfmt = sizeof(sndfmt) / sizeof(sndfmt[0]);

    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    if (!rdhdr_snd()) {
        _close(bf);		/* close the file */
	return (0);
    }
    _close(bf);			/* close the file */

    dfmt = (int) sound->code;
    sdf_dtyp = (dfmt >= 0 && dfmt < nfmt) ? sndfmt[dfmt] : DTYP_UNKNOWN;
    sdf_ofst = sound->ofst;
    sdf_scal = 1;
    setsiz();
    wave_.nsamp = sound->size / sdf_bps[sdf_dtyp];
    wave_.sampsc = (float) sound->rate;
    return (1);
}

int
open_snd()
{
    char    str[40];
    long    ns;
    struct _stat ss;

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    if (_lseek(bf, sdf_ofst, SEEK_SET) == -1L) {
	fprintf(lgf, "SND seeking for data location failed\n");
        _close(bf);
	return (0);
    }
    (void) _fstat(bf, &ss);
    ns = (ss.st_size - sdf_ofst) / sdf_bps[sdf_dtyp];
    if (ns < wave_.nsamp)
	wave_.nsamp = ns;
    if (fpar_.rate > 0) {
	wave_.sampsc = fpar_.rate;
    } else if (wave_.sampsc <= 0) {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

int
open_wav()
{
    char    str[40];
    char   *chkid;
    short   status, bswp;
    long    rifsz, fmtsz, datsz, facsz, hdrsz;
    short   cpu_a = 1;
    char   *cpu_b = (char *) (&cpu_a);
    struct wave_fmt {
	short   format;
	short   channels;
	long    samp_sec;
	long    byte_sec;
	short   block_align;
	short   bits_smp;
    }       wfmt;

    if ((bf = _open(wave_.sdfn, oflg, 0)) < 0)
	return (0);
    bswp = cpu_b[1] ? 1 : 0;
    chkid = (char *) buff;
    chkid[4] = '\0';
    status = _read(bf, (BUF_TYPE) chkid, 4);
    if (status != 4) {
        _close(bf);
	return (0);
    }
    if (strncmp(chkid, "RIFF", 4) != 0) {
        _close(bf);
	return (0);
    }
    status = _read(bf, (BUF_TYPE) &rifsz, 4);
    if (bswp) 
	lswapb(&rifsz, 1);
    status = _read(bf, (BUF_TYPE) chkid, 4);
    if (strncmp(chkid, "WAVE", 4) != 0) {
        _close(bf);
	return (0);
    }
    status = _read(bf, (BUF_TYPE) chkid, 4);
    if (strncmp(chkid, "fmt ", 4) != 0) {
        _close(bf);
	return (0);
    }
    status = _read(bf, (BUF_TYPE) &fmtsz, 4);
    if (bswp) 
	lswapb(&fmtsz, 1);
    status = _read(bf, (BUF_TYPE) &wfmt, sizeof(wfmt));
    if (bswp) {
	wswapb(&(wfmt.format), 2);
	lswapb(&(wfmt.samp_sec), 2);
	wswapb(&(wfmt.block_align), 2);
    }
    if (fmtsz > sizeof(wfmt))
    	(void) _lseek(bf, fmtsz - sizeof(wfmt), 1);
    hdrsz = 7 * 4 + fmtsz;
    status = _read(bf, (BUF_TYPE) chkid, 4);
    while (strncmp(chkid, "fact", 4) == 0) {
        status = _read(bf, (BUF_TYPE) &facsz, 4);
	if (bswp) 
	    lswapb(&facsz, 1);
	(void) _lseek(bf, facsz, 1);
        status = _read(bf, (BUF_TYPE) chkid, 4);
        hdrsz += 2 * 4 + facsz;
    }
    if (bswp)
	lswapb(&hdrsz, 1);
    if (strncmp(chkid, "data", 4) != 0) {
        _close(bf);
	return (0);
    }
    status = _read(bf, (BUF_TYPE) &datsz, 4);
    if (bswp)
	lswapb(&datsz, 1);

    sdf_ofst = 7 * 4 + fmtsz;
    sdf_bswp = bswp;
    sdf_scal = 1;
    if (wfmt.bits_smp == 16) {
        sdf_dtyp = DTYP_LINEAR_16;
        wave_.nsamp = datsz / 2;
    } else {
        sdf_dtyp = DTYP_UNSIGNED_8;
        wave_.nsamp = datsz;
    }
    if (fpar_.rate > 0) {
	wave_.sampsc = fpar_.rate;
    } else if (wfmt.samp_sec > 0) {
	wave_.sampsc = (float) wfmt.samp_sec;
    } else {
	(void) fprintf(lgf, "samples/second: ");
	(void) fgets(str, 80, stdin);
	wave_.sampsc = (float) atof(str);
    }
    setsiz();
    ib = -1;			/* invalidate data read buffer */
    return (1);
}

void
exch_hdr()
{
    lswapb(&sound->key, 1);
    lswapb(&sound->ofst, 1);
    lswapb(&sound->size, 1);
    lswapb(&sound->code, 1);
    lswapb(&sound->rate, 1);
    lswapb(&sound->chan, 1);
}

void
wswapb(short *a, int n)
{
    _swab((char *) a, (char *) a, n * 2);
}

void
lswapb(long *a, int n)
{
    unsigned char *ch, tmp;

    while (n-- > 0) {
	ch = (unsigned char *) a++;
	tmp = ch[3];
	ch[3] = ch[0];
	ch[0] = tmp;
	tmp = ch[2];
	ch[2] = ch[1];
	ch[1] = tmp;
    }
}

void
dswapb(double *a, int n)
{
    unsigned char *ch, tmp;

    while (n-- > 0) {
	ch = (unsigned char *) a++;
	tmp = ch[7];
	ch[7] = ch[0];
	ch[0] = tmp;
	tmp = ch[6];
	ch[6] = ch[1];
	ch[1] = tmp;
	tmp = ch[5];
	ch[5] = ch[2];
	ch[2] = tmp;
	tmp = ch[4];
	ch[4] = ch[3];
	ch[3] = tmp;
    }
}

int
wxchnb(int a)
{
    short   b;
    unsigned char *ch, tmp;

    b = (short) a;
    ch = (unsigned char *) &b;
    tmp = ch[1];
    ch[1] = ch[0];
    ch[0] = tmp;
    a = (int) b;

    return (a);
}

long
lxchnb(long a)
{
    unsigned char *ch, tmp;

    ch = (unsigned char *) &a;
    tmp = ch[3];
    ch[3] = ch[0];
    ch[0] = tmp;
    tmp = ch[2];
    ch[2] = ch[1];
    ch[1] = tmp;

    return (a);
}

void
setsiz()
{
    switch (sdf_dtyp) {
	case DTYP_UNSIGNED_8:
	sdf_bsiz = 1;
	sdf_sblk = BS;
	break;
    case DTYP_LINEAR_8:
	sdf_bsiz = 1;
	sdf_sblk = BS;
	break;
    case DTYP_MULAW_8:
	sdf_bsiz = 1;
	sdf_sblk = BS;
	break;
    case DTYP_LINEAR_16:
	sdf_bsiz = 2;
	sdf_sblk = BS;
	break;
    case DTYP_LINEAR_32:
	sdf_bsiz = 4;
	sdf_sblk = BS / 2;
	break;
    case DTYP_FLOAT:
	sdf_bsiz = 4;
	sdf_sblk = BS / 2;
	break;
    case DTYP_DOUBLE:
	sdf_bsiz = 8;
	sdf_sblk = BS / 4;
	break;
    }
}
