/* tsd.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "specto.h"
#include "cparam.h"
#include "tdmdef.h"

#ifdef WIN32
#include <io.h>
#define BUF_TYPE    void *
#define OPEN_MODE   _O_RDONLY|_O_BINARY
#else
#include <unistd.h>
#define _stat       stat
#define _fstat      fstat
#define _open       open
#define _close      close
#define _read       read
#define _lseek      lseek
#define _swab       swab
#define BUF_TYPE    char *
#define OPEN_MODE   O_RDONLY
#endif

void    leave(int);
void    swab();

extern int sdf_type;
extern FILE *lgf;

float   tm0 = 0;
float   wnap = 0;
int     npts, nbyt, nstp, npm1;

static void swabw();

static int bo = 0;
static int bf = 0;
static int hs0 = 0;
static int oflg = (int) OPEN_MODE;
static float *buf;

int
rd_tsd_hdr()
{
    if (_read(bf, (BUF_TYPE) &pa_, sizeof(pa_)) < 1)
	return (0);

    if (pa_.magic != 0x74736400) {	/* "tsd" */
	if (pa_.magic != 0x00647374)
	    return (0);
	bo = 1;
    }
    if (bo) {
	_swab((char *) &pa_.hdrsiz, (char *) &pa_.hdrsiz, 6 * sizeof(short));
	_swab((char *) &pa_.m, (char *) &pa_.m, 4 * sizeof(short));
	swabw((long *) &pa_.dt, 14);
    }
    if (pa_.ireor != 0) {	/* is it a time step data file ? */
	fprintf(lgf, "ERROR:  not a time-step file\n");
	return (0);
    } else if (pa_.ifrmt != 1) {
	fprintf(lgf, "ERROR:  unsupported format: ifmrt=%d\n", pa_.ifrmt);
	return (0);
    }
    nstp = pa_.ntsf;
    nbyt = pa_.n * sizeof(float);
    npts = pa_.n;
    npm1 = pa_.n - 1;

    return (1);
}


float  *
rd_tsd_data(b)
long    b;
{
    long    offset;
    unsigned int rd;
    static long lastb = -1;

    if (b != lastb && b >= 0 && b < nstp) {
	offset = b * ((long) nbyt) + hs0;
	(void) _lseek(bf, offset, 0);
	rd = _read(bf, (BUF_TYPE) buf, nbyt);
	if (rd <= 0)
	    return (0);
	if (bo)
	    swabw((long *) buf, npts);
	lastb = b;
	wnap = buf[0];
	buf[0] = 0;
    }
    return (buf);
}

int
is_tsd(fn)
char   *fn;
{
    if ((bf = _open(fn, oflg, 0)) < 0)
	return (0);
    if (!rd_tsd_hdr())
	return (0);
    _close(bf);

    return (1);
}

int
open_tsd()
{
    long    nn;
    struct _stat ss;

    if ((bf = _open(wave_.sdfn, oflg, 0)) <= 0)
	return (0);

    if (!rd_tsd_hdr())
	return (0);

    hs0 = pa_.hdrsiz;
    tm0 = (float) pa_.ntsw * pa_.dt;
    buf = (float *) malloc(nbyt);
    if (buf == NULL) {
	fprintf(lgf, "ERROR:  can't allocate buffer for %s file\n", wave_.sdfn);
	leave(1);
    }
    (void) _fstat(bf, &ss);
    nn = (ss.st_size - hs0) / nbyt;
    if (nn < nstp)
	nstp = nn;
    sdf_type = FT_ILS;
    return (1);
}

void
close_tsd()
{
    _close(bf);			/* close the file */
    free(buf);
}


void
swabw(long *p, int n)
{
    unsigned char *ch, tmp;

    while (n-- > 0) {
	ch = (unsigned char *) p++;
	tmp = ch[3];
	ch[3] = ch[0];
	ch[0] = tmp;
	tmp = ch[2];
	ch[2] = ch[1];
	ch[1] = tmp;
    }
}
