/* w32.c - console routines for WIN32 version of SPECTO */

#include <stdio.h>
#include <io.h>
#include <math.h>
#include <direct.h>
#include <windows.h>
#include <windowsx.h>
#include "wspecto.h"
#include "specto.h"
#include "frame.h"

#ifndef WIN32
#define _access access
#define _getcwd getcwd
#define _chdir  chdir
#endif

#define XPAGE	    10.5
#define YPAGE       8.0
#define XTRN(x,y)   (rotate?(y):(x))
#define YTRN(x,y)   (rotate?(x):ypix-(y))

int     guess_file_type();
void    newext();

extern char pfn[256];
extern float dotdiam;
extern int adupi_x;
extern int adupi_y;
extern int bpp;
extern int parft;
extern int noquery;
extern int p_color;
extern int pglen;
extern int p_gmode;
extern int pgwid;
extern void (*pr_idle) (), (*pr_page) (), (*pr_out) (), (*pr_end) ();
extern void (*pr_rule) (), (*pr_line) (), (*pr_gray) ();
extern FILE *lgf;

static void gr_colm(unsigned char *, int);
static void gr_exit();
static void gr_init();
static void gr_line();
static void gr_rect();
static void gr_scrclr();

static int xbyte = 0;
static int  xpix = 640, ypix = 480;

void
co_idle()
{
}

void
co_page(int prompt)
{
    xbyte = 0;
}

void
co_out(unsigned char *buf, int n)
{
    if (xbyte == 0)
        gr_scrclr();
    gr_colm(buf, n);
    xbyte++;
}

void
co_end()
{
    xbyte = 0;
    gr_exit();
}

void
co_rule(xllc, yllc, xlen, ylen)
long    xllc, yllc, xlen, ylen;
{
    gr_rect(xllc, yllc, xlen, ylen, 255);
}

void
co_line(int n, long *x, long *y, double w)
{
    int i, x1, y1, x2, y2;

    for (i = 0; i < n; i++) {
	x2 = (int) x[i];
	y2 = (int) y[i];
	if (i > 0)
	    gr_line(x1, y1, x2, y2, w);
	x1 = x2;
	y1 = y2;
    }
}

void
co_gray(long x, long y, long w, long h, long nx, unsigned char *px, float *mp)
{
    double a0, a1, dj, m1, m2, p1, p2, pm;
    int i, j, n, y1, y2, g;
    static double rpi = 100;	// rectangles per inch

    n = (int) ((h / adupi_y) * rpi + 0.5);
    if (n < nx)
	n = nx;
    if (n > h / 2)
	n = h;
    for (i = 0; i < n; i++) {
	// interpolate i
	dj = (i * (nx - 0.5)) / (n - 1);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	m1 = a0 * mp[j] + a1 * mp[j + 1];
	p1 = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 0.5
	dj = ((i + 0.5) * (nx - 0.5)) / (n - 1);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	pm = a0 * px[j] + a1 * px[j + 1];
	// interpolate i + 1
	dj = ((i + 1) * (nx - 0.5)) / (n - 1);
	j = (int) dj;
	a1 = dj - j;
	a0 = 1 - a1;
	m2 = a0 * mp[j] + a1 * mp[j + 1];
	p2 = a0 * px[j] + a1 * px[j + 1];
	if (i == (n - 1))
	    m2 = 1;
	// compute y range and gray level
	y1 = (int) (y + h * m1 + 0.5);
	y2 = (int) (y + h * m2 + 0.5);
	g = 1 + (int) (((p1 + p2 + pm) / 3) * 253 / 255 + 0.5);
	gr_rect(x, y1, w, y2 - y1, g);
    }
}

void
co_setup()
{
    pglen = xpix;
    pgwid = ypix;
    bpp = 8;
    adupi_x = (int) (pglen / XPAGE + 0.5);
    adupi_y = (int) (pgwid / YPAGE + 0.5);
    dotdiam = (float) (2.0 / (adupi_x + adupi_y));

    pr_idle = co_idle;
    pr_page = co_page;
    pr_out = co_out;
    pr_end = co_end;
    pr_rule = co_rule;
    pr_line = co_line;
    pr_gray = NULL;
    pr_gray = co_gray;

    gr_init();
    gr_scrclr();
}

/*---------------------------------------------------------------------*/

#define FN          256
#define MAXARG      64

int specto(int, char **);

static char cwd[256];
static char szAppName[] = "Specto" ;
static char file_name[256] = "";
static char template[40] = "template.spc";
static int  loop_flag = 0;
static int  resize_flag = 0;
static int  keyval = 0;
static int  orient = 0;	    // 0=landscape, 1=portrait
static int  rotate = 0;
static int  onscrn = 1;
static int  xwin = 640, ywin = 480;
static COLORREF cc[256] = {0};
static FARPROC lpAbout;
static HANDLE hinst;
static HANDLE hAccel;
static HDC hDCMem, hDCScr, hDCDrw;
static HPEN hpen, spen;
static HWND hwnd;
static HBITMAP hOBM, hNBM;
static RECT lpSize;

static void
set_palette()
{
    int     i, m, n, r, g, b;
    static double l = 0.1, s = 1.0, hmx = 300;

    n = 1 << bpp;
    m = n - 1;
    cc[0] = RGB(255, 255, 255);
    cc[m] = RGB(0, 0, 0);
    for (i = 1; i < m; i++) {
	if (p_color == 0) {
	    r = g = b = (255 * (m  - 1 - i)) / (m - 2);
	} else {
	    if (i < (m / 2)) {
		r = 0;
		b = m - (i * 2);
	    } else {
		r = (i - (m / 2)) * 2;
		b = 0;
	    }
	    g = m - r - b;
	}
	cc[i] = RGB(r, g, b);
    }
}

/*---------------------------------------------------------------------*/

void
penwid(double pw)
{
    LOGPEN lpen;
    POINT pnpt;
    static double lpw = 0;

    if (pw != lpw) {
	pnpt.x = (int) (pw * adupi_x + 0.5);
	pnpt.y = (int) (pw * adupi_y + 0.5);
        lpen.lopnStyle = PS_SOLID;
	lpen.lopnWidth = pnpt;
	lpen.lopnColor = RGB(0, 0, 0);
	hpen = CreatePenIndirect(&lpen);
	DeleteObject(SelectPen(hDCDrw, hpen));
	if (onscrn)
	    DeleteObject(SelectPen(hDCScr, hpen));
    }
}

static void
rectangle(int x1, int y1, int x2, int y2, int c)
{
    HBRUSH  b;
    RECT    r;

    r.left = x1;
    r.bottom = y1;
    r.right = x2;
    r.top = y2;
    b = CreateSolidBrush(c);
    FillRect(hDCDrw, &r, b);
    if (onscrn)
        FillRect(hDCScr, &r, b);
    DeleteObject(b);
}

static void
line(int x1, int y1, int x2, int y2)
{
    MoveToEx(hDCDrw, x1, y1, NULL);
    LineTo(hDCDrw, x2, y2);
    SetPixel(hDCDrw, x2, y2, 0);
    if (onscrn){
        MoveToEx(hDCScr, x1, y1, NULL);
	LineTo(hDCScr, x2, y2);
	SetPixel(hDCScr, x2, y2, 0);
    }
}

void
setpix(int x, int y, int c)
{
    SetPixel(hDCDrw, x, y, c);
    if (onscrn)
        SetPixel(hDCScr, x, y, c);
}

/*---------------------------------------------------------------------*/

static void
gr_init()
{
    set_palette();
    hpen = CreatePen(PS_SOLID, 0, 0);
    spen = SelectPen(hDCDrw, hpen);
    if (onscrn) {
	hDCScr = GetDC(hwnd);
        spen = SelectPen(hDCScr, hpen);
    }
}

static void
gr_exit()
{
    DeleteObject(SelectPen(hDCDrw, spen));  // delete hpen, restore spen
    if (onscrn)
        DeleteObject(SelectPen(hDCScr, spen));
}

static void
gr_colm(unsigned char *buf, int n)
{
    int i, j, k, x, y, c, xx;

    xx = xbyte * 8;
    for (i = 0; i < n / bpp; i++) {
        y = i;
        for (j = 0; j < 8; j++) {
	    x = xx + j;
	    if (bpp == 8) {
		c = buf[i * bpp + j];
	    } else {
		c = 0;
		for (k = 0; k < bpp; k++)
		    c |= ((buf[i * bpp + k] >> (7 - j)) & 1) << (bpp - 1 - k);
	    }
	    if (c)
		setpix(XTRN(x,y), YTRN(x,y), cc[c]);
	}
    }
}

static void
gr_scrclr()
{
    rectangle(0, 0, xpix, ypix, cc[0]);
}

static void
gr_line(int x1, int y1, int x2, int y2, double w)
{
    penwid(w);
    line(XTRN(x1,y1), YTRN(x1,y1), XTRN(x2,y2), YTRN(x2,y2));
}

static void
gr_rect(int x, int y, int w, int h, int g)
{
    rectangle(XTRN(x,y), YTRN(x,y), XTRN(x+w,y+h), YTRN(x+w,y+h), cc[g]);
}

/*---------------------------------------------------------------------*/

static void
run_specto()
{
    if (__argc > 1)
        specto(__argc, __argv);
}

static int
getmsg()
{
    MSG   msg;

    while (!keyval && GetMessage(&msg, NULL, 0, 0)) {
        if (!TranslateAccelerator(hwnd, hAccel, &msg)) {
	    TranslateMessage(&msg);
	    DispatchMessage(&msg);
	}
    }
    return (keyval);
}

static void
win_open()
{
    static char *av[4];
    static int ac = 0;
    static char fn[256] = "";
    static char *title = "SPECTO files";
    static char *filter = 
        "SPC Files\0" "*.spc\0"
        "WAV Files\0" "*.wav\0"
        "ILS Files\0" "*.ils\0"
        "TSD Files\0" "*.tsd\0"
        "All Files\0" "*.*\0";
    static OPENFILENAME ofn = {0};

    // Initialize
    strcpy(fn, "");
    reset_value();
    // Fill out the OPENFILENAME struct
    ofn.lStructSize = sizeof( OPENFILENAME );
    ofn.hwndOwner = hwnd;
    ofn.hInstance = hinst;
    ofn.lpstrFile = fn;
    ofn.nMaxFile = MAX_PATH;
    ofn.Flags = OFN_FILEMUSTEXIST;
    ofn.lpstrFilter = filter;
    ofn.lpstrTitle = title;

    // Use the common dialog
    if (GetOpenFileName( &ofn )) {
	strcpy(file_name, fn);
	/* invoke SPECTO program */
	ac = 0;
	av[ac++] = szAppName;
	if (_access(template, 0) == 0 
	    && guess_file_type(file_name, 0) == FT_SDF) {
	    av[ac++] = template;
	}
	av[ac++] = file_name;
	__argc = ac;
	__argv = av;
        _getcwd(cwd, 256);
	run_specto();
    }
}

static void
edit()
{
    char cmd[256];
    static char *txted = "notepad";
    static char *waved = "wwaved";

    if (_access(pfn, 0) == 0) {
	if (parft == FT_SPC) {
	    sprintf(cmd, "%s %s", txted, pfn);
	    WinExec(cmd, 1);
	} else if (parft == FT_SDF) {
	    sprintf(cmd, "%s %s", waved, pfn);
	    WinExec(cmd, 1);
	} 
    }
}

HDC
GetPrinterDC(void)
{

    PRINTDLG pd;  // initialization data

    pd.lStructSize = sizeof(pd);
    pd.hwndOwner = hwnd;
    pd.hDevMode = NULL;
    pd.hDevNames = NULL;
    pd.hDC = 0;
    pd.Flags = PD_RETURNDC;
    pd.nFromPage = 1;
    pd.nToPage = 1;
    pd.nMinPage = 1;
    pd.nMaxPage = 1;
    pd.nCopies = 1;
    if (PrintDlg(&pd)) {
        GlobalFree(pd.hDevMode);
        GlobalFree(pd.hDevNames);
        return (pd.hDC);
    }

    return(0);
}

static void
print_spec()
{
    int	    x, y, sx, sy;
    HDC     hDCPrint;
    static DOCINFO  di = {sizeof(DOCINFO), "WSPECTO", NULL};

    hDCPrint = GetPrinterDC();
    if (hDCPrint) {
	sx = xpix;
	sy = ypix;
	if (orient) {
            x = (GetDeviceCaps(hDCPrint, LOGPIXELSX) * 8);
	    y = (GetDeviceCaps(hDCPrint, LOGPIXELSY) * 21) / 2;
	    rotate = 0;
	} else {
	    x = (GetDeviceCaps(hDCPrint, LOGPIXELSY) * 21) / 2;
            y = (GetDeviceCaps(hDCPrint, LOGPIXELSX) * 8);
	    rotate = 1;
	}
	onscrn = 0;
	xpix = x;
	ypix = (ypix * x) / sx;
	hDCDrw = hDCPrint;
        (void) StartDoc(hDCPrint, &di);
	(void) StartPage(hDCPrint);
	run_specto();
	(void) EndPage(hDCPrint);
        (void) EndDoc(hDCPrint);
        DeleteDC(hDCPrint);
	hDCDrw = hDCMem;
	rotate = 0;
	onscrn = 1;
	xpix = sx;
	ypix = sy;
	run_specto();
    }
}

static OPENFILENAME ofn;
static char     sfn[MAX_PATH] = "plt.eps";
static char    *filter = 
    "PostScript (ps)\0" "*.ps\0"
    "Enhanced Metafile (emf)\0" "*.emf\0"
    "All Files\0" "*.*\0";
static char *dext[] = {"", "ps", "emf", ""};
static char *mode[] = {"", "-Pps", "-Pemf", ""};
static UINT lastfilt = 0;

static UINT_PTR CALLBACK 
OFNHookProc(HWND hdlg, UINT uiMsg, WPARAM wParam, LPARAM lParam)
{
    if (lastfilt != ofn.nFilterIndex) {
        lastfilt = ofn.nFilterIndex;
        ofn.lpstrDefExt = dext[lastfilt];
    }
    return (0);
}

static void
export_spec()
{
    char           *av[8];
    int             ac = 0, i;
    
    strcpy(sfn,pfn);
    newext(sfn, "");
    
    // Fill out the OPENFILENAME struct
    ZeroMemory( &ofn, sizeof( OPENFILENAME ) );
    ofn.lStructSize = sizeof( OPENFILENAME );
    ofn.hwndOwner = hwnd;
    ofn.hInstance = hinst;
    ofn.lpstrFile = sfn;
    ofn.nMaxFile = MAX_PATH;
    ofn.Flags = OFN_OVERWRITEPROMPT | OFN_ENABLEHOOK | OFN_EXPLORER;
    ofn.lpstrFilter = filter;
    ofn.nFilterIndex = lastfilt;
    ofn.lpstrTitle = "PLT Save As...";
    ofn.lpfnHook = OFNHookProc;
    ofn.lpstrDefExt = "";

    // Use the common dialog
    if (GetSaveFileName( &ofn )) { 
	i = ofn.nFilterIndex;
	av[ac++] = "wspecto";
	av[ac++] = mode[i];
	av[ac++] = "-o";
	av[ac++] = sfn;
	if (p_color)
	    av[ac++] = "-c";
	av[ac++] = pfn;
        _chdir(cwd);
	specto(ac, av);
	lastfilt = i;
    }
}

static void
copy_clip()
{
    if (OpenClipboard(NULL)) {
        EmptyClipboard();
        SetClipboardData(CF_BITMAP, hNBM);
        CloseClipboard();
    }
}

/*---------------------------------------------------------------------*/

static int mlb, mrb, mcx, mcy;
static int spcflg = 1;

void
tstr(t, st)
double  t;
char   *st;
{
    if (t < 1e-5)
	sprintf(st, "%4.2f us", t * 1e6);
    else if (t < 1e-4)
	sprintf(st, "%4.1f us", t * 1e6);
    else if (t < 1e-3)
	sprintf(st, "%4.0f us", t * 1e6);
    else if (t < 1e-2)
	sprintf(st, "%4.2f ms", t * 1e3);
    else if (t < 0.1)
	sprintf(st, "%4.1f ms", t * 1e3);
    else if (t < 1)
	sprintf(st, "%4.0f ms", t * 1e3);
    else if (t < 10)
	sprintf(st, "%4.2f s", t);
    else if (t < 100)
	sprintf(st, "%4.1f s", t);
    else if (t < 1e6)
	sprintf(st, "%4.0f s", t);
    else
	sprintf(st, "INF");
}

void
fstr(f, sf)
double  f;
char   *sf;
{
    if (f > 1e12)
	sprintf(sf, "INF");
    else if (f > 1e6)
	sprintf(sf, "%4.2f MHz", f / 1e6);
    else if (f > 1e5)
	sprintf(sf, "%4.0f kHz", f / 1e3);
    else if (f > 1e4)
	sprintf(sf, "%4.1f kHz", f / 1e3);
    else if (f > 1e3)
	sprintf(sf, "%4.2f kHz", f / 1e3);
    else if (f > 1e2)
	sprintf(sf, "%4.0f Hz", f);
    else if (f > 10)
	sprintf(sf, "%4.1f Hz", f);
    else
	sprintf(sf, "%4.2f Hz", f);
}

double
pfmap(double p, int i)
{
    double  f, a, b, c, p0, p1, p2;

    switch (fr_[i].fa_typ) {
    case 1:
	f = fr_[i].fa_min * exp(p * fr_[i].fa_nat);
	break;
    case 2:
	a = fr_[i].fa_pma;
	b = fr_[i].fa_pmb;
	c = fr_[i].fa_pmc;
	p1 = log10(fr_[i].fa_min / a + c) / b;
	p2 = log10(fr_[i].fa_max / a + c) / b;
	p0 = p1 + p * (p2 - p1);
	f = a * (pow(10.0, p0 * b) - c);
	break;
    default:
	f = fr_[i].fa_min + p * (fr_[i].fa_max - fr_[i].fa_min);
    }
    return (f);
}

void
info_line(HDC hdc, POINTS mc)
{
    double  x, y, t, f, p, v, m, d, sps;
    int     i, ix1, iy1, ix2, iy2, iw1, iw2, cx, cy, iv, nwn, nft;
    char    s[80], st[20], sf[20], *sdf;
    static  double eps = 1e-9;

    if (adupi_x <= 0 || adupi_y <= 0)
	return;
    cx = mc.x;
    cy = ypix - mc.y;
    x = cx / (double) adupi_x;
    y = cy / (double) adupi_y;
    t = 0;
    f = 0;
    *s = '\0';
    for (i = 0; i < nfr; i++) {
	ix1 = fr_[i].ix_min;
	iy1 = fr_[i].iy_min;
	ix2 = fr_[i].ix_max;
	iy2 = fr_[i].iy_max;
	iw1 = fr_[i].iw_min;
	iw2 = fr_[i].iw_max;
        sdf = fr_[i].sd_fn;
        nwn = fr_[i].wn_wid;
        nft = fr_[i].wn_nft;
        sps = fr_[i].wv_sps;
	if (ix2 <= ix1 || iy2 <= iy1)
	    continue;
	if (cx >= ix1 && cx <= ix2) {
	    p = ((double) (cx - ix1)) / (ix2 - ix1);
	    t = fr_[i].ta_min + p * (fr_[i].ta_max - fr_[i].ta_min);
	    tstr(t, st);
	    if (cy >= iy1 && cy <= iy2) {
		p = ((float) (cy - iy1)) / (iy2 - iy1);
		f = pfmap(p, i);
		fstr(f, sf);
                v = spec_value(sdf, t, f, sps, nwn, nft);
                m = (fr_[i].sm_max < eps) ? eps : fr_[i].sm_max;
                d = (v < eps) ? -999 : 20 * log10(v / m);
		sprintf(s, " (t,f,v) = (%s, %s, %.1f dB) ", st, sf, d);
	    } else if (cy >= iw1 && cy <= iw2) {
                iv = wave_value(sdf, t, sps);
		sprintf(s, " (t,v) = (%s, %d) ", st, iv);
	    } else {
		sprintf(s, " t = %s ", st);
	    }
	    break;
	}
    }
    if (*s == '\0')
        sprintf(s, " (x,y) = (%3.2f in, %3.2f in)", x, y);
    for (i = strlen(s); i < 70; i++)
	s[i] = ' ';
    s[i] = '\0';
    TextOut(hdc, 0, 3, s, strlen(s));
}

void
play_sdf(char *sdf)
{
  PlaySound(sdf,NULL,SND_FILENAME);
}

void
play_wave(HDC hdc, POINTS mc)
{
    double  x, y, t, f, sps;
    int     i, ix1, iy1, ix2, iy2, iw1, iw2, cx, cy, nwn, nft;
    char    s[80], *sdf;
    static  double eps = 1e-9;

    if (adupi_x <= 0 || adupi_y <= 0)
	return;
    cx = mc.x;
    cy = ypix - mc.y;
    x = cx / (double) adupi_x;
    y = cy / (double) adupi_y;
    t = 0;
    f = 0;
    *s = '\0';
    for (i = 0; i < nfr; i++) {
	ix1 = fr_[i].ix_min;
	iy1 = fr_[i].iy_min;
	ix2 = fr_[i].ix_max;
	iy2 = fr_[i].iy_max;
	iw1 = fr_[i].iw_min;
	iw2 = fr_[i].iw_max;
        sdf = fr_[i].sd_fn;
        nwn = fr_[i].wn_wid;
        nft = fr_[i].wn_nft;
        sps = fr_[i].wv_sps;
	if (ix2 <= ix1 || iy2 <= iy1)
	    continue;
	if (cx >= ix1 && cx <= ix2) {
	    if (cy >= iy1 && cy <= iy2) {	    // spectrum
		sprintf(s, "s: sdf = %s ", sdf);
	    } else if (cy >= iw1 && cy <= iw2) {    // waveform
		sprintf(s, "w: sdf = %s ", sdf);
	    } else {				    // other
		sprintf(s, "o: sdf = %s ", sdf);
	    }
	    sprintf(s, "sdf = %s ", sdf);
	    play_sdf(sdf);
	    break;
	}
    }
    for (i = strlen(s); i < 70; i++)
	s[i] = ' ';
    s[i] = '\0';
    TextOut(hdc, 0, 3, s, strlen(s));
}

/*---------------------------------------------------------------------*/

/* SpecMenu - called on WM_COMMAND messages */

static void
SpecMenu(HWND hwnd, INT idval)
{
    HMENU hmenu;
    UINT uIDCheckItem;
    UINT uCheck;
    static char fn[256] = "";

    switch (idval) {

    case EXIT:
        PostMessage(hwnd, WM_CLOSE, 0, 0L);
        break;

    case OPEN_FILE:
        win_open(fn);
        break;

    case EDIT_FILE:
        edit() ;
        break;

    case PRINT_FILE:
	print_spec() ;
        break;

    case EXPORT_FILE:
	export_spec() ;
        break;

    case COPY_CLIP:
        copy_clip() ;
        break;

    case REPLOT:
	run_specto();
        break;

    case VIEW_COLOR:
        p_color = !p_color;
	hmenu = GetMenu(hwnd);
	uIDCheckItem = VIEW_COLOR;
	uCheck = MF_BYCOMMAND | (p_color ? MF_CHECKED : MF_UNCHECKED);
	(void) CheckMenuItem(hmenu, uIDCheckItem, uCheck);
	set_palette();
	run_specto();
        break;

    case MN_HELP_ABOUT:
        DialogBox(hinst, MAKEINTRESOURCE(ABOUTBOX), hwnd, (DLGPROC)lpAbout );
        break;

    }
}

/*---------------------------------------------------------------------*/

static void
paint_scr(HWND hWindow, PAINTSTRUCT ps)
{
    HDC  hdc;
    INT  x, y, w, h;

    x = ps.rcPaint.left;
    y = ps.rcPaint.top;
    w = ps.rcPaint.right - ps.rcPaint.left;
    h = ps.rcPaint.bottom - ps.rcPaint.top;
    hdc = GetDC(hwnd);
    BitBlt(hdc, x, y, w, h, hDCMem, x, y, SRCCOPY); 
    ReleaseDC(hwnd, hdc);
}

/*---------------------------------------------------------------------*/

static LRESULT CALLBACK 
WndProc (HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps ;

    switch (iMsg) {

    case WM_CREATE:
        break ;

    case WM_COMMAND:
        SpecMenu(hwnd, GET_WM_COMMAND_ID(wParam, lParam));
        break ;

    case WM_SIZE:
        SetWindowPos(hwnd, HWND_TOP, 0, 0, xwin, ywin, SWP_NOMOVE);
        break ;

    case WM_PAINT :
        BeginPaint(hwnd, (LPPAINTSTRUCT) &ps);
        paint_scr(hwnd, ps);
        EndPaint(hwnd, (LPPAINTSTRUCT) &ps);
        break ;

    case WM_KEYDOWN:
        if (wParam == VK_HOME) {
            keyval = (FN | 71);
        } else if (wParam == VK_END) {
            keyval = (FN | 79);
        } else if (wParam == VK_LEFT) {
            keyval = (FN | 75);
        } else if (wParam == VK_RIGHT) {
            keyval = (FN | 77);
        } else if (wParam == VK_UP) {
            keyval = (FN | 72);
        } else if (wParam == VK_DOWN) {
            keyval = (FN | 80);
        }
        break ;

    case WM_LBUTTONDOWN:
        info_line(GetDC(hwnd), MAKEPOINTS(lParam));
        break ;

    case WM_RBUTTONDOWN:
        play_wave(GetDC(hwnd), MAKEPOINTS(lParam));
        break ;

    case WM_DESTROY :
        PostQuitMessage (0) ;
	keyval = 3;
	break ;

    default:
	return DefWindowProc (hwnd, iMsg, wParam, lParam) ;
    }
    return 0 ;
}

/*
 *  FUNCTION: AboutDlgProc(HWND, WORD, WORD, LONG)
 *
 *  PURPOSE:  Processes messages for "About" dialog box
 *
 *  MESSAGES:
 *
 *      WM_INITDIALOG - initialize dialog box
 *      WM_COMMAND    - Input received
 */

static LRESULT CALLBACK
AboutDlgProc(HWND hDlg, WORD message, WORD wParam, LONG lParam)
{
    if(message == WM_COMMAND) {
        EndDialog(hDlg, TRUE);
        return(TRUE);
    }
    if (message == WM_INITDIALOG)
        return(TRUE);
    else
        return(FALSE);
}

static void
win_size()
{
    HDC     hdc;
  
    GetClientRect(hwnd, (LPRECT) &lpSize);
    xpix = lpSize.right - lpSize.left;
    ypix = lpSize.bottom - lpSize.top;

    hdc = GetDC(hwnd);
    hDCMem = CreateCompatibleDC(hdc);
    hNBM = CreateCompatibleBitmap(hdc,xpix,ypix); 
    hOBM = SelectObject(hDCMem, hNBM);
    ReleaseDC(hwnd, hdc);
    FillRect(hDCMem, &lpSize, GetStockObject(GRAY_BRUSH));
    hDCDrw = hDCMem;
}

static int
win_init(HINSTANCE hI, char *title)
{
    int         ws;
    LPSTR       cn = "specto";
    WNDCLASS    wndclass;

    if (hinst == 0) {
        hinst = hI;
        hAccel = LoadAccelerators(hinst, (LPSTR)"MAINACC");
        lpAbout = MakeProcInstance((FARPROC)AboutDlgProc, hinst);
        wndclass.style         = 0 ;
        wndclass.lpfnWndProc   = WndProc ;
	wndclass.cbClsExtra    = 0 ;
        wndclass.cbWndExtra    = 0 ;
        wndclass.hInstance     = hinst;
        wndclass.hIcon         = LoadIcon(hinst, MAKEINTRESOURCE(WSPECTO_ICON)) ;
        wndclass.hCursor       = LoadCursor(NULL, IDC_ARROW) ;
        wndclass.hbrBackground = GetStockObject(GRAY_BRUSH);
        wndclass.lpszMenuName  = MAKEINTRESOURCE(WSPECTO) ;
        wndclass.lpszClassName = cn;
        if(!RegisterClass(&wndclass))
            return(FALSE);
    }
    if (hwnd == 0) {
        if (resize_flag)
            ws = WS_OVERLAPPEDWINDOW;
        else
            ws = WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX;
        xwin = GetSystemMetrics(SM_CXSCREEN) * 7 / 8,
        ywin = (int) (xwin * 8.0 / 10.5);
        hwnd = CreateWindow(cn, (LPSTR) title, ws, CW_USEDEFAULT, 0,
            xwin, ywin, (HANDLE) NULL, (HANDLE) NULL, (HANDLE) hinst, NULL);
    }

    ShowWindow(hwnd, SW_SHOW);
    UpdateWindow(hwnd);

    // Initialize bitmaps
    win_size();

    _getcwd(cwd, 256);

    return(TRUE);
}

/*---------------------------------------------------------------------*/

int WINAPI 
WinMain (HINSTANCE hI, HINSTANCE hPI, char * CL, int CS)
{
    win_init(hI, szAppName);
    run_specto();
    while (keyval != 3)
	getmsg();

    return (0);
}
